@@ -8,12 +8,61 @@
 (defpackage #:sb-cover
   (:use #:cl #:sb-c)
   (:export #:report
+           #:get-coverage
            #:reset-coverage #:clear-coverage
            #:restore-coverage #:restore-coverage-from-file
            #:save-coverage #:save-coverage-in-file
            #:store-coverage-data))
 
 (in-package #:sb-cover)
+(eval-when (:compile-toplevel :load-toplevel :execute)
+  (setf (sb-int:system-package-p *package*) t))
+
+(defmacro code-coverage-hashtable () `(car sb-c:*code-coverage-info*))
+
+;;;; New coverage representation.
+;;;; One byte per coverage mark is stored in the unboxed constants of the code.
+;;;; x86[-64] use a slightly different but not significantly different
+;;; representation of the marks from other architectures.
+(defun %find-coverage-map (code)
+  (declare (type (or sb-kernel:simple-fun
+                     sb-kernel:code-component
+                     symbol)
+                 code))
+  (etypecase code
+   (sb-kernel:simple-fun
+    (%find-coverage-map (sb-kernel:fun-code-header code)))
+   (symbol
+    (%find-coverage-map (or (macro-function code) (fdefinition code))))
+   (sb-kernel:code-component
+    (let ((n (sb-kernel:code-header-words code)))
+      (let ((map (sb-kernel:code-header-ref code (1- n))))
+        (when (typep map '(cons (eql sb-c::coverage-map)))
+          (return-from %find-coverage-map (values (cdr map) code))))
+      ;; if code-boxed-words can't be an odd number, try one more slot
+      #-(or x86 x86-64)
+      (let ((map (sb-kernel:code-header-ref code (- n 2))))
+        (when (typep map '(cons (eql sb-c::coverage-map)))
+          (return-from %find-coverage-map (values (cdr map) code))))))))
+
+#+arm64
+(declaim (ftype (sb-int:sfunction (t) (simple-array (unsigned-byte 8) (*))) code-coverage-marks))
+;;; Coverage marks are in the raw bytes following the jump tables
+;;; preceding any other unboxed constants. This way we don't have to store
+;;; a pointer to the coverage marks since their location is implicit.
+(defun code-coverage-marks (code)
+  #-arm64
+  (let ((insts (sb-kernel:code-instructions code)))
+    (sb-sys:sap+ insts (ash (sb-kernel:code-jump-table-words code)
+                            sb-vm:word-shift)))
+  #+arm64
+  (let* ((words (sb-kernel:code-header-words code))
+         (last (sb-kernel:code-header-ref code (- words 2))))
+    (if (vectorp last)
+        last
+        (sb-kernel:code-header-ref code (- words 3)))))
+
+;;;;
 
 (declaim (type (member :whole :car) *source-path-mode*))
 (defvar *source-path-mode* :whole)
@@ -28,11 +77,121 @@
 into the database when the FASL files (produced by compiling
 STORE-COVERAGE-DATA optimization policy set to 3) are loaded again into the
 image."
-  (sb-c::clear-code-coverage))
+  (sb-c:clear-code-coverage))
 
-(defun reset-coverage ()
+(macrolet
+    ((do-instrumented-code ((var &optional result) &body body)
+       ;; Scan list of weak-pointers to all coverage-instrumented code,
+       ;; binding VAR to each object, and removing broken weak-pointers.
+       `(let ((predecessor sb-c:*code-coverage-info*))
+          (loop
+           (let ((cell (cdr predecessor)))
+             (unless cell (return ,result))
+             (let ((,var (sb-ext:weak-pointer-value (car cell))))
+               (if ,var
+                   (progn ,@body (setq predecessor cell))
+                   (rplacd predecessor (cdr cell))))))))
+     (empty-mark-word ()
+       #+(or x86-64 x86) 0
+       #-(or x86-64 x86) sb-ext:most-positive-word)
+     (byte-marked-p (byte)
+       #+(or x86-64 x86) `(/= ,byte 0)
+       #-(or x86-64 x86) `(/= ,byte #xFF)))
+
+;;; Retun just the list of soure paths in CODE that are marked covered.
+(defun get-coverage (code)
+  (multiple-value-bind (map code) (%find-coverage-map code)
+    (when map
+      (sb-int:collect ((paths))
+        #-arm64
+        (sb-sys:with-pinned-objects (code)
+          (let ((sap (code-coverage-marks code)))
+            (dotimes (i (length map) (paths))
+              (when (byte-marked-p (sb-sys:sap-ref-8 sap i))
+                (paths (svref map i))))))
+        #+arm64
+        (let ((marks (code-coverage-marks code)))
+          (dotimes (i (length map) (paths))
+            (when (byte-marked-p (aref marks i))
+              (paths (svref map i)))))))))
+
+(defun reset-coverage (&optional object)
   "Reset all coverage data back to the `Not executed` state."
-  (sb-c::reset-code-coverage))
+  (cond (object ; reset only this object
+         (multiple-value-bind (map code) (%find-coverage-map object)
+           (when map
+             #-arm64
+             (sb-sys:with-pinned-objects (code)
+               (let ((sap (code-coverage-marks code)))
+                 (dotimes (i (ceiling (length map) sb-vm:n-word-bytes))
+                   (setf (sb-sys:sap-ref-word sap (ash i sb-vm:n-word-bytes))
+                         (empty-mark-word)))))
+             #+arm64
+             (fill (code-coverage-marks code) #xFF))))
+        (t                              ; reset everything
+         (do-instrumented-code (code)
+           (reset-coverage code))
+         (sb-c:reset-code-coverage))))
+
+;;; Transfer data from new-style coverage marks into old-style.
+;;; Update only data for FILENAME if supplied, or all files if NIL.
+;;; Ideally we could report directly from the new data where applicable,
+;;; however this is, for the time being, perfectly backward-compatibile.
+(defun refresh-coverage-info (&optional filename)
+  (declare (ignorable filename))
+  ;; NAMESTRING->PATH-TABLES maps a namestring to a hashtable which maps
+  ;; source paths to the legacy coverage record for that path in that file,
+  ;;   e.g. (1 4 1) -> ((1 4 1) . SB-C::%CODE-COVERAGE-UNMARKED%)
+  (let ((namestring->path-tables (make-hash-table :test 'equal))
+        (coverage-records (code-coverage-hashtable))
+        (n-marks 0))
+    (do-instrumented-code (code)
+      (sb-int:binding* ((map (%find-coverage-map code) :exit-if-null)
+                        (namestring
+                         (sb-int:debug-source-namestring
+                          (sb-c::debug-info-source (sb-kernel:%code-debug-info code)))
+                         :exit-if-null)
+                        (legacy-coverage-marks
+                         (and (or (null filename) (string= namestring filename))
+                              (gethash namestring coverage-records))
+                         :exit-if-null)
+                        (path-lookup-table
+                         (gethash namestring namestring->path-tables)))
+        ;; Build the source path -> marked map for this file if not seen yet.
+        ;; It is of course redundant to have both representations.
+        (unless path-lookup-table
+          (setf path-lookup-table (make-hash-table :test 'equal)
+                (gethash namestring namestring->path-tables) path-lookup-table)
+          (dolist (item legacy-coverage-marks)
+            (setf (gethash (car item) path-lookup-table) item)))
+        #-arm64
+        (sb-sys:with-pinned-objects (code)
+          (let ((sap (code-coverage-marks code)))
+            (dotimes (i (length map))   ; for each recorded mark
+              (when (byte-marked-p (sb-sys:sap-ref-8 sap i))
+                (incf n-marks)
+                ;; Set the legacy coverage mark for each path it touches
+                ;; One mark byte for x86[-64] corresponds to a union of source paths.
+                ;; For everybody else, one mark is one path.
+                (dolist (path #+(or x86 x86-64) (svref map i)
+                              #-(or x86 x86-64) (list (svref map i)))
+                  (let ((found (gethash path path-lookup-table)))
+                    (if found
+                        (rplacd found t)
+                        #+nil
+                        (warn "Missing coverage entry for ~S in ~S"
+                              path namestring))))))))
+        #+arm64
+        (let ((marks (code-coverage-marks code)))
+          (dotimes (i (length map))     ; for each recorded mark
+            (when (byte-marked-p (aref marks i))
+              (incf n-marks)
+              (let ((found (gethash (svref map i) path-lookup-table)))
+                (if found
+                    (rplacd found t))))))))
+    (values coverage-records n-marks)))
+
+) ; end MACROLET
 
 (defun save-coverage ()
   "Returns an opaque representation of the current code coverage state.
@@ -40,15 +199,17 @@ The only operation that may be done on the state is passing it to
 RESTORE-COVERAGE. The representation is guaranteed to be readably printable.
 A representation that has been printed and read back will work identically
 in RESTORE-COVERAGE."
-  (loop for file being the hash-keys of sb-c::*code-coverage-info*
+  (refresh-coverage-info)
+  (loop for file being the hash-keys of (code-coverage-hashtable)
         using (hash-value states)
         collect (cons file states)))
 
 (defun restore-coverage (coverage-state)
   "Restore the code coverage data back to an earlier state produced by
 SAVE-COVERAGE."
+  ;; This does not update coverage stored in code object headers
   (loop for (file . states) in coverage-state
-        do (let ((image-states (gethash file sb-c::*code-coverage-info*))
+        do (let ((image-states (gethash file (code-coverage-hashtable)))
                  (table (make-hash-table :test 'equal)))
              (when image-states
                (loop for cons in image-states
@@ -79,7 +240,21 @@ result to RESTORE-COVERAGE."
         (restore-coverage (read stream))))
     (values)))
 
+(defun pathname-as-directory (pathname &optional (errorp t))
+  (let ((pathname (merge-pathnames pathname)))
+    (if (and (member (pathname-name pathname) '(nil :unspecific))
+             (member (pathname-type pathname) '(nil :unspecific)))
+        pathname
+        (if errorp
+            (error "~S does not designate a directory" pathname)
+            (make-pathname :directory (append (or (pathname-directory pathname)
+                                                  (list :relative))
+                                              (list (file-namestring pathname)))
+                           :name nil :type nil :version nil
+                           :defaults pathname)))))
+
 (defun report (directory &key ((:form-mode *source-path-mode*) :whole)
+               (if-matches 'identity)
                (external-format :default))
   "Print a code coverage report of all instrumented files into DIRECTORY.
 If DIRECTORY does not exist, it will be created. The main report will be
@@ -90,26 +265,39 @@ If the keyword argument FORM-MODE has the value :CAR, the annotations in
 the coverage report will be placed on the CARs of any cons-forms, while if
 it has the value :WHOLE the whole form will be annotated (the default).
 The former mode shows explicitly which forms were instrumented, while the
-latter mode is generally easier to read."
-  (let ((paths)
-        (*default-pathname-defaults* (merge-pathnames (pathname directory))))
-    (ensure-directories-exist *default-pathname-defaults*)
+latter mode is generally easier to read.
+
+The keyword argument IF-MATCHES should be a designator for a function
+of one argument, called for the namestring of each file with code
+coverage info. If it returns true, the file's info is included in the
+report, otherwise ignored. The default value is CL:IDENTITY.
+"
+  (let* ((paths)
+         (directory (pathname-as-directory directory))
+         (defaults (translate-logical-pathname directory)))
+    (ensure-directories-exist defaults)
+    (when (eq if-matches 'identity)
+      (refresh-coverage-info)) ; update all
     (maphash (lambda (k v)
                (declare (ignore v))
-               (let* ((n (format nil "~(~{~2,'0X~}~)"
-                                (coerce (sb-md5:md5sum-string
-                                         (sb-ext:native-namestring k))
-                                        'list)))
-                      (path (make-pathname :name n :type "html")))
-                 (when (probe-file k)
-                   (with-open-file (stream path
-                                           :direction :output
-                                           :if-exists :supersede
-                                           :if-does-not-exist :create)
-                     (push (list* k n (report-file k stream external-format))
-                           paths)))))
-             *code-coverage-info*)
-    (let ((report-file (make-pathname :name "cover-index" :type "html")))
+               (when (funcall if-matches k)
+                 (unless (eq if-matches 'identity)
+                   (refresh-coverage-info k)) ; update one file
+                 (let* ((pk (translate-logical-pathname k))
+                        (n (format nil "~(~{~2,'0X~}~)"
+                                   (coerce (sb-md5:md5sum-string
+                                            (sb-ext:native-namestring pk))
+                                           'list))))
+                   (when (probe-file k)
+                     (with-open-file (stream (make-pathname :name n :type "html"
+                                                            :defaults directory)
+                                             :direction :output
+                                             :if-exists :supersede
+                                             :if-does-not-exist :create)
+                       (push (list* k n (report-file k stream external-format))
+                             paths))))))
+             (code-coverage-hashtable))
+    (let ((report-file (make-pathname :name "cover-index" :type "html" :defaults directory)))
       (with-open-file (stream report-file
                               :direction :output :if-exists :supersede
                               :if-does-not-exist :create)
@@ -153,6 +341,11 @@ latter mode is generally easier to read."
     (* 100
        (/ (ok-of count) (all-of count)))))
 
+;; This special is bound to the package in which forms will be read,
+;; and reassigned as read-and-record-source-map detects what look like
+;; "IN-PACKAGE" forms.
+(defvar *current-package*)
+
 (defun report-file (file html-stream external-format)
   "Print a code coverage report of FILE into the stream HTML-STREAM."
   (format html-stream "<html><head>")
@@ -162,15 +355,15 @@ latter mode is generally easier to read."
          (states (make-array (length source)
                              :initial-element 0
                              :element-type '(unsigned-byte 4)))
+         (hashtable (code-coverage-hashtable))
          ;; Convert the code coverage records to a more suitable format
          ;; for this function.
-         (expr-records (convert-records (gethash file *code-coverage-info*)
-                                        :expression))
-         (branch-records (convert-records (gethash file *code-coverage-info*)
-                                          :branch))
+         (expr-records (convert-records (gethash file hashtable) :expression))
+         (branch-records (convert-records (gethash file hashtable) :branch))
          ;; Cache the source-maps
          (maps (with-input-from-string (stream source)
-                 (loop with map = nil
+                 (loop with *current-package* = (find-package "CL-USER")
+                       with map = nil
                        with form = nil
                        with eof = nil
                        for i from 0
@@ -393,7 +586,7 @@ table.summary tr.subheading td { text-align: left; font-weight: bold; padding-le
            unless (member (caar record) '(:then :else))
            collect (list mode
                          (car record)
-                         (if (sb-c::code-coverage-record-marked record)
+                         (if (sb-c:code-coverage-record-marked record)
                              1
                              2))))
     (:branch
@@ -403,7 +596,7 @@ table.summary tr.subheading td { text-align: left; font-weight: bold; padding-le
            (when (member (car path) '(:then :else))
              (setf (gethash (cdr path) hash)
                    (logior (gethash (cdr path) hash 0)
-                           (ash (if (sb-c::code-coverage-record-marked record)
+                           (ash (if (sb-c:code-coverage-record-marked record)
                                     1
                                     2)
                                 (if (eql (car path) :then)
@@ -429,7 +622,7 @@ table.summary tr.subheading td { text-align: left; font-weight: bold; padding-le
   "Return a macro character function that does the same as FN, but
 additionally stores the result together with the stream positions
 before and after of calling FN in the hashtable SOURCE-MAP."
-  (declare (type function fn))
+  (declare (type (or function symbol) fn))
   (lambda (stream char)
     (declare (optimize debug safety))
     (let ((start (file-position stream))
@@ -445,13 +638,17 @@ before and after of calling FN in the hashtable SOURCE-MAP."
 The source locations are stored in SOURCE-MAP."
   (let* ((tab (copy-readtable readtable))
          (*readtable* tab))
+    ;; It is unspecified whether doing (SET-MACRO-CHARACTER c1 fn)
+    ;; and then (SET-DISPATCH-MACRO-CHARACTER c1 c2 fn) should be allowed.
+    ;; Portability concerns aside, it doesn't work in the latest code,
+    ;; but first changing the function for #. and then # in that order works.
+    (suppress-sharp-dot tab)
     (dotimes (code 128)
       (let ((char (code-char code)))
         (multiple-value-bind (fn term) (get-macro-character char tab)
           (when fn
             (set-macro-character char (make-source-recorder fn source-map)
                                  term tab)))))
-    (suppress-sharp-dot tab)
     (set-macro-character #\(
                          (make-source-recorder
                           (make-recording-read-list source-map)
@@ -482,15 +679,17 @@ The source locations are stored in SOURCE-MAP."
                            ((sb-impl::whitespace[2]p nextchar)
                             (setq nextchar (sb-impl::flush-whitespace stream))))
                      (rplacd listtail
-                             ;; Return list containing last thing.
-                             (car (sb-impl::read-after-dot stream nextchar)))
+                             (sb-impl::read-after-dot
+                              stream nextchar (if *read-suppress* 0 -1)))
                      (return (cdr thelist)))
                     ;; Put back NEXTCHAR so that we can read it normally.
                     (t (unread-char nextchar stream)))))
           ;; Next thing is not an isolated dot.
-          (let ((start (file-position stream))
-                (listobj (sb-impl::read-maybe-nothing stream firstchar))
-                (end (file-position stream)))
+          (sb-int:binding*
+              ((start (file-position stream))
+               ((winp obj) (sb-impl::read-maybe-nothing stream firstchar))
+               (listobj (if (not (zerop winp)) (list obj)))
+               (end (file-position stream)))
             ;; allows the possibility that a comment was read
             (when listobj
              (unless (consp (car listobj))
@@ -509,15 +708,51 @@ The source locations are stored in SOURCE-MAP."
                                         (apply sharp-dot args)))
                                     readtable))))
 
+;;; The detection logic for "IN-PACKAGE" is stolen from swank's
+;;; source-path-parser.lisp.
+;;;
+;;; We look for lines that start with "(in-package " or
+;;; "(cl:in-package ", without leading whitespace.
+(defun starts-with-p (string prefix)
+  (declare (type string string prefix))
+  (not (mismatch string prefix
+                 :end1 (min (length string) (length prefix))
+                 :test #'char-equal)))
+
+(defun extract-package (line)
+  (declare (type string line))
+  (let ((*package* *current-package*))
+    (second (read-from-string line))))
+
+(defun look-for-in-package-form (string)
+  (when (or (starts-with-p string "(in-package ")
+            (starts-with-p string "(cl:in-package "))
+    (let ((package (find-package (extract-package string))))
+      (when package
+        (setf *current-package* package)))))
+
+(defun look-for-in-package-form-in-stream (stream start-position end-position)
+  "Scans the stream between start-position up to end-position for
+   something that looks like an in-package form. If it does find
+   something, the function updates *current-package*. In all cases,
+   the stream is reset to end-position on exit."
+  (assert (file-position stream start-position))  ; rewind the stream
+  (loop until (>= (file-position stream) end-position)
+     do (look-for-in-package-form (or (read-line stream nil)
+                                      (return))))
+  (assert (file-position stream end-position)))
+
 (defun read-and-record-source-map (stream)
   "Read the next object from STREAM.
 Return the object together with a hashtable that maps
 subexpressions of the object to stream positions."
   (let* ((source-map (make-hash-table :test #'eq))
-         (*readtable* (make-source-recording-readtable *readtable* source-map))
          (start (file-position stream))
-         (form (read stream))
+         (form (let ((*readtable* (make-source-recording-readtable *readtable* source-map))
+                     (*package* *current-package*))
+                 (read stream)))
          (end (file-position stream)))
+    (look-for-in-package-form-in-stream stream start end)
     ;; ensure that at least FORM is in the source-map
     (unless (gethash form source-map)
       (push (list start end nil)
