@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;;; DEFKNOWNs
 
@@ -51,8 +51,6 @@
   ())
 (defknown %local-alien-addr (local-alien-info t) (alien (* t))
   (flushable movable))
-(defknown dispose-local-alien (local-alien-info t) t
-  ())
 
 (defknown %cast (alien-value alien-type) alien
   (flushable movable))
@@ -63,13 +61,21 @@
   (flushable movable))
 (defknown deport-alloc (alien alien-type) t
   (flushable movable))
-(defknown extract-alien-value (system-area-pointer unsigned-byte alien-type) t
+(defknown %alien-value (system-area-pointer unsigned-byte alien-type) t
   (flushable))
-(defknown deposit-alien-value (system-area-pointer unsigned-byte alien-type t) t
+(defknown (setf %alien-value) (t system-area-pointer unsigned-byte alien-type) t
   ())
 
-(defknown alien-funcall (alien-value &rest *) *
+(defknown alien-funcall (alien-value &rest t) *
   (any recursive))
+
+(defknown sb-alien::string-to-c-string (simple-string t) (or (simple-array (unsigned-byte 8) (*))
+                                                             simple-base-string)
+    (movable flushable))
+(defknown sb-alien::c-string-to-string (system-area-pointer t t) simple-string
+    (movable flushable))
+(defknown sb-alien::c-string-external-format * *
+        (movable flushable))
 
 ;;;; cosmetic transforms
 
@@ -101,9 +107,8 @@
              (field (find slot-name (alien-record-type-fields alien-type)
                           :key #'alien-record-field-name)))
         (unless field
-          (abort-ir1-transform "~S doesn't have a slot named ~S"
-                               alien
-                               slot-name))
+          (abort-ir1-transform "~S~% doesn't have a slot named ~S"
+                               type slot-name))
         (values (alien-record-field-offset field)
                 (alien-record-field-type field))))))
 
@@ -117,12 +122,12 @@
         (return (make-alien-type-type slot-type))))
     *wild-type*))
 
-(deftransform slot ((alien slot) * * :important t)
+(deftransform slot ((alien slot))
   (multiple-value-bind (slot-offset slot-type)
       (find-slot-offset-and-type alien slot)
-    `(extract-alien-value (alien-sap alien)
-                          ,slot-offset
-                          ',slot-type)))
+    `(%alien-value (alien-sap alien)
+                   ,slot-offset
+                   ',slot-type)))
 
 #+nil ;; ### But what about coercions?
 (defoptimizer (%set-slot derive-type) ((alien slot value))
@@ -136,13 +141,13 @@
           (return type))))
     *wild-type*))
 
-(deftransform %set-slot ((alien slot value) * * :important t)
+(deftransform %set-slot ((alien slot value))
   (multiple-value-bind (slot-offset slot-type)
       (find-slot-offset-and-type alien slot)
-    `(deposit-alien-value (alien-sap alien)
-                          ,slot-offset
-                          ',slot-type
-                          value)))
+    `(setf (%alien-value (alien-sap alien)
+                         ,slot-offset
+                         ',slot-type)
+           value)))
 
 (defoptimizer (%slot-addr derive-type) ((alien slot))
   (block nil
@@ -154,11 +159,11 @@
                  (make-alien-pointer-type :to slot-type)))))
     *wild-type*))
 
-(deftransform %slot-addr ((alien slot) * * :important t)
+(deftransform %slot-addr ((alien slot))
   (multiple-value-bind (slot-offset slot-type)
       (find-slot-offset-and-type alien slot)
     (/noshow "in DEFTRANSFORM %SLOT-ADDR, creating %SAP-ALIEN")
-    `(%sap-alien (sap+ (alien-sap alien) (/ ,slot-offset sb!vm:n-byte-bits))
+    `(%sap-alien (sap+ (alien-sap alien) (/ ,slot-offset sb-vm:n-byte-bits))
                  ',(make-alien-pointer-type :to slot-type))))
 
 ;;;; DEREF support
@@ -167,10 +172,7 @@
   (let ((alien-type (lvar-type alien)))
     (unless (alien-type-type-p alien-type)
       (give-up-ir1-transform))
-    (let ((alien-type (alien-type-type-alien-type alien-type)))
-      (if (alien-type-p alien-type)
-          alien-type
-          (give-up-ir1-transform)))))
+    (alien-type-type-alien-type alien-type)))
 
 (defun find-deref-element-type (alien)
   (let ((alien-type (find-deref-alien-type alien)))
@@ -217,7 +219,7 @@
            (give-up-ir1-transform "Element alignment is unknown."))
          (if (null dims)
              (values nil 0 element-type)
-             (let* ((arg (gensym))
+             (let* ((arg (sb-xc:gensym))
                     (args (list arg))
                     (offsetexpr arg))
                (dolist (dim (cdr dims))
@@ -240,13 +242,13 @@
       (return (make-alien-type-type (find-deref-element-type alien))))
     *wild-type*))
 
-(deftransform deref ((alien &rest indices) * * :important t)
+(deftransform deref ((alien &rest indices))
   (multiple-value-bind (indices-args offset-expr element-type)
       (compute-deref-guts alien indices)
     `(lambda (alien ,@indices-args)
-       (extract-alien-value (alien-sap alien)
-                            ,offset-expr
-                            ',element-type))))
+       (%alien-value (alien-sap alien)
+                     ,offset-expr
+                     ',element-type))))
 
 #+nil ;; ### Again, the value might be coerced.
 (defoptimizer (%set-deref derive-type) ((alien value &rest noise))
@@ -260,14 +262,14 @@
         (return type)))
     *wild-type*))
 
-(deftransform %set-deref ((alien value &rest indices) * * :important t)
+(deftransform %set-deref ((alien value &rest indices))
   (multiple-value-bind (indices-args offset-expr element-type)
       (compute-deref-guts alien indices)
     `(lambda (alien value ,@indices-args)
-       (deposit-alien-value (alien-sap alien)
-                            ,offset-expr
-                            ',element-type
-                            value))))
+       (setf (%alien-value (alien-sap alien)
+                           ,offset-expr
+                           ',element-type)
+             value))))
 
 (defoptimizer (%deref-addr derive-type) ((alien &rest noise))
   (declare (ignore noise))
@@ -278,12 +280,12 @@
                 :to (find-deref-element-type alien)))))
     *wild-type*))
 
-(deftransform %deref-addr ((alien &rest indices) * * :important t)
+(deftransform %deref-addr ((alien &rest indices))
   (multiple-value-bind (indices-args offset-expr element-type)
       (compute-deref-guts alien indices)
     (/noshow "in DEFTRANSFORM %DEREF-ADDR, creating (LAMBDA .. %SAP-ALIEN)")
     `(lambda (alien ,@indices-args)
-       (%sap-alien (sap+ (alien-sap alien) (/ ,offset-expr sb!vm:n-byte-bits))
+       (%sap-alien (sap+ (alien-sap alien) (/ ,offset-expr sb-vm:n-byte-bits))
                    ',(make-alien-pointer-type :to element-type)))))
 
 ;;;; support for aliens on the heap
@@ -304,9 +306,9 @@
         (return (make-alien-type-type type))))
     *wild-type*))
 
-(deftransform %heap-alien ((info) * * :important t)
+(deftransform %heap-alien ((info) ((constant-arg heap-alien-info)) *)
   (multiple-value-bind (sap type) (heap-alien-sap-and-type info)
-    `(extract-alien-value ,sap 0 ',type)))
+    `(%alien-value ,sap 0 ',type)))
 
 #+nil ;; ### Again, deposit value might change the type.
 (defoptimizer (%set-heap-alien derive-type) ((info value))
@@ -319,9 +321,9 @@
           (return type))))
     *wild-type*))
 
-(deftransform %set-heap-alien ((info value) (heap-alien-info *) * :important t)
+(deftransform %set-heap-alien ((info value) (heap-alien-info t) *)
   (multiple-value-bind (sap type) (heap-alien-sap-and-type info)
-    `(deposit-alien-value ,sap 0 ',type value)))
+    `(setf (%alien-value ,sap 0 ',type) value)))
 
 (defoptimizer (%heap-alien-addr derive-type) ((info))
   (block nil
@@ -331,7 +333,7 @@
         (return (make-alien-type-type (make-alien-pointer-type :to type)))))
     *wild-type*))
 
-(deftransform %heap-alien-addr ((info) * * :important t)
+(deftransform %heap-alien-addr ((info))
   (multiple-value-bind (sap type) (heap-alien-sap-and-type info)
     (/noshow "in DEFTRANSFORM %HEAP-ALIEN-ADDR, creating %SAP-ALIEN")
     `(%sap-alien ,sap ',(make-alien-pointer-type :to type))))
@@ -339,9 +341,12 @@
 
 ;;;; support for local (stack or register) aliens
 
-(deftransform make-local-alien ((info) * * :important t)
+(defun alien-info-constant-or-abort (info)
   (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+    (abort-ir1-transform "Local alien info isn't constant?")))
+
+(deftransform make-local-alien ((info))
+  (alien-info-constant-or-abort info)
   (let* ((info (lvar-value info))
          (alien-type (local-alien-info-type info))
          (bits (alien-type-bits alien-type)))
@@ -351,112 +356,91 @@
     (/noshow (local-alien-info-force-to-memory-p info))
     (/noshow alien-type (unparse-alien-type alien-type) (alien-type-bits alien-type))
     (if (local-alien-info-force-to-memory-p info)
-        #!+(or x86 x86-64)
+        #+(or x86 x86-64)
         `(%primitive alloc-alien-stack-space
                      ,(ceiling (alien-type-bits alien-type)
-                               sb!vm:n-byte-bits))
-        #!-(or x86 x86-64)
+                               sb-vm:n-byte-bits))
+        #-(or x86 x86-64)
         `(%primitive alloc-number-stack-space
                      ,(ceiling (alien-type-bits alien-type)
-                               sb!vm:n-byte-bits))
+                               sb-vm:n-byte-bits))
         (let* ((alien-rep-type-spec (compute-alien-rep-type alien-type))
                (alien-rep-type (specifier-type alien-rep-type-spec)))
           (cond ((csubtypep (specifier-type 'system-area-pointer)
                             alien-rep-type)
                  '(int-sap 0))
                 ((ctypep 0 alien-rep-type) 0)
-                ((ctypep 0.0f0 alien-rep-type) 0.0f0)
-                ((ctypep 0.0d0 alien-rep-type) 0.0d0)
+                ((ctypep $0.0f0 alien-rep-type) $0.0f0)
+                ((ctypep $0.0d0 alien-rep-type) $0.0d0)
                 (t
                  (compiler-error
                   "Aliens of type ~S cannot be represented immediately."
                   (unparse-alien-type alien-type))))))))
 
-(deftransform note-local-alien-type ((info var) * * :important t)
-  ;; FIXME: This test and error occur about a zillion times. They
-  ;; could be factored into a function.
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+(deftransform note-local-alien-type ((info var))
+  (alien-info-constant-or-abort info)
   (let ((info (lvar-value info)))
-    (/noshow "in DEFTRANSFORM NOTE-LOCAL-ALIEN-TYPE" info)
-    (/noshow (local-alien-info-force-to-memory-p info))
     (unless (local-alien-info-force-to-memory-p info)
       (let ((var-node (lvar-uses var)))
-        (/noshow var-node (ref-p var-node))
-        (when (ref-p var-node)
+        (when (and (ref-p var-node)
+                   (constant-reference-p var-node))
           (propagate-to-refs (ref-leaf var-node)
                              (specifier-type
                               (compute-alien-rep-type
                                (local-alien-info-type info))))))))
   nil)
 
-(deftransform local-alien ((info var) * * :important t)
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+(deftransform local-alien ((info var))
+  (alien-info-constant-or-abort info)
   (let* ((info (lvar-value info))
          (alien-type (local-alien-info-type info)))
     (/noshow "in DEFTRANSFORM LOCAL-ALIEN" info alien-type)
     (/noshow (local-alien-info-force-to-memory-p info))
     (if (local-alien-info-force-to-memory-p info)
-        `(extract-alien-value var 0 ',alien-type)
+        `(%alien-value var 0 ',alien-type)
         `(naturalize var ',alien-type))))
 
-(deftransform %local-alien-forced-to-memory-p ((info) * * :important t)
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+(deftransform %local-alien-forced-to-memory-p ((info))
+  (alien-info-constant-or-abort info)
   (let ((info (lvar-value info)))
     (local-alien-info-force-to-memory-p info)))
 
-(deftransform %set-local-alien ((info var value) * * :important t)
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+(deftransform %set-local-alien ((info var value))
+  (alien-info-constant-or-abort info)
   (let* ((info (lvar-value info))
          (alien-type (local-alien-info-type info)))
     (if (local-alien-info-force-to-memory-p info)
-        `(deposit-alien-value var 0 ',alien-type value)
+        `(setf (%alien-value var 0 ',alien-type) value)
         '(error "This should be eliminated as dead code."))))
 
 (defoptimizer (%local-alien-addr derive-type) ((info var))
+  (declare (ignore var))
   (if (constant-lvar-p info)
       (let* ((info (lvar-value info))
              (alien-type (local-alien-info-type info)))
         (make-alien-type-type (make-alien-pointer-type :to alien-type)))
       *wild-type*))
 
-(deftransform %local-alien-addr ((info var) * * :important t)
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
+(deftransform %local-alien-addr ((info var))
+  (alien-info-constant-or-abort info)
   (let* ((info (lvar-value info))
          (alien-type (local-alien-info-type info)))
     (/noshow "in DEFTRANSFORM %LOCAL-ALIEN-ADDR, creating %SAP-ALIEN")
     (if (local-alien-info-force-to-memory-p info)
         `(%sap-alien var ',(make-alien-pointer-type :to alien-type))
         (error "This shouldn't happen."))))
-
-(deftransform dispose-local-alien ((info var) * * :important t)
-  (unless (constant-lvar-p info)
-    (abort-ir1-transform "Local alien info isn't constant?"))
-  (let* ((info (lvar-value info))
-         (alien-type (local-alien-info-type info)))
-    (if (local-alien-info-force-to-memory-p info)
-      #!+(or x86 x86-64) `(%primitive dealloc-alien-stack-space
-                          ,(ceiling (alien-type-bits alien-type)
-                                    sb!vm:n-byte-bits))
-      #!-(or x86 x86-64) `(%primitive dealloc-number-stack-space
-                          ,(ceiling (alien-type-bits alien-type)
-                                    sb!vm:n-byte-bits))
-      nil)))
 
 ;;;; %CAST
 
 (defoptimizer (%cast derive-type) ((alien type))
+  (declare (ignore alien))
   (or (when (constant-lvar-p type)
         (let ((alien-type (lvar-value type)))
           (when (alien-type-p alien-type)
             (make-alien-type-type alien-type))))
       *wild-type*))
 
-(deftransform %cast ((alien target-type) * * :important t)
+(deftransform %cast ((alien target-type))
   (unless (constant-lvar-p target-type)
     (give-up-ir1-transform
      "The alien type is not constant, so access cannot be open coded."))
@@ -470,7 +454,7 @@
 
 ;;;; ALIEN-SAP, %SAP-ALIEN, %ADDR, etc.
 
-(deftransform alien-sap ((alien) * * :important t)
+(deftransform alien-sap ((alien))
   (let ((alien-node (lvar-uses alien)))
     (typecase alien-node
       (combination
@@ -487,12 +471,10 @@
       (make-alien-type-type (lvar-value type))
       *wild-type*))
 
-(deftransform %sap-alien ((sap type) * * :important t)
+(deftransform %sap-alien ((sap type))
+  "optimize away %SAP-ALIEN"
   (give-up-ir1-transform
-   ;; FIXME: The hardcoded newline here causes more-than-usually
-   ;; screwed-up formatting of the optimization note output.
-   "could not optimize away %SAP-ALIEN: forced to do runtime ~@
-    allocation of alien-value structure"))
+   "forced to do runtime allocation of alien-value structure"))
 
 ;;;; NATURALIZE/DEPORT/EXTRACT/DEPOSIT magic
 
@@ -507,144 +489,27 @@
                result)
            (error (condition)
                   (compiler-error "~A" condition)))))
-  (deftransform naturalize ((object type) * * :important t)
+  (deftransform naturalize ((object type))
     (%computed-lambda #'compute-naturalize-lambda type))
-  (deftransform deport ((alien type) * * :important t)
+  (deftransform deport ((alien type))
     (%computed-lambda #'compute-deport-lambda type))
-  (deftransform deport-alloc ((alien type) * * :important t)
+  (deftransform deport-alloc ((alien type))
     (%computed-lambda #'compute-deport-alloc-lambda type))
-  (deftransform extract-alien-value ((sap offset type) * * :important t)
+  (deftransform %alien-value ((sap offset type))
     (%computed-lambda #'compute-extract-lambda type))
-  (deftransform deposit-alien-value ((sap offset type value) * * :important t)
+  (deftransform (setf %alien-value) ((value sap offset type))
     (%computed-lambda #'compute-deposit-lambda type)))
 
-;;;; a hack to clean up divisions
-
-(defun count-low-order-zeros (thing)
-  (typecase thing
-    (lvar
-     (if (constant-lvar-p thing)
-         (count-low-order-zeros (lvar-value thing))
-         (count-low-order-zeros (lvar-uses thing))))
-    (combination
-     (case (let ((name (lvar-fun-name (combination-fun thing))))
-             (or (modular-version-info name :untagged nil) name))
-       ((+ -)
-        (let ((min most-positive-fixnum)
-              (itype (specifier-type 'integer)))
-          (dolist (arg (combination-args thing) min)
-            (if (csubtypep (lvar-type arg) itype)
-                (setf min (min min (count-low-order-zeros arg)))
-                (return 0)))))
-       (*
-        (let ((result 0)
-              (itype (specifier-type 'integer)))
-          (dolist (arg (combination-args thing) result)
-            (if (csubtypep (lvar-type arg) itype)
-                (setf result (+ result (count-low-order-zeros arg)))
-                (return 0)))))
-       (ash
-        (let ((args (combination-args thing)))
-          (if (= (length args) 2)
-              (let ((amount (second args)))
-                (if (constant-lvar-p amount)
-                    (max (+ (count-low-order-zeros (first args))
-                            (lvar-value amount))
-                         0)
-                    0))
-              0)))
-       (t
-        0)))
-    (integer
-     (if (zerop thing)
-         most-positive-fixnum
-         (do ((result 0 (1+ result))
-              (num thing (ash num -1)))
-             ((logbitp 0 num) result))))
-    (cast
-     (count-low-order-zeros (cast-value thing)))
-    (t
-     0)))
-
-(deftransform / ((numerator denominator) (integer integer))
-  "convert x/2^k to shift"
-  (unless (constant-lvar-p denominator)
-    (give-up-ir1-transform))
-  (let* ((denominator (lvar-value denominator))
-         (bits (1- (integer-length denominator))))
-    (unless (and (> denominator 0) (= (ash 1 bits) denominator))
-      (give-up-ir1-transform))
-    (let ((alignment (count-low-order-zeros numerator)))
-      (unless (>= alignment bits)
-        (give-up-ir1-transform))
-      `(ash numerator ,(- bits)))))
-
-(deftransform ash ((value amount))
-  (let ((value-node (lvar-uses value)))
-    (unless (combination-p value-node)
-      (give-up-ir1-transform))
-    (let ((inside-fun-name (lvar-fun-name (combination-fun value-node))))
-      (multiple-value-bind (prototype width)
-          (modular-version-info inside-fun-name :untagged nil)
-        (unless (eq (or prototype inside-fun-name) 'ash)
-          (give-up-ir1-transform))
-        (when (and width (not (constant-lvar-p amount)))
-          (give-up-ir1-transform))
-        (let ((inside-args (combination-args value-node)))
-          (unless (= (length inside-args) 2)
-            (give-up-ir1-transform))
-          (let ((inside-amount (second inside-args)))
-            (unless (and (constant-lvar-p inside-amount)
-                         (not (minusp (lvar-value inside-amount))))
-              (give-up-ir1-transform)))
-          (splice-fun-args value inside-fun-name 2)
-          (if width
-              `(lambda (value amount1 amount2)
-                 (logand (ash value (+ amount1 amount2))
-                         ,(1- (ash 1 (+ width (lvar-value amount))))))
-              `(lambda (value amount1 amount2)
-                 (ash value (+ amount1 amount2)))))))))
-
 ;;;; ALIEN-FUNCALL support
 
 (deftransform alien-funcall ((function &rest args)
-                             ((alien (* t)) &rest *) *
-                             :important t)
+                             ((alien (* t)) &rest t) *)
   (let ((names (make-gensym-list (length args))))
     (/noshow "entering first DEFTRANSFORM ALIEN-FUNCALL" function args)
     `(lambda (function ,@names)
        (alien-funcall (deref function) ,@names))))
 
-;;; Frame pointer, program counter conses. In each thread it's bound
-;;; locally or not bound at all.
-(defvar *saved-fp-and-pcs*)
-
-#!+:c-stack-is-control-stack
-(declaim (inline invoke-with-saved-fp-and-pc))
-#!+:c-stack-is-control-stack
-(defun invoke-with-saved-fp-and-pc (fn)
-  (declare #-sb-xc-host (muffle-conditions compiler-note)
-           (optimize (speed 3)))
-  (let* ((fp-and-pc (cons (%caller-frame)
-                          (sap-int (%caller-pc)))))
-    (declare (truly-dynamic-extent fp-and-pc))
-    (let ((*saved-fp-and-pcs* (if (boundp '*saved-fp-and-pcs*)
-                                  (cons fp-and-pc *saved-fp-and-pcs*)
-                                  (list fp-and-pc))))
-      (declare (truly-dynamic-extent *saved-fp-and-pcs*))
-      (funcall fn))))
-
-(defun find-saved-fp-and-pc (fp)
-  (when (boundp '*saved-fp-and-pcs*)
-    (dolist (x *saved-fp-and-pcs*)
-      (when (#!+:stack-grows-downward-not-upward
-             sap>
-             #!-:stack-grows-downward-not-upward
-             sap<
-             (int-sap (get-lisp-obj-address (car x))) fp)
-        (return (values (car x) (cdr x)))))))
-
-(deftransform alien-funcall ((function &rest args) * * :node node :important t)
+(deftransform alien-funcall ((function &rest args) * * :node node)
   (let ((type (lvar-type function)))
     (unless (alien-type-type-p type)
       (give-up-ir1-transform "can't tell function type at compile time"))
@@ -652,7 +517,8 @@
     (let ((alien-type (alien-type-type-alien-type type)))
       (unless (alien-fun-type-p alien-type)
         (give-up-ir1-transform))
-      (let ((arg-types (alien-fun-type-arg-types alien-type)))
+      (let ((arg-types (alien-fun-type-arg-types alien-type))
+            (ignore-fun))
         (unless (= (length args) (length arg-types))
           (abort-ir1-transform
            "wrong number of arguments; expected ~W, got ~W"
@@ -667,9 +533,26 @@
           (let ((return-type (alien-fun-type-result-type alien-type))
                 ;; Innermost, we DEPORT the parameters (e.g. by taking SAPs
                 ;; to them) and do the call.
-                (body `(%alien-funcall (deport function ',alien-type)
-                                       ',alien-type
-                                       ,@(deports))))
+                (body
+                 ;; If FUNCTION's source looks like
+                 ;;  (%SAP-ALIEN (FOREIGN-SYMBOL-SAP "sym") #<anything>)
+                 ;; then snarf out the string and use it as the funarg
+                 ;; unless the backend lacks the CALL-OUT-NAMED vop.
+                 `(%alien-funcall
+                   ,(or (when (and (gethash 'call-out-named *backend-parsed-vops*)
+                                   (lvar-matches function :fun-names '(%sap-alien)
+                                                          :arg-count 2))
+                          (let ((sap (first (combination-args (lvar-use function)))))
+                            (when (lvar-matches sap :fun-names '(foreign-symbol-sap)
+                                                    :arg-count 1)
+                              (let ((sym (first (combination-args (lvar-use sap)))))
+                                (when (and (constant-lvar-p sym)
+                                           (stringp (lvar-value sym)))
+                                  (setq ignore-fun t)
+                                  (lvar-value sym))))))
+                        `(deport function ',alien-type))
+                   ',alien-type
+                   ,@(deports))))
             ;; Wrap that in a WITH-PINNED-OBJECTS to ensure the values
             ;; the SAPs are taken for won't be moved by the GC. (If
             ;; needed: some alien types won't need it).
@@ -701,31 +584,33 @@
             ;; Remember this frame to make sure that we can get back
             ;; to it later regardless of how the foreign stack looks
             ;; like.
-            #!+:c-stack-is-control-stack
-            (when (policy node (<= speed debug))
-              (setf body `(invoke-with-saved-fp-and-pc (lambda () ,body))))
+            #+c-stack-is-control-stack
+            (when (policy node (= 3 alien-funcall-saves-fp-and-pc))
+              (setf body `(invoke-with-saved-fp (lambda () ,body))))
             (/noshow "returning from DEFTRANSFORM ALIEN-FUNCALL" (params) body)
             `(lambda (function ,@(params))
+               ,@(when ignore-fun '((declare (ignore function))))
                (declare (optimize (let-conversion 3)))
                ,body)))))))
 
 (defoptimizer (%alien-funcall derive-type) ((function type &rest args))
   (declare (ignore function args))
-  (unless (constant-lvar-p type)
+  (unless (and (constant-lvar-p type)
+               (alien-fun-type-p (lvar-value type)))
     (error "Something is broken."))
-  (let ((type (lvar-value type)))
-    (unless (alien-fun-type-p type)
-      (error "Something is broken."))
-    (values-specifier-type
-     (compute-alien-rep-type
-      (alien-fun-type-result-type type)
-      :result))))
+  (let ((spec (compute-alien-rep-type
+               (alien-fun-type-result-type (lvar-value type))
+               :result)))
+    (if (eq spec '*) *wild-type* (values-specifier-type spec))))
 
 (defoptimizer (%alien-funcall ltn-annotate)
               ((function type &rest args) node ltn-policy)
+  (declare (ignore type ltn-policy))
   (setf (basic-combination-info node) :funny)
   (setf (node-tail-p node) nil)
-  (annotate-ordinary-lvar function)
+  (unless (and (constant-lvar-p function)
+               (stringp function))
+    (annotate-ordinary-lvar function))
   (dolist (arg args)
     (annotate-ordinary-lvar arg)))
 
@@ -739,73 +624,131 @@
                   (lvar-value type)
                   (error "Something is broken.")))
         (lvar (node-lvar call))
-        (args args)
-        #!+x86
+        ;; KLUDGE: On ARM systems our register pressure is so high
+        ;; that if we process register args before stack args we end
+        ;; up with all of our non-descriptor regs either
+        ;; component-live (NFP) or wired (everything else) and can't
+        ;; actually process the stack args.  Processing the arguments
+        ;; in reverse order here doesn't change the semantics, but we
+        ;; deal with all of the stack arguments before the wired
+        ;; register arguments become live.
+        (args #-arm args #+arm (reverse args))
+        #+c-stack-is-control-stack
         (stack-pointer (make-stack-pointer-tn)))
     (multiple-value-bind (nsp stack-frame-size arg-tns result-tns)
         (make-call-out-tns type)
-      #!+x86
-      (progn
-        (vop set-fpu-word-for-c call block)
-        (vop current-stack-pointer call block stack-pointer))
+      #+x86
+      (vop set-fpu-word-for-c call block)
+      ;; Save the stack pointer, it will get aligned and subtracting
+      ;; the size will not restore the original value, and some
+      ;; things, like SB-C::CALL-VARIABLE, use the stack pointer to
+      ;; calculate the number of saved values.
+      ;; See alien.impure.lisp/:stack-misalignment
+      #+c-stack-is-control-stack
+      (vop current-stack-pointer call block stack-pointer)
       (vop alloc-number-stack-space call block stack-frame-size nsp)
-      (dolist (tn arg-tns)
-        ;; On PPC, TN might be a list. This is used to indicate
-        ;; something special needs to happen. See below.
-        ;;
-        ;; FIXME: We should implement something better than this.
-        (let* ((first-tn (if (listp tn) (car tn) tn))
-               (arg (pop args))
-               (sc (tn-sc first-tn))
-               (scn (sc-number sc))
-               #!-(or x86 x86-64) (temp-tn (make-representation-tn
-                                            (tn-primitive-type first-tn) scn))
-               (move-arg-vops (svref (sc-move-arg-vops sc) scn)))
-          (aver arg)
-          (unless (= (length move-arg-vops) 1)
-            (error "no unique move-arg-vop for moves in SC ~S" (sc-name sc)))
-          #!+(or x86 x86-64) (emit-move-arg-template call
-                                                     block
-                                                     (first move-arg-vops)
-                                                     (lvar-tn call block arg)
-                                                     nsp
-                                                     first-tn)
-          #!-(or x86 x86-64) (progn
-                               (emit-move call
-                                          block
-                                          (lvar-tn call block arg)
-                                          temp-tn)
-                               (emit-move-arg-template call
-                                                       block
-                                                       (first move-arg-vops)
-                                                       temp-tn
-                                                       nsp
-                                                       first-tn))
-          #!+(and ppc darwin)
-          (when (listp tn)
-            ;; This means that we have a float arg that we need to
-            ;; also copy to some int regs. The list contains the TN
-            ;; for the float as well as the TNs to use for the int
-            ;; arg.
-            (destructuring-bind (float-tn i1-tn &optional i2-tn)
-                tn
-              (if i2-tn
-                  (vop sb!vm::move-double-to-int-arg call block
-                       float-tn i1-tn i2-tn)
-                  (vop sb!vm::move-single-to-int-arg call block
-                       float-tn i1-tn))))))
+      ;; KLUDGE: This is where the second half of the ARM
+      ;; register-pressure change lives (see above).
+      (dolist (tn #-arm arg-tns #+arm (reverse arg-tns))
+        (if (functionp tn)
+            (funcall tn (pop args) call block nsp)
+            ;; On PPC, TN might be a list. This is used to indicate
+            ;; something special needs to happen. See below.
+            ;;
+            ;; FIXME: We should implement something better than this.
+            (let* ((first-tn (if (listp tn) (car tn) tn))
+                   (arg (pop args))
+                   (sc (tn-sc first-tn))
+                   (scn (sc-number sc))
+                   (move-arg-vops (svref (sc-move-arg-vops sc) scn)))
+              (aver arg)
+              (unless (= (length move-arg-vops) 1)
+                (error "no unique move-arg-vop for moves in SC ~S" (sc-name sc)))
+
+              (cond
+                #+arm-softfp
+                ((and (proper-list-of-length-p tn 3)
+                      (symbolp (third tn)))
+                 (emit-template call block
+                                (template-or-lose (third tn))
+                                (reference-tn (lvar-tn call block arg) nil)
+                                (reference-tn-list (butlast tn) t)))
+                (t
+                 (when (eq (sb-kind (sc-sb sc)) :unbounded) ;; stacks are unbounded
+                   ;; Avoid allocating this TN on the caller's stack
+                   (setf (tn-kind first-tn) :arg-pass))
+                 #+(or x86 x86-64)
+                 (emit-move-arg-template call
+                                         block
+                                         (first move-arg-vops)
+                                         (lvar-tn call block arg)
+                                         nsp
+                                         first-tn)
+                 #-(or x86 x86-64)
+                 (let* ((primitive-type (tn-primitive-type first-tn))
+                        ;; If the destination is a stack TN make sure
+                        ;; the temporary TN is a register.
+                        (scn (if (sc-number-stack-p sc)
+                                 (car (primitive-type-scs primitive-type))
+                                 scn))
+                        (temp-tn (make-representation-tn primitive-type scn)))
+                   (emit-move call block (lvar-tn call block arg) temp-tn)
+                   (emit-move-arg-template call
+                                           block
+                                           (first move-arg-vops)
+                                           temp-tn
+                                           nsp
+                                           first-tn))))
+              #+(and ppc darwin)
+              (when (listp tn)
+                ;; This means that we have a float arg that we need to
+                ;; also copy to some int regs. The list contains the TN
+                ;; for the float as well as the TNs to use for the int
+                ;; arg.
+                (destructuring-bind (float-tn i1-tn &optional i2-tn)
+                    tn
+                  (if i2-tn
+                      (vop sb-vm::move-double-to-int-arg call block
+                           float-tn i1-tn i2-tn)
+                      (vop sb-vm::move-single-to-int-arg call block
+                           float-tn i1-tn)))))))
       (aver (null args))
-      (unless (listp result-tns)
-        (setf result-tns (list result-tns)))
-      (let ((arg-tns (flatten-list arg-tns)))
-        (vop* call-out call block
-              ((lvar-tn call block function)
-               (reference-tn-list arg-tns nil))
-              ((reference-tn-list result-tns t))))
-      #!-x86
-      (vop dealloc-number-stack-space call block stack-frame-size)
-      #!+x86
-      (progn
+      (let* ((result-tns (ensure-list result-tns))
+             (arg-operands
+              (reference-tn-list (remove-if-not #'tn-p (flatten-list arg-tns)) nil))
+             (result-operands
+              (reference-tn-list (remove-if-not #'tn-p result-tns) t)))
+        (cond #+#.(cl:if (sb-c::vop-existsp :named sb-vm::call-out-named) '(and) '(or))
+              ((and (constant-lvar-p function) (stringp (lvar-value function)))
+               (vop* call-out-named call block (arg-operands) (result-operands)
+                     (lvar-value function)
+                     (sb-alien::alien-fun-type-varargs type)))
+              (t
+               (vop* call-out call block
+                     ((lvar-tn call block function) arg-operands)
+                     (result-operands))))
+        #-c-stack-is-control-stack
+        (vop dealloc-number-stack-space call block stack-frame-size)
+        #+c-stack-is-control-stack
         (vop reset-stack-pointer call block stack-pointer)
-        (vop set-fpu-word-for-lisp call block))
-      (move-lvar-result call block result-tns lvar))))
+        #+x86
+        (vop set-fpu-word-for-lisp call block)
+        (cond
+          #+arm-softfp
+          ((and lvar
+                (fourth result-tns))
+           (emit-template call block
+                          (template-or-lose (fourth result-tns))
+                          (reference-tn-list (butlast result-tns 2) nil)
+                          (reference-tn (third result-tns) t))
+           (move-lvar-result call block (list (third result-tns)) lvar))
+          (t
+           (move-lvar-result call block result-tns lvar)))))))
+
+(deftransform sb-alien::c-string-external-format ((type)
+                                                  ((constant-arg sb-alien::alien-c-string-type)))
+  (let ((format (sb-alien::alien-c-string-type-external-format
+                 (lvar-value type))))
+    (if (eq format :default)
+        `(sb-alien::default-c-string-external-format)
+        `',format)))
