@@ -10,44 +10,12 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
-(defvar *args* ()
-  #!+sb-doc
-  "This variable is bound to the format arguments when an error is signalled
-  by BARF or BURP.")
-
-(defvar *ignored-errors* (make-hash-table :test 'equal))
-
-;;; A definite inconsistency has been detected. Signal an error with
-;;; *args* bound to the list of the format args.
+;;; A definite inconsistency has been detected. Signal an error.
 (declaim (ftype (function (string &rest t) (values)) barf))
-(defun barf (string &rest *args*)
-  (unless (gethash string *ignored-errors*)
-    (restart-case
-        (apply #'error string *args*)
-      (continue ()
-        :report "Ignore this error.")
-      (ignore-all ()
-        :report "Ignore this and all future occurrences of this error."
-        (setf (gethash string *ignored-errors*) t))))
-  (values))
-
-(defvar *burp-action* :warn
-  #!+sb-doc
-  "Action taken by the BURP function when a possible compiler bug is detected.
-  One of :WARN, :ERROR or :NONE.")
-(declaim (type (member :warn :error :none) *burp-action*))
-
-;;; Called when something funny but possibly correct is noticed.
-;;; Otherwise similar to BARF.
-(declaim (ftype (function (string &rest t) (values)) burp))
-(defun burp (string &rest *args*)
-  (ecase *burp-action*
-    (:warn (apply #'warn string *args*))
-    (:error (apply #'cerror "press on anyway." string *args*))
-    (:none))
-  (values))
+(defun barf (string &rest args)
+  (apply #'bug string args))
 
 ;;; *SEEN-BLOCKS* is a hashtable with true values for all blocks which
 ;;; appear in the DFO for one of the specified components.
@@ -56,8 +24,8 @@
 ;;; reached by recursing on top level functions.
 ;;; FIXME: Is it really only LAMBDAs, not e.g. FUNCTIONALs? Then
 ;;; shouldn't it be *SEEN-LAMBDAS*?
-(defvar *seen-blocks* (make-hash-table :test 'eq))
-(defvar *seen-funs* (make-hash-table :test 'eq))
+(defvar *seen-blocks*)
+(defvar *seen-funs*)
 
 ;;; Barf if NODE is in a block which wasn't reached during the graph
 ;;; walk.
@@ -68,9 +36,9 @@
   (values))
 
 ;;; Check everything that we can think of for consistency. When a
-;;; definite inconsistency is detected, we BARF. Possible problems
-;;; just cause us to BURP. Our argument is a list of components, but
-;;; we also look at the *FREE-VARS*, *FREE-FUNS* and *CONSTANTS*.
+;;; definite inconsistency is detected, we BARF.
+;;; Our argument is a list of components, but
+;;; we also look at the *IR1-NAMESPACE*.
 ;;;
 ;;; First we do a pre-pass which finds all the CBLOCKs and CLAMBDAs,
 ;;; testing that they are linked together properly and entering them
@@ -78,79 +46,81 @@
 ;;; the actual code and control flow. Finally, we scan the global leaf
 ;;; hashtables, looking for lossage.
 (declaim (ftype (function (list) (values)) check-ir1-consistency))
-(defun check-ir1-consistency (components)
-  (clrhash *seen-blocks*)
-  (clrhash *seen-funs*)
-  (dolist (c components)
-    (let* ((head (component-head c))
-           (tail (component-tail c)))
-      (unless (and (null (block-pred head))
-                   (null (block-succ tail)))
-        (barf "~S is malformed." c))
+(defun check-ir1-consistency (components &aux (ns *ir1-namespace*))
+  (let ((*seen-blocks* (make-hash-table :test 'eq))
+        (*seen-funs* (make-hash-table :test 'eq)))
+    (unwind-protect
+         (progn
+           (dolist (c components)
+             (let* ((head (component-head c))
+                    (tail (component-tail c)))
+               (unless (and (null (block-pred head))
+                            (null (block-succ tail)))
+                 (barf "~S is malformed." c))
 
-      (do ((prev nil block)
-           (block head (block-next block)))
-          ((null block)
-           (unless (eq prev tail)
-             (barf "wrong TAIL for DFO, ~S in ~S" prev c)))
-        (setf (gethash block *seen-blocks*) t)
-        (unless (eq (block-prev block) prev)
-          (barf "bad PREV for ~S, should be ~S" block prev))
-        (unless (or (eq block tail)
-                    (eq (block-component block) c))
-          (barf "~S is not in ~S." block c)))
-#|
-      (when (or (loop-blocks c) (loop-inferiors c))
-        (do-blocks (block c :both)
-          (setf (block-flag block) nil))
-        (check-loop-consistency c nil)
-        (do-blocks (block c :both)
-          (unless (block-flag block)
-            (barf "~S was not in any loop." block))))
-|#
-    ))
+               (do ((prev nil block)
+                    (block head (block-next block)))
+                   ((null block)
+                    (unless (eq prev tail)
+                      (barf "wrong TAIL for DFO, ~S in ~S" prev c)))
+                 (setf (gethash block *seen-blocks*) t)
+                 (unless (eq (block-prev block) prev)
+                   (barf "bad PREV for ~S, should be ~S" block prev))
+                 (unless (or (eq block tail)
+                             (eq (block-component block) c))
+                   (barf "~S is not in ~S." block c)))
+               (let ((component-outer-loop (component-outer-loop c)))
+                 (when (or (loop-blocks component-outer-loop)
+                           (loop-inferiors component-outer-loop))
+                   (do-blocks (block c :both)
+                     (setf (block-flag block) nil))
+                   (check-loop-consistency component-outer-loop nil)
+                   (do-blocks (block c :both)
+                     (unless (block-flag block)
+                       (barf "~S was not in any loop." block)))))))
+           (check-fun-consistency components)
 
-  (check-fun-consistency components)
+           (dolist (c components)
+             (do ((block (block-next (component-head c)) (block-next block)))
+                 ((null (block-next block)))
+               (check-block-consistency block)))
 
-  (dolist (c components)
-    (do ((block (block-next (component-head c)) (block-next block)))
-        ((null (block-next block)))
-      (check-block-consistency block)))
+           (maphash (lambda (k v)
+                      (declare (ignore k))
+                      (unless (or (eq v :deprecated)
+                                  (constant-p v)
+                                  (and (global-var-p v)
+                                       (member (global-var-kind v)
+                                               '(:global :special :unknown))))
+                        (barf "strange FREE-VARS entry: ~S" v))
+                      (when (leaf-p v)
+                        (dolist (n (leaf-refs v))
+                          (check-node-reached n)))
+                      (when (basic-var-p v)
+                        (dolist (n (basic-var-sets v))
+                          (check-node-reached n))))
+                    (free-vars ns))
 
-  (maphash (lambda (k v)
-             (declare (ignore k))
-             (unless (or (constant-p v)
-                         (and (global-var-p v)
-                              (member (global-var-kind v)
-                                      '(:global :special :unknown))))
-               (barf "strange *FREE-VARS* entry: ~S" v))
-             (dolist (n (leaf-refs v))
-               (check-node-reached n))
-             (when (basic-var-p v)
-               (dolist (n (basic-var-sets v))
-                 (check-node-reached n))))
-           *free-vars*)
+           (maphash (lambda (k v)
+                      (declare (ignore k))
+                      (unless (constant-p v)
+                        (barf "strange CONSTANTS entry: ~S" v))
+                      (dolist (n (leaf-refs v))
+                        (check-node-reached n)))
+                    (eql-constants ns))
 
-  (maphash (lambda (k v)
-             (declare (ignore k))
-             (unless (constant-p v)
-               (barf "strange *CONSTANTS* entry: ~S" v))
-             (dolist (n (leaf-refs v))
-               (check-node-reached n)))
-           *constants*)
-
-  (maphash (lambda (k v)
-             (declare (ignore k))
-             (unless (or (functional-p v)
-                         (and (global-var-p v)
-                              (eq (global-var-kind v) :global-function)))
-               (barf "strange *FREE-FUNS* entry: ~S" v))
-             (dolist (n (leaf-refs v))
-               (check-node-reached n)))
-           *free-funs*)
-  (clrhash *seen-funs*)
-  (clrhash *seen-blocks*)
-  (values))
+           (maphash (lambda (k v)
+                      (declare (ignore k))
+                      (unless (or (functional-p v)
+                                  (and (global-var-p v)
+                                       (eq (global-var-kind v) :global-function)))
+                        (barf "strange FREE-FUNS entry: ~S" v))
+                      (dolist (n (leaf-refs v))
+                        (check-node-reached n)))
+                    (free-funs ns)))
+      (clrhash *seen-blocks*)
+      (clrhash *seen-funs*))
+    (values)))
 
 ;;;; function consistency checking
 
@@ -273,12 +243,11 @@
 
 ;;;; loop consistency checking
 
-#|
 ;;; Descend through the loop nesting and check that the tree is well-formed
 ;;; and that all blocks in the loops are known blocks. We also mark each block
 ;;; that we see so that we can do a check later to detect blocks that weren't
 ;;; in any loop.
-(declaim (ftype (function (loop (or loop null)) (values)) check-loop-consistency))
+(declaim (ftype (function (cloop (or cloop null)) (values)) check-loop-consistency))
 (defun check-loop-consistency (loop superior)
   (unless (eq (loop-superior loop) superior)
     (barf "wrong superior in ~S, should be ~S" loop superior))
@@ -307,7 +276,7 @@
   (values))
 
 ;;; Check that Block is either in Loop or an inferior.
-(declaim (ftype (function (block loop) (values)) check-loop-block))
+(declaim (ftype (function (cblock cloop) (values)) check-loop-block))
 (defun check-loop-block (block loop)
   (unless (gethash block *seen-blocks*)
     (barf "unseen block ~S in loop info for ~S" block loop))
@@ -320,8 +289,6 @@
       (barf "~S is in loop info for ~S but not in the loop." block loop)))
   (values))
 
-|#
-
 ;;; Check a block for consistency at the general flow-graph level, and
 ;;; call CHECK-NODE-CONSISTENCY on each node to locally check for
 ;;; semantic consistency.
@@ -463,8 +430,7 @@
      (let ((leaf (ref-leaf node)))
        (when (functional-p leaf)
          (if (eq (functional-kind leaf) :toplevel-xep)
-             (unless (eq (component-kind (block-component (node-block node)))
-                         :toplevel)
+             (unless (component-toplevelish-p (block-component (node-block node)))
                (barf ":TOPLEVEL-XEP ref in non-top-level component: ~S"
                      node))
              (check-fun-reached leaf node)))))
@@ -532,6 +498,7 @@
      (dolist (exit (entry-exits node))
        (unless (node-deleted exit)
          (check-node-reached node))))
+    (enclose)
     (exit
      (let ((entry (exit-entry node))
            (value (exit-value node)))
@@ -572,10 +539,11 @@
         (barf "The WRITE-P in ~S isn't ~S." vop write-p))
       (unless (find-in #'tn-ref-next-ref ref vop-refs)
         (barf "~S not found in REFS for ~S" ref vop))
-      (unless (find-in #'tn-ref-next ref
-                       (if (tn-ref-write-p ref)
-                           (tn-writes (tn-ref-tn ref))
-                           (tn-reads (tn-ref-tn ref))))
+      (unless (or (eq (tn-kind (tn-ref-tn ref)) :unused)
+                  (find-in #'tn-ref-next ref
+                           (if (tn-ref-write-p ref)
+                               (tn-writes (tn-ref-tn ref))
+                               (tn-reads (tn-ref-tn ref)))))
         (barf "~S not found in reads/writes for its TN" ref))
 
       (let ((target (tn-ref-target ref)))
@@ -789,14 +757,8 @@
 (defun check-block-conflicts (component)
   (do-ir2-blocks (block component)
     (do ((conf (ir2-block-global-tns block)
-               (global-conflicts-next-blockwise conf))
-         (prev nil conf))
+               (global-conflicts-next-blockwise conf)))
         ((null conf))
-      (when prev
-        (unless (> (tn-number (global-conflicts-tn conf))
-                   (tn-number (global-conflicts-tn prev)))
-          (barf "~S and ~S out of order in ~S" prev conf block)))
-
       (unless (find-in #'global-conflicts-next-tnwise
                        conf
                        (tn-global-conflicts
@@ -854,12 +816,14 @@
            (do ((scs scs (cdr scs))
                 (op ops (tn-ref-across op)))
                ((null scs))
-             (let ((load-tn (tn-ref-load-tn op)))
-               (unless (eq (svref (car scs)
-                                  (sc-number
-                                   (tn-sc
-                                    (or load-tn (tn-ref-tn op)))))
-                           t)
+             (let ((tn (tn-ref-tn op))
+                   (load-tn (tn-ref-load-tn op)))
+               (unless (or (eq (tn-kind tn) :unused)
+                           (eq (svref (car scs)
+                                      (sc-number
+                                       (tn-sc
+                                        (or load-tn tn))))
+                               t))
                  (barf "operand restriction not satisfied: ~S" op))))))
     (do-ir2-blocks (block component)
       (do ((vop (ir2-block-last-vop block) (vop-prev vop)))
@@ -871,36 +835,77 @@
 
 ;;;; data structure dumping routines
 
+;;; IR2 blocks don't get a convenient number until LIFETIME-ANALYZE which is
+;;; too late when trying to talk about an ir2 block during IR2-OPTIMIZE.
+;;; So instead of accessing them by the block-number, use an opaque ID.
+;;; And avoid confusion with all the other opaque IDs flying around
+;;; by using alphabetics instead.
+;;; Generate this numbering sequence for up to 5 alphabetic "digits":
+;;; A..Z, AA,AB ... AZ, BA ... BZ, ZA .. ZZ, AAA, AAB ...
+(defun id->string (id)
+  (declare (type (integer 1 *) id))
+  (flet ((n-digit-string (n-digits val)
+           (let ((str (make-string n-digits :element-type 'base-char)))
+             (loop for i downfrom (1- n-digits) to 0
+                   do (multiple-value-bind (quo rem) (floor val 26)
+                        (setf (aref str i) (code-char (+ (char-code #\A) rem))
+                              val quo)))
+             str)))
+    (do ((min 0) (n-digits 1 (1+ n-digits)))
+        (nil)
+      (let ((span (expt 26 n-digits)))
+        (when (<= id (+ min span))
+          (return (n-digit-string n-digits (- id (1+ min)))))
+        (incf min span)))))
+;;; Invert the above mapping
+(defun string->id (string)
+  (let ((val 0))
+    (dotimes (i (length string))
+      (setq val (+ (* val 26) (- (char-code (char string i)) (char-code #\A)))))
+    (dotimes (i (1- (length string)))
+      (incf val (expt 26 (1+ i))))
+    (1+ val)))
+
+(eval-when (#+sb-xc :compile-toplevel)
+  (loop for i from 1 to 1000 ; sanity check the stringifying algorithm
+        do (assert (= (string->id (id->string i)) i))))
+
 ;;; When we print CONTINUATIONs and TNs, we assign them small numeric
 ;;; IDs so that we can get a handle on anonymous objects given a
 ;;; printout.
-;;;
-;;; FIXME:
-;;;   * Perhaps this machinery should be #!+SB-SHOW.
-;;;   * Probably the hash tables should either be weak hash tables,
-;;;     or only allocated within a single compilation unit. Otherwise
-;;;     there will be a tendency for them to grow without bound and
-;;;     keep garbage from being collected.
-(macrolet ((def (counter vto vfrom fto ffrom)
-             `(progn
-                (declaim (type hash-table ,vto ,vfrom))
-                (defvar ,vto (make-hash-table :test 'eq))
-                (defvar ,vfrom (make-hash-table :test 'eql))
-                (declaim (type fixnum ,counter))
-                (defvar ,counter 0)
-
-                (defun ,fto (x)
-                  (or (gethash x ,vto)
-                      (let ((num (incf ,counter)))
-                        (setf (gethash num ,vfrom) x)
-                        (setf (gethash x ,vto) num))))
-
-                (defun ,ffrom (num)
-                  (values (gethash num ,vfrom))))))
-  (def *continuation-number* *continuation-numbers* *number-continuations*
-       cont-num num-cont)
-  (def *tn-id* *tn-ids* *id-tns* tn-id id-tn)
-  (def *label-id* *id-labels* *label-ids* label-id id-label))
+(macrolet ((def (array-getter fto ffrom type) ; "to" = to number/id, resp. "from"
+             (let ((xform (if (eq array-getter 'objmap-id-to-ir2block)
+                              'id->string
+                              'identity)))
+               `(progn
+                  (export '(,fto ,ffrom))
+                  (defun ,fto (x)
+                    (declare (type ,type x))
+                    (let* ((map *compilation*)
+                           (ht (objmap-obj-to-id map)))
+                      (or (gethash x ht)
+                          (let ((array (,array-getter map)))
+                            (when (null array)
+                              (setf array (make-array 20 :fill-pointer 0
+                                                      :adjustable t)
+                                    (,array-getter map) array))
+                            (setf (gethash x ht)
+                                  (,xform (1+ (vector-push-extend x array))))))))
+                  (defun ,ffrom (num)
+                    (let ((array (,array-getter *compilation*)))
+                      ,@(when (eq array-getter 'objmap-id-to-ir2block)
+                          '((setq num (string->id num))))
+                      ;; numbers start from 1, not 0
+                      (and (< (1- num) (length array)) (aref array (1- num)))))))))
+  (def objmap-id-to-node node-id id-node node)
+  (def objmap-id-to-comp component-id id-component component)
+  (def objmap-id-to-leaf leaf-id id-leaf leaf)
+  (def objmap-id-to-cont %cont-num num-cont (or ctran lvar))
+  (def objmap-id-to-ir2block ir2-block-id id-ir2-block ir2-block)
+  (def objmap-id-to-tn tn-id id-tn tn)
+  (def objmap-id-to-label label-id id-label label))
+;; Do not give NIL a continuation number
+(defun cont-num (x) (if x (%cont-num x)))
 
 ;;; Print a terse one-line description of LEAF.
 (defun print-leaf (leaf &optional (stream *standard-output*))
@@ -911,7 +916,13 @@
     (global-var
      (format stream "~S {~A}" (leaf-debug-name leaf) (global-var-kind leaf)))
     (functional
-     (format stream "~S ~S" (type-of leaf) (functional-debug-name leaf)))))
+     (format stream "~S ~S ~S" (type-of leaf) (functional-debug-name leaf)
+             (mapcar #'leaf-debug-name
+                     (typecase leaf
+                       (clambda
+                        (lambda-vars leaf))
+                       (optional-dispatch
+                        (optional-dispatch-arglist leaf))))))))
 
 ;;; Attempt to find a block given some thing that has to do with it.
 (declaim (ftype (sfunction (t) cblock) block-or-lose))
@@ -928,7 +939,7 @@
     (integer (ctran-block (num-cont thing)))
     (functional (lambda-block (main-entry thing)))
     (null (error "Bad thing: ~S." thing))
-    (symbol (block-or-lose (gethash thing *free-funs*)))))
+    (symbol (block-or-lose (gethash thing (free-funs *ir1-namespace*))))))
 
 ;;; Print cN.
 (defun print-ctran (cont)
@@ -945,17 +956,21 @@
         do (format stream "~:[u~;d~]v~D~@[ ~]"
                    (lvar-dynamic-extent lvar) (cont-num lvar) rest)))
 
+(defvar *debug-print-types* nil)
+
 ;;; Print out the nodes in BLOCK in a format oriented toward
 ;;; representing what the code does.
 (defun print-nodes (block)
   (setq block (block-or-lose block))
   (pprint-logical-block (nil nil)
-    (format t "~:@_IR1 block ~D start c~D"
+    (format t "~:@_IR1 block ~D~@[ start c~D~]"
             (block-number block) (cont-num (block-start block)))
     (when (block-delete-p block)
       (format t " <deleted>"))
 
     (pprint-newline :mandatory)
+    (when (block-start-cleanup block)
+      (format t "cleanup ~s~%" (cleanup-kind (block-start-cleanup block))))
     (awhen (block-info block)
       (format t "start stack: ")
       (print-lvar-stack (ir2-block-start-stack it))
@@ -979,7 +994,8 @@
              (dolist (arg (basic-combination-args node))
                (if arg
                    (print-lvar arg)
-                   (format t "<none> ")))))
+                   (format t "<none> "))))
+           (format t "{derived ~a}" (type-specifier (basic-combination-derived-type node))))
           (cset
            (write-string "set ")
            (print-leaf (set-var node))
@@ -1026,7 +1042,18 @@
              (format t "cast v~D ~A[~S -> ~S]" (cont-num value)
                      (if (cast-%type-check node) #\+ #\-)
                      (cast-type-to-check node)
-                     (cast-asserted-type node)))))
+                     (cast-asserted-type node))))
+          (no-op
+           (princ "no-op"))
+          (enclose
+           (write-string "enclose ")
+           (dolist (leaf (enclose-funs node))
+             (print-leaf leaf)
+             (write-char #\space))))
+        (when (and *debug-print-types*
+                   (valued-node-p node))
+          (write-char #\space)
+          (princ (type-specifier (node-derived-type node))))
         (pprint-newline :mandatory)))
 
     (awhen (block-info block)
@@ -1035,7 +1062,9 @@
       (pprint-newline :mandatory))
     (let ((succ (block-succ block)))
       (format t "successors~{ c~D~}~%"
-              (mapcar (lambda (x) (cont-num (block-start x))) succ))))
+              (mapcar (lambda (x) (cont-num (block-start x))) succ)))
+    (when (block-end-cleanup block)
+      (format t "cleanup ~s~%" (cleanup-kind (block-end-cleanup block)))))
   (values))
 
 ;;; Print the guts of a TN. (logic shared between PRINT-OBJECT (TN T)
@@ -1049,7 +1078,14 @@
           (t
            (format stream "t~D" (tn-id tn))))
     (when (and (tn-sc tn) (tn-offset tn))
-      (format stream "[~A]" (location-print-name tn)))))
+      (format stream "[~A]" (location-print-name tn)))
+    (format stream " ~s~@[ ~a~]~@[ ~a~]" (tn-kind tn)
+            (and *debug-print-types*
+                 (and (tn-sc tn)
+                      (sc-name (tn-sc tn))))
+            (and *debug-print-types*
+                 (and (tn-primitive-type tn)
+                      (primitive-type-name (tn-primitive-type tn)))))))
 
 ;;; Print the TN-REFs representing some operands to a VOP, linked by
 ;;; TN-REF-ACROSS.
@@ -1060,6 +1096,7 @@
         ((null ref))
       (let ((tn (tn-ref-tn ref))
             (ltn (tn-ref-load-tn ref)))
+        (awhen (tn-ref-memory-access ref) (format t "@~A" it))
         (cond ((not ltn)
                (print-tn-guts tn))
               (t
@@ -1073,16 +1110,29 @@
 ;;; necessary.
 (defun print-vop (vop)
   (pprint-logical-block (*standard-output* nil)
-    (princ (vop-info-name (vop-info vop)))
+    (princ (vop-name vop))
     (princ #\space)
     (pprint-indent :current 0)
     (print-operands (vop-args vop))
     (pprint-newline :linear)
     (when (vop-codegen-info vop)
-      (princ (with-output-to-string (stream)
-               (let ((*print-level* 1)
-                     (*print-length* 3))
-                 (format stream "{~{~S~^ ~}} " (vop-codegen-info vop)))))
+      (case (vop-name vop)
+       (multiway-branch-if-eq
+        (princ (vop-codegen-info vop)))
+       (t
+        (princ (%with-output-to-string (stream)
+                 ;; Current print depth varies based on whether PRINT-VOP
+                 ;; is called by DESCRIBE-IR2-COMPONENT or TRACE-INSTRUCTION,
+                 ;; so any fixed value of *PRINT-LEVEL* changes its effect
+                 ;; depending on the call context. Resetting depth to 0 seems
+                 ;; like the best way to get consistent output.
+                 ;; We shouldn't bind the printer limits to NIL, because
+                 ;; hairy internal objects such as PHYSENV can be printed.
+                 ;; See also the comment above FUNCALL-WITH-DEBUG-IO-SYNTAX.
+                 (let (#-sb-xc-host (*current-level-in-print* 0)
+                       (*print-level* 2)
+                       (*print-length* 15))
+                   (format stream "{~{~S~^ ~}} " (vop-codegen-info vop)))))))
       (pprint-newline :linear))
     (when (vop-results vop)
       (princ "=> ")
@@ -1095,9 +1145,15 @@
   (pprint-logical-block (*standard-output* nil)
     (cond
       ((eq (block-info (ir2-block-block block)) block)
-       (format t "~:@_IR2 block ~D start c~D~:@_"
+       (format t "~:@_IR2 block ~A~@[(#~d)~]~@[ start c~D~]~:@_"
+               (ir2-block-id block)
                (ir2-block-number block)
                (cont-num (block-start (ir2-block-block block))))
+       (when (boundp '*2block-info*)
+         (let ((info (gethash block *2block-info*)))
+           (format t "pred=~:A succ=~:A~:@_"
+                   (mapcar #'ir2-block-id (car info))
+                   (mapcar #'ir2-block-id (cdr info)))))
        (let ((label (ir2-block-%label block)))
          (when label
            (format t "L~D:~:@_" (label-id label)))))
@@ -1132,7 +1188,7 @@
          do (pprint-logical-block (*standard-output* nil)
               (if full
                   (print-nodes block)
-                  (format t "IR1 block ~D start c~D"
+                  (format t "IR1 block ~D~@[ start c~D~]"
                           (block-number block)
                           (cont-num (block-start block))))
               (pprint-indent :block 4)
@@ -1167,7 +1223,7 @@
         (format t "~&~A...~%" condition))))
   (values))
 
-(defvar *list-conflicts-table* (make-hash-table :test 'eq))
+(defvar *list-conflicts-table*)
 
 ;;; Add all ALWAYS-LIVE TNs in BLOCK to the conflicts. TN is ignored
 ;;; when it appears in the global conflicts.
@@ -1198,7 +1254,6 @@
                (when k
                  (res k)))
              *list-conflicts-table*)
-    (clrhash *list-conflicts-table*)
     (res)))
 
 ;;; Return a list of a the TNs that conflict with TN. Sort of, kind
@@ -1207,24 +1262,26 @@
   (aver (member (tn-kind tn) '(:normal :environment :debug-environment)))
   (let ((confs (tn-global-conflicts tn)))
     (cond (confs
-           (clrhash *list-conflicts-table*)
-           (do ((conf confs (global-conflicts-next-tnwise conf)))
-               ((null conf))
-             (format t "~&#<block ~D kind ~S>~%"
-                     (block-number (ir2-block-block (global-conflicts-block
-                                                     conf)))
-                     (global-conflicts-kind conf))
-             (let ((block (global-conflicts-block conf)))
-               (add-always-live-tns block tn)
-               (if (eq (global-conflicts-kind conf) :live)
-                   (add-all-local-tns block)
-                   (let ((bconf (global-conflicts-conflicts conf))
-                         (ltns (ir2-block-local-tns block)))
-                     (dotimes (i (ir2-block-local-tn-count block))
-                       (when (/= (sbit bconf i) 0)
-                         (setf (gethash (svref ltns i) *list-conflicts-table*)
-                               t)))))))
-           (listify-conflicts-table))
+           (let ((*list-conflicts-table* (make-hash-table :test 'eq)))
+             (unwind-protect
+                  (do ((conf confs (global-conflicts-next-tnwise conf)))
+                      ((null conf)
+                       (listify-conflicts-table))
+                    (format t "~&#<block ~D kind ~S>~%"
+                            (block-number (ir2-block-block (global-conflicts-block
+                                                            conf)))
+                            (global-conflicts-kind conf))
+                    (let ((block (global-conflicts-block conf)))
+                      (add-always-live-tns block tn)
+                      (if (eq (global-conflicts-kind conf) :live)
+                          (add-all-local-tns block)
+                          (let ((bconf (global-conflicts-conflicts conf))
+                                (ltns (ir2-block-local-tns block)))
+                            (dotimes (i (ir2-block-local-tn-count block))
+                              (when (/= (sbit bconf i) 0)
+                                (setf (gethash (svref ltns i) *list-conflicts-table*)
+                                      t)))))))
+               (clrhash *list-conflicts-table*))))
           (t
            (let* ((block (tn-local tn))
                   (ltns (ir2-block-local-tns block))
@@ -1245,9 +1302,37 @@
                (res)))))))
 
 (defun nth-vop (thing n)
-  #!+sb-doc
   "Return the Nth VOP in the IR2-BLOCK pointed to by THING."
   (let ((block (block-info (block-or-lose thing))))
     (do ((i 0 (1+ i))
          (vop (ir2-block-start-vop block) (vop-next vop)))
         ((= i n) vop))))
+
+(defun ir1-to-dot (component output-file)
+  (with-open-file (stream output-file :if-exists :supersede
+                                      :if-does-not-exist :create
+                                      :direction :output)
+    (write-line "digraph G {" stream)
+    (do-blocks (block component)
+      (when (typep (block-out block) '(cons (eql :graph) cons))
+        (format stream "~a[style=filled color=~a];~%"
+                (block-number block)
+                (case (cadr (block-out block))
+                  (:marked "green")
+                  (:remarked "aquamarine")
+                  (:start "lightblue")
+                  (:end "red"))))
+      (let ((succ (block-pred block)))
+        (when succ
+          (loop for succ in succ
+                for attr = "[style=bold]" then ""
+                do
+                (format stream "~a -> ~a~a;~%"
+                        (block-number block)
+                        (block-number succ)
+                        attr)))
+        (when (nle-block-p block)
+          (format stream "~a -> ~a [style=dotted];~%"
+                  (block-number block)
+                  (block-number (nle-block-entry-block block))))))
+    (write-line "}" stream)))
