@@ -10,7 +10,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; Data object ref/set stuff.
 
@@ -30,7 +30,6 @@
   (:results)
   (:generator 1
     (storew value object offset lowtag)))
-
 
 ;;;; Symbol hacking VOPs:
 
@@ -54,11 +53,11 @@
 ;;; Symbol-Value of NIL is NIL.
 ;;;
 (define-vop (symbol-value checked-cell-ref)
-  (:translate symbol-value)
+  (:translate symeval)
   (:generator 9
     (move obj-temp object)
     (loadw value obj-temp symbol-value-slot other-pointer-lowtag)
-    (let ((err-lab (generate-error-code vop unbound-symbol-error obj-temp)))
+    (let ((err-lab (generate-error-code vop 'unbound-symbol-error obj-temp)))
       (inst xor temp value unbound-marker-widetag)
       (inst beq temp err-lab)
       (inst nop))))
@@ -85,7 +84,7 @@
 (define-vop (fast-symbol-value cell-ref)
   (:variant symbol-value-slot other-pointer-lowtag)
   (:policy :fast)
-  (:translate symbol-value))
+  (:translate symeval))
 
 (define-vop (symbol-hash)
   (:policy :fast-safe)
@@ -106,9 +105,9 @@
 ;;; On unithreaded builds these are just copies of the non-global versions.
 (define-vop (%set-symbol-global-value set))
 (define-vop (symbol-global-value symbol-value)
-  (:translate symbol-global-value))
+  (:translate sym-global-val))
 (define-vop (fast-symbol-global-value fast-symbol-value)
-  (:translate symbol-global-value))
+  (:translate sym-global-val))
 
 ;;;; Fdefinition (fdefn) objects.
 
@@ -116,6 +115,8 @@
   (:variant fdefn-fun-slot other-pointer-lowtag))
 
 (define-vop (safe-fdefn-fun)
+  (:translate safe-fdefn-fun)
+  (:policy :fast-safe)
   (:args (object :scs (descriptor-reg) :target obj-temp))
   (:results (value :scs (descriptor-reg any-reg)))
   (:vop-var vop)
@@ -124,7 +125,7 @@
   (:generator 10
     (move obj-temp object)
     (loadw value obj-temp fdefn-fun-slot other-pointer-lowtag)
-    (let ((err-lab (generate-error-code vop undefined-fun-error obj-temp)))
+    (let ((err-lab (generate-error-code vop 'undefined-fun-error obj-temp)))
       (inst beq value null-tn err-lab))
     (inst nop)))
 
@@ -137,32 +138,37 @@
   (:temporary (:scs (non-descriptor-reg)) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
-    (let ((normal-fn (gen-label)))
       (load-type type function (- fun-pointer-lowtag))
       (inst nop)
-      (inst xor type simple-fun-header-widetag)
+      (inst xor type simple-fun-widetag)
       (inst beq type normal-fn)
       (inst addu lip function
-            (- (ash simple-fun-code-offset word-shift)
+            (- (ash simple-fun-insts-offset word-shift)
                fun-pointer-lowtag))
-      (inst li lip (make-fixup "closure_tramp" :foreign))
-      (emit-label normal-fn)
+      ;; XXX: it is questionable to load LIP with a value that is not an
+      ;; interior pointer relative to a boxed register. If there were any
+      ;; tagged pointer below LIP, then LIP might get "fixed" (i.e. ruined)
+      ;; by a GC if interrupted after this instruction. And it's even more
+      ;; questionable because 'li' is in fact two instructions.
+      ;; However, there are mitigating factors:
+      ;;   - there should be no boxed objects at a lower address.
+      ;;   - lack of thread support makes a GC interrupt somewhat unlikely,
+      ;;     (though any interrupt could call into lisp, also unlikely)
+      (inst li lip (make-fixup 'closure-tramp :assembly-routine))
+    NORMAL-FN
       (storew lip fdefn fdefn-raw-addr-slot other-pointer-lowtag)
       (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
-      (move result function))))
+      (move result function)))
 
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
+  (:args (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
     (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
-    (inst li temp (make-fixup "undefined_tramp" :foreign))
-    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (move result fdefn)))
-
+    (inst li temp (make-fixup 'undefined-tramp :assembly-routine))
+    (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 
 ;;;; Binding and Unbinding.
@@ -170,8 +176,9 @@
 ;;; BIND -- Establish VAL as a binding for SYMBOL.  Save the old value and
 ;;; the symbol on the binding stack and stuff the new value into the
 ;;; symbol.
+;;; See the "Chapter 9: Specials" of the SBCL Internals Manual.
 
-(define-vop (bind)
+(define-vop (dynbind)
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
   (:temporary (:scs (descriptor-reg)) temp)
@@ -237,11 +244,19 @@
   funcallable-instance-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %funcallable-instance-info)
 
-(define-vop (closure-ref slot-ref)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-ref)
+  (:args (object :scs (descriptor-reg)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (loadw value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
-(define-vop (closure-init slot-set)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-init)
+  (:args (object :scs (descriptor-reg))
+         (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
@@ -261,16 +276,15 @@
 
 ;;;; Instance hackery:
 
-(define-vop (instance-length)
+(define-vop ()
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
   (:results (res :scs (unsigned-reg)))
   (:result-types positive-fixnum)
   (:generator 4
-    (loadw temp struct 0 instance-pointer-lowtag)
-    (inst srl res temp n-widetag-bits)))
+    (loadw res struct 0 instance-pointer-lowtag)
+    (inst srl res res instance-length-shift)))
 
 (define-full-reffer instance-index-ref * instance-slots-offset
   instance-pointer-lowtag (descriptor-reg any-reg) * %instance-ref)
@@ -292,167 +306,116 @@
 
 ;;;; raw instance slot accessors
 
-(define-vop (raw-instance-ref/word)
-  (:translate %raw-instance-ref/word)
-  (:policy :fast-safe)
-  (:args (object :scs (descriptor-reg))
-         (index :scs (any-reg)))
-  (:arg-types * positive-fixnum)
-  (:results (value :scs (unsigned-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
-  (:temporary (:scs (interior-reg)) lip)
-  (:result-types unsigned-num)
-  (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset n-word-bytes)
-    (inst addu lip offset object)
-    (inst lw value lip (- (* instance-slots-offset n-word-bytes)
-                          instance-pointer-lowtag))))
+(macrolet ((def (suffix sc primtype)
+             `(progn
+                (define-vop ()
+                  (:translate ,(symbolicate "%RAW-INSTANCE-REF/" suffix))
+                  (:policy :fast-safe)
+                  (:args (object :scs (descriptor-reg))
+                         (index :scs (any-reg)))
+                  (:arg-types * positive-fixnum)
+                  (:results (value :scs (,sc)))
+                  (:temporary (:scs (interior-reg)) lip)
+                  (:result-types ,primtype)
+                  (:generator 5
+                    (inst addu lip object index)
+                    (loadw value lip instance-slots-offset instance-pointer-lowtag)))
 
-(define-vop (raw-instance-set/word)
-  (:translate %raw-instance-set/word)
-  (:policy :fast-safe)
-  (:args (object :scs (descriptor-reg))
-         (index :scs (any-reg))
-         (value :scs (unsigned-reg) :target result))
-  (:arg-types * positive-fixnum unsigned-num)
-  (:results (result :scs (unsigned-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
-  (:temporary (:scs (interior-reg)) lip)
-  (:result-types unsigned-num)
-  (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset n-word-bytes)
-    (inst addu lip offset object)
-    (inst sw value lip (- (* instance-slots-offset n-word-bytes)
-                          instance-pointer-lowtag))
-    (move result value)))
+                (define-vop ()
+                  (:translate ,(symbolicate "%RAW-INSTANCE-SET/" suffix))
+                  (:policy :fast-safe)
+                  (:args (object :scs (descriptor-reg))
+                         (index :scs (any-reg))
+                         (value :scs (,sc)))
+                  (:arg-types * positive-fixnum ,primtype)
+                  (:temporary (:scs (interior-reg)) lip)
+                  (:generator 5
+                    (inst addu lip object index)
+                    (storew value lip instance-slots-offset instance-pointer-lowtag))))))
+  (def word unsigned-reg unsigned-num)
+  (def signed-word signed-reg signed-num))
 
-(define-vop (raw-instance-ref/single)
+(define-vop ()
   (:translate %raw-instance-ref/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg)))
   (:arg-types * positive-fixnum)
   (:results (value :scs (single-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
   (:result-types single-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset n-word-bytes)
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (inst lwc1 value lip (- (* instance-slots-offset n-word-bytes)
                             instance-pointer-lowtag))))
 
-(define-vop (raw-instance-set/single)
+(define-vop ()
   (:translate %raw-instance-set/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (single-reg) :target result))
+         (value :scs (single-reg)))
   (:arg-types * positive-fixnum single-float)
-  (:results (result :scs (single-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types single-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset n-word-bytes)
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (inst swc1 value lip (- (* instance-slots-offset n-word-bytes)
-                            instance-pointer-lowtag))
-    (unless (location= result value)
-      (inst fmove :single result value))))
+                            instance-pointer-lowtag))))
 
-(define-vop (raw-instance-ref/double)
+(define-vop ()
   (:translate %raw-instance-ref/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg)))
   (:arg-types * positive-fixnum)
   (:results (value :scs (double-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
   (:result-types double-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 2 n-word-bytes))
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
                                instance-pointer-lowtag)))
       (ecase *backend-byte-order*
-        (:big-endian (inst lwc1 value lip immediate-offset))
-        (:little-endian (inst lwc1-odd value lip immediate-offset))))
+        (:big-endian (inst lwc1-odd value lip immediate-offset))
+        (:little-endian (inst lwc1 value lip immediate-offset))))
     (let ((immediate-offset (- (* (1+ instance-slots-offset) n-word-bytes)
                                instance-pointer-lowtag)))
       (ecase *backend-byte-order*
-        (:big-endian (inst lwc1-odd value lip immediate-offset))
-        (:little-endian (inst lwc1 value lip immediate-offset))))))
+        (:big-endian (inst lwc1 value lip immediate-offset))
+        (:little-endian (inst lwc1-odd value lip immediate-offset))))))
 
-(define-vop (raw-instance-set/double)
+(define-vop ()
   (:translate %raw-instance-set/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (double-reg) :target result))
+         (value :scs (double-reg)))
   (:arg-types * positive-fixnum double-float)
-  (:results (result :scs (double-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types double-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 2 n-word-bytes))
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
                                instance-pointer-lowtag)))
-      (ecase *backend-byte-order*
-        (:big-endian (inst swc1 value lip immediate-offset))
-        (:little-endian (inst swc1-odd value lip immediate-offset))))
-    (let ((immediate-offset (- (* (1+ instance-slots-offset) n-word-bytes)
-                               instance-pointer-lowtag)))
       (ecase *backend-byte-order*
         (:big-endian (inst swc1-odd value lip immediate-offset))
         (:little-endian (inst swc1 value lip immediate-offset))))
-    (unless (location= result value)
-      (inst fmove :double result value))))
+    (let ((immediate-offset (- (* (1+ instance-slots-offset) n-word-bytes)
+                               instance-pointer-lowtag)))
+      (ecase *backend-byte-order*
+        (:big-endian (inst swc1 value lip immediate-offset))
+        (:little-endian (inst swc1-odd value lip immediate-offset))))))
 
-(define-vop (raw-instance-ref/complex-single)
+(define-vop ()
   (:translate %raw-instance-ref/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg)))
   (:arg-types * positive-fixnum)
   (:results (value :scs (complex-single-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
   (:result-types complex-single-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 2 n-word-bytes))
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (inst lwc1
           (complex-single-reg-real-tn value)
           lip
@@ -464,42 +427,23 @@
           (- (* (1+ instance-slots-offset) n-word-bytes)
              instance-pointer-lowtag))))
 
-(define-vop (raw-instance-set/complex-single)
+(define-vop ()
   (:translate %raw-instance-set/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-single-reg) :target result))
+         (value :scs (complex-single-reg)))
   (:arg-types * positive-fixnum complex-single-float)
-  (:results (result :scs (complex-single-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types complex-single-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 2 n-word-bytes))
-    (inst addu lip offset object)
-    (let ((value-real (complex-single-reg-real-tn value))
-          (result-real (complex-single-reg-real-tn result)))
-      (inst swc1
-            value-real
-            lip
-            (- (* instance-slots-offset n-word-bytes)
-               instance-pointer-lowtag))
-      (unless (location= result-real value-real)
-        (inst fmove :single result-real value-real)))
-    (let ((value-imag (complex-single-reg-imag-tn value))
-          (result-imag (complex-single-reg-imag-tn result)))
-      (inst swc1
-            value-imag
-            lip
-            (- (* (1+ instance-slots-offset) n-word-bytes)
-               instance-pointer-lowtag))
-      (unless (location= result-imag value-imag)
-        (inst fmove :single result-imag value-imag)))))
+    (inst addu lip object index)
+    (inst swc1 (complex-single-reg-real-tn value)
+          lip (- (* instance-slots-offset n-word-bytes)
+                 instance-pointer-lowtag))
+    (inst swc1
+          (complex-single-reg-imag-tn value)
+          lip (- (* (1+ instance-slots-offset) n-word-bytes)
+                 instance-pointer-lowtag))))
 
 (define-vop (raw-instance-ref/complex-double)
   (:translate %raw-instance-ref/complex-double)
@@ -508,109 +452,102 @@
          (index :scs (any-reg)))
   (:arg-types * positive-fixnum)
   (:results (value :scs (complex-double-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
   (:result-types complex-double-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 4 n-word-bytes))
-    (inst addu lip offset object)
+    (inst addu lip object index)
     (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
                                instance-pointer-lowtag)))
       (ecase *backend-byte-order*
-        (:big-endian (inst lwc1
+        (:big-endian (inst lwc1-odd
                            (complex-double-reg-real-tn value)
                            lip
                            immediate-offset))
-        (:little-endian (inst lwc1-odd
+        (:little-endian (inst lwc1
                               (complex-double-reg-real-tn value)
                               lip
                               immediate-offset))))
     (let ((immediate-offset (- (* (1+ instance-slots-offset) n-word-bytes)
                                instance-pointer-lowtag)))
       (ecase *backend-byte-order*
-        (:big-endian (inst lwc1-odd
+        (:big-endian (inst lwc1
                            (complex-double-reg-real-tn value)
                            lip
                            immediate-offset))
-        (:little-endian (inst lwc1
+        (:little-endian (inst lwc1-odd
                               (complex-double-reg-real-tn value)
                               lip
                               immediate-offset))))
     (let ((immediate-offset (- (* (+ instance-slots-offset 2) n-word-bytes)
                                instance-pointer-lowtag)))
+      (ecase *backend-byte-order*
+        (:big-endian (inst lwc1-odd
+                           (complex-double-reg-imag-tn value)
+                           lip
+                           immediate-offset))
+        (:little-endian (inst lwc1
+                              (complex-double-reg-imag-tn value)
+                              lip
+                              immediate-offset))))
+    (let ((immediate-offset (- (* (+ instance-slots-offset 3) n-word-bytes)
+                               instance-pointer-lowtag)))
       (ecase *backend-byte-order*
         (:big-endian (inst lwc1
                            (complex-double-reg-imag-tn value)
                            lip
                            immediate-offset))
         (:little-endian (inst lwc1-odd
-                              (complex-double-reg-imag-tn value)
-                              lip
-                              immediate-offset))))
-    (let ((immediate-offset (- (* (+ instance-slots-offset 3) n-word-bytes)
-                               instance-pointer-lowtag)))
-      (ecase *backend-byte-order*
-        (:big-endian (inst lwc1-odd
-                           (complex-double-reg-imag-tn value)
-                           lip
-                           immediate-offset))
-        (:little-endian (inst lwc1
                               (complex-double-reg-imag-tn value)
                               lip
                               immediate-offset))))))
 
-(define-vop (raw-instance-set/complex-double)
+(define-vop ()
   (:translate %raw-instance-set/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (complex-double-reg) :target result))
+         (value :scs (complex-double-reg)))
   (:arg-types * positive-fixnum complex-double-float)
-  (:results (result :scs (complex-double-reg)))
-  (:temporary (:scs (non-descriptor-reg)) offset)
   (:temporary (:scs (interior-reg)) lip)
-  (:result-types complex-double-float)
   (:generator 5
-    (loadw offset object 0 instance-pointer-lowtag)
-    (inst srl offset n-widetag-bits)
-    (inst sll offset n-fixnum-tag-bits)
-    (inst subu offset index)
-    (inst subu offset (* 4 n-word-bytes))
-    (inst addu lip offset object)
-    (let ((value-real (complex-double-reg-real-tn value))
-          (result-real (complex-double-reg-real-tn result)))
+    (inst addu lip object index)
+    (let ((value-real (complex-double-reg-real-tn value)))
       (let ((immediate-offset (- (* instance-slots-offset n-word-bytes)
                                  instance-pointer-lowtag)))
         (ecase *backend-byte-order*
-          (:big-endian (inst swc1
+          (:big-endian (inst swc1-odd
                              value-real
                              lip
                              immediate-offset))
-          (:little-endian (inst swc1-odd
+          (:little-endian (inst swc1
                                 value-real
                                 lip
                                 immediate-offset))))
       (let ((immediate-offset (- (* (1+ instance-slots-offset) n-word-bytes)
                                  instance-pointer-lowtag)))
         (ecase *backend-byte-order*
-          (:big-endian (inst swc1-odd
+          (:big-endian (inst swc1
                              value-real
                              lip
                              immediate-offset))
-          (:little-endian (inst swc1
+          (:little-endian (inst swc1-odd
                                 value-real
                                 lip
-                                immediate-offset))))
-      (unless (location= result-real value-real)
-        (inst fmove :double result-real value-real)))
-    (let ((value-imag (complex-double-reg-imag-tn value))
-          (result-imag (complex-double-reg-imag-tn result)))
+                                immediate-offset)))))
+    (let ((value-imag (complex-double-reg-imag-tn value)))
       (let ((immediate-offset (- (* (+ instance-slots-offset 2) n-word-bytes)
                                  instance-pointer-lowtag)))
+        (ecase *backend-byte-order*
+          (:big-endian (inst swc1-odd
+                             value-imag
+                             lip
+                             immediate-offset))
+          (:little-endian (inst swc1
+                                value-imag
+                                lip
+                                immediate-offset))))
+      (let ((immediate-offset (- (* (+ instance-slots-offset 3) n-word-bytes)
+                                 instance-pointer-lowtag)))
         (ecase *backend-byte-order*
           (:big-endian (inst swc1
                              value-imag
@@ -619,17 +556,4 @@
           (:little-endian (inst swc1-odd
                                 value-imag
                                 lip
-                                immediate-offset))))
-      (let ((immediate-offset (- (* (+ instance-slots-offset 3) n-word-bytes)
-                                 instance-pointer-lowtag)))
-        (ecase *backend-byte-order*
-          (:big-endian (inst swc1-odd
-                             value-imag
-                             lip
-                             immediate-offset))
-          (:little-endian (inst swc1
-                                value-imag
-                                lip
-                                immediate-offset))))
-      (unless (location= result-imag value-imag)
-        (inst fmove :double result-imag value-imag)))))
+                                immediate-offset)))))))
