@@ -9,50 +9,41 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
-
-;;;; Interfaces to IR2 conversion:
-
-;;; Return a wired TN describing the N'th full call argument passing
-;;; location.
-(!def-vm-support-routine standard-arg-location (n)
-  (declare (type unsigned-byte n))
-  (if (< n register-arg-count)
-      (make-wired-tn *backend-t-primitive-type* register-arg-scn
-                     (elt *register-arg-offsets* n))
-      (make-wired-tn *backend-t-primitive-type* control-stack-arg-scn n)))
-
+(in-package "SB-VM")
+(defconstant arg-count-sc (make-sc+offset immediate-arg-scn nargs-offset))
+(defconstant closure-sc (make-sc+offset descriptor-reg-sc-number lexenv-offset))
 
 ;;; Make a passing location TN for a local call return PC.  If
 ;;; standard is true, then use the standard (full call) location,
 ;;; otherwise use any legal location.  Even in the non-standard case,
 ;;; this may be restricted by a desire to use a subroutine call
 ;;; instruction.
-(!def-vm-support-routine make-return-pc-passing-location (standard)
+(defun make-return-pc-passing-location (standard)
   (if standard
-      (make-wired-tn *backend-t-primitive-type* register-arg-scn lra-offset)
-      (make-restricted-tn *backend-t-primitive-type* register-arg-scn)))
+      (make-wired-tn *backend-t-primitive-type* descriptor-reg-sc-number lra-offset)
+      (make-restricted-tn *backend-t-primitive-type* descriptor-reg-sc-number)))
 
 ;;; This is similar to MAKE-RETURN-PC-PASSING-LOCATION, but makes a
 ;;; location to pass OLD-FP in. This is (obviously) wired in the
 ;;; standard convention, but is totally unrestricted in non-standard
 ;;; conventions, since we can always fetch it off of the stack using
 ;;; the arg pointer.
-(!def-vm-support-routine make-old-fp-passing-location (standard)
-  (if standard
-      (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset)
-      (make-normal-tn *fixnum-primitive-type*)))
+(defun make-old-fp-passing-location ()
+  (make-wired-tn *fixnum-primitive-type* immediate-arg-scn ocfp-offset))
+
+(defconstant old-fp-passing-offset
+  (make-sc+offset descriptor-reg-sc-number ocfp-offset))
 
 ;;; Make the TNs used to hold OLD-FP and RETURN-PC within the current
 ;;; function. We treat these specially so that the debugger can find
 ;;; them at a known location.
-(!def-vm-support-routine make-old-fp-save-location (env)
+(defun make-old-fp-save-location (env)
   (specify-save-tn
    (physenv-debug-live-tn (make-normal-tn *fixnum-primitive-type*) env)
    (make-wired-tn *fixnum-primitive-type*
                   control-stack-arg-scn
                   ocfp-save-offset)))
-(!def-vm-support-routine make-return-pc-save-location (env)
+(defun make-return-pc-save-location (env)
   (specify-save-tn
    (physenv-debug-live-tn (make-normal-tn *backend-t-primitive-type*) env)
    (make-wired-tn *backend-t-primitive-type*
@@ -62,38 +53,8 @@
 ;;; Make a TN for the standard argument count passing location.  We
 ;;; only need to make the standard location, since a count is never
 ;;; passed when we are using non-standard conventions.
-(!def-vm-support-routine make-arg-count-location ()
+(defun make-arg-count-location ()
   (make-wired-tn *fixnum-primitive-type* immediate-arg-scn nargs-offset))
-
-
-;;; Make a TN to hold the number-stack frame pointer.  This is
-;;; allocated once per component, and is component-live.
-(!def-vm-support-routine make-nfp-tn ()
-  (component-live-tn
-   (make-wired-tn *fixnum-primitive-type* immediate-arg-scn nfp-offset)))
-
-(!def-vm-support-routine make-stack-pointer-tn ()
-  (make-normal-tn *fixnum-primitive-type*))
-
-(!def-vm-support-routine make-number-stack-pointer-tn ()
-  (make-normal-tn *fixnum-primitive-type*))
-
-;;; Return a list of TNs that can be used to represent an unknown-values
-;;; continuation within a function.
-(!def-vm-support-routine make-unknown-values-locations ()
-  (list (make-stack-pointer-tn)
-        (make-normal-tn *fixnum-primitive-type*)))
-
-;;; This function is called by the ENTRY-ANALYZE phase, allowing
-;;; VM-dependent initialization of the IR2-COMPONENT structure.  We push
-;;; placeholder entries in the Constants to leave room for additional
-;;; noise in the code object header.
-(!def-vm-support-routine select-component-format (component)
-  (declare (type component component))
-  (dotimes (i code-constants-offset)
-    (vector-push-extend nil
-                        (ir2-component-constants (component-info component))))
-  (values))
 
 ;;;; Frame hackery:
 
@@ -140,42 +101,44 @@
     (storew value frame-pointer (tn-offset variable-home-tn))))
 
 (define-vop (xep-allocate-frame)
-  (:info start-lab copy-more-arg-follows)
-  (:ignore copy-more-arg-follows)
-  (:vop-var vop)
-  (:temporary (:scs (non-descriptor-reg)) temp)
+  (:info start-lab)
+  ;; KLUDGE: Specify an explicit offset for TEMP because NARGS is a
+  ;; non-descriptor-reg, but is also live, yet the register allocator
+  ;; does not know that it is, and if TEMP collides NARGS and
+  ;; COMPUTE-CODE-FROM-LIP needs TEMP then we run into trouble very
+  ;; quickly.
+  (:temporary (:scs (non-descriptor-reg) :offset nl6-offset) temp)
   (:generator 1
     ;; Make sure the function is aligned, and drop a label pointing to this
     ;; function header.
     (emit-alignment n-lowtag-bits)
-    (trace-table-entry trace-table-fun-prologue)
     (emit-label start-lab)
     ;; Allocate function header.
     (inst simple-fun-header-word)
-    (dotimes (i (1- simple-fun-code-offset))
-      (inst word 0))
-    (let* ((entry-point (gen-label)))
+    (inst .skip (* (1- simple-fun-insts-offset) n-word-bytes))
+    (let ((entry-point (gen-label)))
       (emit-label entry-point)
-      (inst compute-code-from-lip code-tn lip-tn entry-point temp))
       ;; FIXME alpha port has a ### note here saying we should "save it
       ;; on the stack" so that GC sees it. No idea what "it" is -dan 20020110
-    ;; Build our stack frames.
+      (inst compute-code-from-lip code-tn lip-tn entry-point temp))))
+
+(define-vop (xep-setup-sp)
+  (:vop-var vop)
+  (:generator 1
     (inst addi csp-tn cfp-tn
           (* n-word-bytes (sb-allocated-size 'control-stack)))
     (let ((nfp-tn (current-nfp-tn vop)))
       (when nfp-tn
-        (let* ((nbytes (bytes-needed-for-non-descriptor-stack-frame)))
+        (let ((nbytes (bytes-needed-for-non-descriptor-stack-frame)))
           (when (> nbytes number-stack-displacement)
             (inst stwu nsp-tn nsp-tn (- nbytes))
-            (inst addi nfp-tn nsp-tn number-stack-displacement)))))
-    (trace-table-entry trace-table-normal)))
+            (inst addi nfp-tn nsp-tn number-stack-displacement)))))))
 
 (define-vop (allocate-frame)
   (:results (res :scs (any-reg))
             (nfp :scs (any-reg)))
   (:info callee)
   (:generator 2
-    (trace-table-entry trace-table-fun-prologue)
     (move res csp-tn)
     (inst addi csp-tn csp-tn
           (* n-word-bytes (sb-allocated-size 'control-stack)))
@@ -183,8 +146,7 @@
       (let* ((nbytes (bytes-needed-for-non-descriptor-stack-frame)))
         (when (> nbytes number-stack-displacement)
           (inst stwu nsp-tn nsp-tn (- (bytes-needed-for-non-descriptor-stack-frame)))
-          (inst addi nfp nsp-tn number-stack-displacement))))
-    (trace-table-entry trace-table-normal)))
+          (inst addi nfp nsp-tn number-stack-displacement))))))
 
 ;;; Allocate a partial frame for passing stack arguments in a full call.  Nargs
 ;;; is the number of arguments passed.  If no stack arguments are passed, then
@@ -272,7 +234,7 @@ default-value-8
            (type unsigned-byte nvals) (type tn move-temp temp))
   (if (<= nvals 1)
       (progn
-        (sb!assem:without-scheduling ()
+        (sb-assem:without-scheduling ()
           (note-this-location vop :single-value-return)
           (move csp-tn ocfp-tn)
           (inst nop))
@@ -281,7 +243,7 @@ default-value-8
             (defaulting-done (gen-label))
             (default-stack-vals (gen-label)))
         ;; Branch off to the MV case.
-        (sb!assem:without-scheduling ()
+        (sb-assem:without-scheduling ()
           (note-this-location vop :unknown-return)
           (if (> nvals register-arg-count)
               (inst addic. temp nargs-tn (- (fixnumize register-arg-count)))
@@ -321,16 +283,14 @@ default-value-8
 
             (let ((defaults (defaults)))
               (when defaults
-                (assemble (*elsewhere*)
+                (assemble (:elsewhere)
                   (emit-label default-stack-vals)
-                  (trace-table-entry trace-table-fun-prologue)
                   (do ((remaining defaults (cdr remaining)))
                       ((null remaining))
                     (let ((def (car remaining)))
                       (emit-label (car def))
                       (store-stack-tn (cdr def) null-tn)))
-                  (inst b defaulting-done)
-                  (trace-table-entry trace-table-normal))))))
+                  (inst b defaulting-done))))))
 
         (inst compute-code-from-lra code-tn lra-tn lra-label temp)))
   (values))
@@ -358,7 +318,7 @@ default-value-8
   (declare (type tn args nargs start count temp))
   (let ((variable-values (gen-label))
         (done (gen-label)))
-    (sb!assem:without-scheduling ()
+    (sb-assem:without-scheduling ()
       (inst b variable-values)
       (inst nop))
 
@@ -370,8 +330,7 @@ default-value-8
 
     (emit-label done)
 
-    (assemble (*elsewhere*)
-      (trace-table-entry trace-table-fun-prologue)
+    (assemble (:elsewhere)
       (emit-label variable-values)
       (inst compute-code-from-lra code-tn lra-tn lra-label temp)
       (do ((arg *register-arg-tns* (rest arg))
@@ -380,8 +339,7 @@ default-value-8
         (storew (first arg) args i))
       (move start args)
       (move count nargs)
-      (inst b done)
-      (trace-table-entry trace-table-normal)))
+      (inst b done)))
   (values))
 
 
@@ -396,9 +354,17 @@ default-value-8
               values-start)
   (:temporary (:sc any-reg :offset nargs-offset
                :from :eval :to (:result 1))
-              nvals)
-  (:temporary (:scs (non-descriptor-reg)) temp))
+              nvals))
 
+
+;;; This hook in the codegen pass lets us insert code before fall-thru entry
+;;; points, local-call entry points, and tail-call entry points.  The default
+;;; does nothing.
+(defun emit-block-header (start-label trampoline-label fall-thru-p alignp)
+  (declare (ignore fall-thru-p alignp))
+  (when trampoline-label
+    (emit-label trampoline-label))
+  (emit-label start-label))
 
 
 ;;;; Local call with unknown values convention return:
@@ -435,7 +401,6 @@ default-value-8
   (:temporary (:sc any-reg :offset ocfp-offset :from (:eval 0)) ocfp)
   (:ignore arg-locs args ocfp)
   (:generator 5
-    (trace-table-entry trace-table-call-site)
     (let ((label (gen-label))
           (cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
@@ -453,8 +418,7 @@ default-value-8
       ;; alpha uses (maybe-load-stack-nfp-tn cur-nfp nfp-save temp)
       ;; instead of the clause below
       (when cur-nfp
-        (load-stack-tn cur-nfp nfp-save)))
-    (trace-table-entry trace-table-normal)))
+        (load-stack-tn cur-nfp nfp-save)))))
 
 
 ;;; Non-TR local call for a variable number of return values passed according
@@ -476,7 +440,6 @@ default-value-8
   (:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 20
-    (trace-table-entry trace-table-call-site)
     (let ((label (gen-label))
           (cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
@@ -494,8 +457,7 @@ default-value-8
       (note-this-location vop :unknown-return)
       (receive-unknown-values values-start nvals start count label temp)
       (when cur-nfp
-        (load-stack-tn cur-nfp nfp-save)))
-    (trace-table-entry trace-table-normal)))
+        (load-stack-tn cur-nfp nfp-save)))))
 
 
 ;;;; Local call with known values return:
@@ -519,7 +481,6 @@ default-value-8
   (:temporary (:sc control-stack :offset nfp-save-offset) nfp-save)
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:generator 5
-    (trace-table-entry trace-table-call-site)
     (let ((label (gen-label))
           (cur-nfp (current-nfp-tn vop)))
       (when cur-nfp
@@ -535,8 +496,7 @@ default-value-8
       (emit-return-pc label)
       (note-this-location vop :known-return)
       (when cur-nfp
-        (load-stack-tn cur-nfp nfp-save)))
-    (trace-table-entry trace-table-normal)))
+        (load-stack-tn cur-nfp nfp-save)))))
 
 ;;; Return from known values call.  We receive the return locations as
 ;;; arguments to terminate their lifetimes in the returning function.  We
@@ -556,7 +516,6 @@ default-value-8
   (:ignore val-locs vals)
   (:vop-var vop)
   (:generator 6
-    (trace-table-entry trace-table-fun-epilogue)
     (maybe-load-stack-tn old-fp-temp old-fp)
     (maybe-load-stack-tn return-pc-temp return-pc)
     (move csp-tn cfp-tn)
@@ -566,8 +525,7 @@ default-value-8
               (- (bytes-needed-for-non-descriptor-stack-frame)
                  number-stack-displacement))))
     (move cfp-tn old-fp-temp)
-    (inst j return-pc-temp (- n-word-bytes other-pointer-lowtag))
-    (trace-table-entry trace-table-normal)))
+    (inst j return-pc-temp (- n-word-bytes other-pointer-lowtag))))
 
 
 ;;;; Full call:
@@ -617,10 +575,12 @@ default-value-8
      (:args
       ,@(unless (eq return :tail)
           '((new-fp :scs (any-reg) :to :eval)))
-
-      ,(if named
-           '(name :target name-pass)
-           '(arg-fun :target lexenv))
+      ,@(case named
+          ((nil)
+           '((arg-fun :target lexenv)))
+          (:direct)
+          (t
+           '((name :target name-pass))))
 
       ,@(when (eq return :tail)
           '((old-fp :target old-fp-pass)
@@ -639,6 +599,7 @@ default-value-8
     (:vop-var vop)
     (:info ,@(unless (or variable (eq return :tail)) '(arg-locs))
            ,@(unless variable '(nargs))
+           ,@(when (eq named :direct) '(fun))
            ,@(when (eq return :fixed) '(nvals))
            step-instrumenting)
 
@@ -656,20 +617,24 @@ default-value-8
      (:temporary (:sc descriptor-reg
                   :offset lra-offset
                   :from (:argument ,(if (eq return :tail) 2 1))
-                  :to :eval)
+                  :to (:result 0))
                  return-pc-pass)
 
-     ,(if named
-          `(:temporary (:sc descriptor-reg :offset fdefn-offset ; -dan
-                            :from (:argument ,(if (eq return :tail) 0 1))
-                            :to :eval)
-                       name-pass)
-          `(:temporary (:sc descriptor-reg :offset lexenv-offset
-                            :from (:argument ,(if (eq return :tail) 0 1))
-                            :to :eval)
-                       lexenv))
-     (:temporary (:scs (descriptor-reg) :from (:argument 0) :to :eval)
-                 function)
+     ,@(case named
+         ((t)
+          `((:temporary (:sc descriptor-reg :offset fdefn-offset
+                         :from (:argument ,(if (eq return :tail) 0 1))
+                         :to :eval)
+                        name-pass)))
+         ((nil)
+          `((:temporary (:sc descriptor-reg :offset lexenv-offset
+                         :from (:argument ,(if (eq return :tail) 0 1))
+                         :to :eval)
+                        lexenv))))
+
+     ,@(unless (eq named :direct)
+         '((:temporary (:scs (descriptor-reg) :from (:argument 0) :to :eval)
+            function)))
      (:temporary (:sc any-reg :offset nargs-offset :to :eval)
                  nargs-pass)
 
@@ -696,7 +661,6 @@ default-value-8
                      (if (eq return :tail) 0 10)
                      15
                      (if (eq return :unknown) 25 0))
-       (trace-table-entry trace-table-call-site)
 
        (let* ((cur-nfp (current-nfp-tn vop))
               ,@(unless (eq return :tail)
@@ -767,70 +731,74 @@ default-value-8
                 (insert-step-instrumenting (callable-tn)
                   ;; Conditionally insert a conditional trap:
                   (when step-instrumenting
-                    ;; Get the symbol-value of SB!IMPL::*STEPPING*
-                    #!-sb-thread
-                    (load-symbol-value stepping sb!impl::*stepping*)
-                    #!+sb-thread
+                    ;; Get the symbol-value of SB-IMPL::*STEPPING*
+                    #-sb-thread
+                    (load-symbol-value stepping sb-impl::*stepping*)
+                    #+sb-thread
                     (loadw stepping thread-base-tn thread-stepping-slot)
-                    (inst cmpw stepping null-tn)
+                    (inst cmpwi stepping 0)
                     ;; If it's not null, trap.
                     (inst beq step-done-label)
                     ;; CONTEXT-PC will be pointing here when the
                     ;; interrupt is handled, not after the UNIMP.
-                    (note-this-location vop :step-before-vop)
+                    (note-this-location vop :internal-error)
                     ;; Construct a trap code with the low bits from
                     ;; SINGLE-STEP-AROUND-TRAP and the high bits from
                     ;; the register number of CALLABLE-TN.
                     (inst unimp (logior single-step-around-trap
                                         (ash (reg-tn-encoding callable-tn)
-                                             5)))
+                                             8)))
                     (emit-label step-done-label))))
-           ,@(if named
-                 `((sc-case name
-                     (descriptor-reg (move name-pass name))
-                     (control-stack
-                      (loadw name-pass cfp-tn (tn-offset name))
-                      (do-next-filler))
-                     (constant
-                      (loadw name-pass code-tn (tn-offset name)
-                             other-pointer-lowtag)
-                      (do-next-filler)))
-                   ;; The step instrumenting must be done after
-                   ;; FUNCTION is loaded, but before ENTRY-POINT is
-                   ;; calculated.
-                   (insert-step-instrumenting name-pass)
-                   ;; The raw-addr (ENTRY-POINT) will be one of:
-                   ;; closure_tramp, undefined_tramp, or somewhere
-                   ;; within a simple-fun object.  If the latter, then
-                   ;; it is essential (due to it being an interior
-                   ;; pointer) that the function itself be in a
-                   ;; register before the raw-addr is loaded.
-                   (sb!assem:without-scheduling ()
-                     (loadw function name-pass fdefn-fun-slot
-                            other-pointer-lowtag)
-                     (loadw entry-point name-pass fdefn-raw-addr-slot
-                            other-pointer-lowtag))
-                   (do-next-filler))
-                 `((sc-case arg-fun
-                     (descriptor-reg (move lexenv arg-fun))
-                     (control-stack
-                      (loadw lexenv cfp-tn (tn-offset arg-fun))
-                      (do-next-filler))
-                     (constant
-                      (loadw lexenv code-tn (tn-offset arg-fun)
-                             other-pointer-lowtag)
-                      (do-next-filler)))
-                   (loadw function lexenv closure-fun-slot
-                    fun-pointer-lowtag)
-                   (do-next-filler)
-                   ;; The step instrumenting must be done before
-                   ;; after FUNCTION is loaded, but before ENTRY-POINT
-                   ;; is calculated.
-                   (insert-step-instrumenting function)
-                   (inst addi entry-point function
-                    (- (ash simple-fun-code-offset word-shift)
-                     fun-pointer-lowtag))
-                   ))
+           (declare (ignorable #'insert-step-instrumenting))
+           ,@(case named
+               ((t)
+                `((sc-case name
+                    (descriptor-reg (move name-pass name))
+                    (control-stack
+                     (loadw name-pass cfp-tn (tn-offset name))
+                     (do-next-filler))
+                    (constant
+                     (loadw name-pass code-tn (tn-offset name)
+                         other-pointer-lowtag)
+                     (do-next-filler)))
+                  ;; The step instrumenting must be done after
+                  ;; FUNCTION is loaded, but before ENTRY-POINT is
+                  ;; calculated.
+                  (insert-step-instrumenting name-pass)
+                  ;; The raw-addr (ENTRY-POINT) will be one of:
+                  ;; closure-tramp, undefined-tramp, or somewhere
+                  ;; within a simple-fun object.  If the latter, then
+                  ;; it is essential (due to it being an interior
+                  ;; pointer) that the function itself be in a
+                  ;; register before the raw-addr is loaded.
+                  (sb-assem:without-scheduling ()
+                    (loadw function name-pass fdefn-fun-slot
+                        other-pointer-lowtag)
+                    (loadw entry-point name-pass fdefn-raw-addr-slot
+                        other-pointer-lowtag))
+                  (do-next-filler)))
+               ((nil)
+                `((sc-case arg-fun
+                    (descriptor-reg (move lexenv arg-fun))
+                    (control-stack
+                     (loadw lexenv cfp-tn (tn-offset arg-fun))
+                     (do-next-filler))
+                    (constant
+                     (loadw lexenv code-tn (tn-offset arg-fun)
+                         other-pointer-lowtag)
+                     (do-next-filler)))
+                  (loadw function lexenv closure-fun-slot
+                      fun-pointer-lowtag)
+                  (do-next-filler)
+                  ;; The step instrumenting must be done before
+                  ;; after FUNCTION is loaded, but before ENTRY-POINT
+                  ;; is calculated.
+                  (insert-step-instrumenting function)
+                  (inst addi entry-point function
+                        (- (ash simple-fun-insts-offset word-shift)
+                           fun-pointer-lowtag))))
+               (:direct
+                `((inst lwz entry-point null-tn (static-fun-offset fun)))))
            (loop
              (if filler
                  (do-next-filler)
@@ -857,16 +825,17 @@ default-value-8
                                         lra-label temp)
                 (when cur-nfp
                   (load-stack-tn cur-nfp nfp-save))))
-             (:tail)))
-       (trace-table-entry trace-table-normal))))
-
+             (:tail))))))
 
 (define-full-call call nil :fixed nil)
 (define-full-call call-named t :fixed nil)
+(define-full-call static-call-named :direct :fixed nil)
 (define-full-call multiple-call nil :unknown nil)
 (define-full-call multiple-call-named t :unknown nil)
+(define-full-call static-multiple-call-named :direct :unknown nil)
 (define-full-call tail-call nil :tail nil)
 (define-full-call tail-call-named t :tail nil)
+(define-full-call static-tail-call-named :direct :tail nil)
 
 (define-full-call call-variable nil :fixed t)
 (define-full-call multiple-call-variable nil :unknown t)
@@ -897,7 +866,7 @@ default-value-8
         (inst addi nsp-tn cur-nfp
               (- (bytes-needed-for-non-descriptor-stack-frame)
                  number-stack-displacement))))
-    (inst lr temp (make-fixup 'tail-call-variable :assembly-routine))
+    (load-asm-rtn-addr temp 'tail-call-variable)
     (inst mtlr temp)
     (inst blr)))
 
@@ -914,7 +883,6 @@ default-value-8
   (:temporary (:scs (interior-reg)) lip)
   (:vop-var vop)
   (:generator 6
-    (trace-table-entry trace-table-fun-epilogue)
     (move lra return-pc)
     ;; Clear the number stack.
     (let ((cur-nfp (current-nfp-tn vop)))
@@ -926,8 +894,7 @@ default-value-8
     (move csp-tn cfp-tn)
     (move cfp-tn old-fp)
     ;; Out of here.
-    (lisp-return lra lip :offset 2)
-    (trace-table-entry trace-table-normal)))
+    (lisp-return lra lip :offset 2)))
 
 ;;; Do unknown-values return of a fixed number of values.  The Values are
 ;;; required to be set up in the standard passing locations.  Nvals is the
@@ -958,7 +925,6 @@ default-value-8
   (:temporary (:scs (interior-reg)) lip)
   (:vop-var vop)
   (:generator 6
-    (trace-table-entry trace-table-fun-epilogue)
     (move lra return-pc)
     ;; Clear the number stack.
     (let ((cur-nfp (current-nfp-tn vop)))
@@ -985,8 +951,7 @@ default-value-8
              (dolist (reg (subseq (list a0 a1 a2 a3) nvals))
                (move reg null-tn)))
            ;; And away we go.
-           (lisp-return lra lip)))
-    (trace-table-entry trace-table-normal)))
+           (lisp-return lra lip)))))
 
 ;;; Do unknown-values return of an arbitrary number of values (passed
 ;;; on the stack.)  We check for the common case of a single return
@@ -1008,7 +973,6 @@ default-value-8
   (:temporary (:sc any-reg) temp)
   (:vop-var vop)
   (:generator 13
-    (trace-table-entry trace-table-fun-epilogue)
     (move lra lra-arg)
     (let ((not-single (gen-label)))
       ;; Clear the number stack.
@@ -1030,21 +994,12 @@ default-value-8
       (move old-fp old-fp-arg)
       (move vals vals-arg)
       (move nvals nvals-arg)
-      (inst lr temp (make-fixup 'return-multiple :assembly-routine))
+      (load-asm-rtn-addr temp 'return-multiple)
       (inst mtlr temp)
-      (inst blr))
-    (trace-table-entry trace-table-normal)))
+      (inst blr))))
 
 ;;;; XEP hackery:
 
-;;; We don't need to do anything special for regular functions.
-(define-vop (setup-environment)
-  (:info label)
-  (:ignore label)
-  (:generator 0
-    ;; Don't bother doing anything.
-    ))
-
 ;;; Get the lexical environment from its passing location.
 (define-vop (setup-closure-environment)
   (:temporary (:sc descriptor-reg :offset lexenv-offset :target closure
@@ -1062,57 +1017,98 @@ default-value-8
 (define-vop (copy-more-arg)
   (:temporary (:sc any-reg :offset nl0-offset) result)
   (:temporary (:sc any-reg :offset nl1-offset) count)
-  (:temporary (:sc any-reg :offset nl2-offset) src)
   (:temporary (:sc any-reg :offset nl3-offset) dst)
   (:temporary (:sc descriptor-reg :offset l0-offset) temp)
   (:info fixed)
+  (:vop-var vop)
   (:generator 20
-    (let ((loop (gen-label))
-          (do-regs (gen-label))
-          (done (gen-label)))
-      (when (< fixed register-arg-count)
-        ;; Save a pointer to the results so we can fill in register args.
-        ;; We don't need this if there are more fixed args than reg args.
-        (move result csp-tn))
-      ;; Allocate the space on the stack.
-      (cond ((zerop fixed)
-             (inst cmpwi nargs-tn 0)
-             (inst add csp-tn csp-tn nargs-tn)
-             (inst beq done))
+    ;; Compute the end of the fixed stack frame (start of the MORE arg
+    ;; area) into RESULT.
+    (inst addi result cfp-tn
+          (* n-word-bytes (sb-allocated-size 'control-stack)))
+    ;; We can set up the NFP any time we want (can't we?), so lets get
+    ;; it out of the way early.
+    (let ((nfp-tn (current-nfp-tn vop)))
+      (when nfp-tn
+        (let ((nbytes (bytes-needed-for-non-descriptor-stack-frame)))
+          (when (> nbytes number-stack-displacement)
+            (inst stwu nsp-tn nsp-tn (- nbytes))
+            (inst addi nfp-tn nsp-tn number-stack-displacement)))))
+    ;; Compute the end of the MORE arg area (and our overall frame
+    ;; allocation) into the stack pointer, skipping the rest of the
+    ;; VOP if there are no MORE args.  FIXME: If there are more fixed
+    ;; args than there are slots allocated for the control stack frame
+    ;; then this can expose some of the MORE args to being clobbered
+    ;; if we take an asynchronous interrupt before they are copied to
+    ;; their proper location.
+    (cond ((zerop fixed)
+           (inst cmpwi nargs-tn 0)
+           (inst add csp-tn result nargs-tn)
+           (inst beq DONE))
+          (t
+           (inst addic. count nargs-tn (- (fixnumize fixed)))
+           (inst ble (assemble (:elsewhere)
+                       FIX-CSP
+                       ;; If the fixed args aren't fully supplied, the
+                       ;; (INST ADD CSP-TN RESULT COUNT) below could
+                       ;; expose part of our stack area, so branch to
+                       ;; *ELSEWHERE* and do a straight move instead.
+                       (move csp-tn result)
+                       (inst b DONE)
+                       ;; VALUES here to prevent ASSEMBLE from
+                       ;; emitting it as a label (again).
+                       (values FIX-CSP)))
+           (inst add csp-tn result count)))
+    (when (< fixed register-arg-count)
+      ;; We must stop when we run out of stack args, not when we run out of
+      ;; more args.
+      (inst addic. count nargs-tn (- (fixnumize register-arg-count)))
+      ;; Everything of interest is in registers.
+      (inst ble DO-REGS))
+    ;; Initialize dst to be end of stack.
+    (move dst csp-tn)
+
+    LOOP
+    ;; FIXME: This loop can end up copying more argument words than
+    ;; strictly necessary.  This doesn't cause incorrect operation,
+    ;; but is slightly wasteful.
+    (let ((delta (- (sb-allocated-size 'control-stack) fixed)))
+      (cond ((zerop delta)
+             ;; If DELTA is zero then all of the MORE args are in the
+             ;; right place, so we don't really need to do anything.
+             )
+            ((plusp delta)
+             ;; If DELTA is positive then the allocated stack frame is
+             ;; overlapping some of the MORE args, and we need to copy
+             ;; the list starting from the end (so that we don't
+             ;; overwrite any elements before they're copied).
+             (inst lwz temp dst (- (* (1+ delta) n-word-bytes)))
+             (inst stwu temp dst (- n-word-bytes))
+             (inst cmpw dst result)
+             (inst bgt LOOP))
             (t
-             (inst addic. count nargs-tn (- (fixnumize fixed)))
-             (inst ble done)
-             (inst add csp-tn csp-tn count)))
-      (when (< fixed register-arg-count)
-        ;; We must stop when we run out of stack args, not when we run out of
-        ;; more args.
-        (inst addic. count nargs-tn (- (fixnumize register-arg-count)))
-        ;; Everything of interest is in registers.
-        (inst ble do-regs))
-      ;; Initialize dst to be end of stack.
-      (move dst csp-tn)
-      ;; Initialize src to be end of args.
-      (inst add src cfp-tn nargs-tn)
+             ;; If DELTA is negative then the start of the MORE args
+             ;; is beyond the end of the allocated stack frame, and we
+             ;; need to copy the list from the start in order to close
+             ;; the gap.
+             (inst lwz temp result (- (* delta n-word-bytes)))
+             (inst stw temp result 0)
+             (inst addi result result n-word-bytes)
+             (inst cmpw dst result)
+             (inst bgt LOOP))))
 
-      (emit-label loop)
-      ;; *--dst = *--src, --count
-      (inst lwzu temp src (- n-word-bytes))
-      (inst addic. count count (- (fixnumize 1)))
-      (inst stwu temp dst (- n-word-bytes))
-      (inst bgt loop)
-
-      (emit-label do-regs)
-      (when (< fixed register-arg-count)
-        ;; Now we have to deposit any more args that showed up in registers.
-        (inst subic. count nargs-tn (fixnumize fixed))
-        (do ((i fixed (1+ i)))
-            ((>= i register-arg-count))
-          ;; Don't deposit any more than there are.
-          (inst beq done)
-          (inst subic. count count (fixnumize 1))
-          ;; Store it relative to the pointer saved at the start.
-          (storew (nth i *register-arg-tns*) result (- i fixed))))
-      (emit-label done))))
+    DO-REGS
+    (when (< fixed register-arg-count)
+      ;; Now we have to deposit any more args that showed up in registers.
+      (inst subic. count nargs-tn (fixnumize fixed))
+      (do ((i fixed (1+ i)))
+          ((>= i register-arg-count))
+        ;; Don't deposit any more than there are.
+        (inst beq DONE)
+        (inst subic. count count (fixnumize 1))
+        ;; Store it relative to the pointer saved at the start.
+        (storew (nth i *register-arg-tns*) result (- i fixed))))
+    DONE))
 
 
 ;;; More args are stored consecutively on the stack, starting
@@ -1122,8 +1118,23 @@ default-value-8
   (:variant 0 0)
   (:translate %more-arg))
 
+(define-vop (more-arg-or-nil)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg) :to (:result 1))
+         (count :scs (any-reg)))
+  (:info index)
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:result-types *)
+  (:generator 3
+    (inst cmpwi count (fixnumize index))
+    (move value null-tn)
+    (inst ble done)
+    (inst lwz value object (ash index word-shift))
+    done))
+
 ;;; Turn more arg (context, count) into a list.
-(define-vop (listify-rest-args)
+(define-vop ()
+  (:translate %listify-rest-args)
   (:args (context-arg :target context :scs (descriptor-reg))
          (count-arg :target count :scs (any-reg)))
   (:arg-types * tagged-num)
@@ -1133,7 +1144,6 @@ default-value-8
   (:temporary (:scs (non-descriptor-reg) :from :eval) dst)
   (:temporary (:sc non-descriptor-reg :offset nl3-offset) pa-flag)
   (:results (result :scs (descriptor-reg)))
-  (:translate %listify-rest-args)
   (:policy :safe)
   (:node-var node)
   (:generator 20
@@ -1149,7 +1159,7 @@ default-value-8
       (inst beq done)
 
     ;; We need to do this atomically.
-    (pseudo-atomic (pa-flag)
+    (pseudo-atomic (pa-flag :sync nil)
       ;; Allocate a cons (2 words) for each item.
       (if dx-p
           (progn
@@ -1161,7 +1171,7 @@ default-value-8
             (inst add csp-tn csp-tn temp))
           (progn
             (inst slwi temp count 1)
-            (allocation result temp list-pointer-lowtag
+            (allocation 'list temp list-pointer-lowtag result
                         :temp-tn dst
                         :flag-tn pa-flag)
             (move dst result)))
@@ -1199,9 +1209,9 @@ default-value-8
 ;;; preventing this info from being returned as values.  What we do is
 ;;; compute (- SUPPLIED FIXED), and return a pointer that many words
 ;;; below the current stack top.
-(define-vop (more-arg-context)
+(define-vop ()
   (:policy :fast-safe)
-  (:translate sb!c::%more-arg-context)
+  (:translate sb-c::%more-arg-context)
   (:args (supplied :scs (any-reg)))
   (:arg-types tagged-num (:constant fixnum))
   (:info fixed)
@@ -1215,56 +1225,37 @@ default-value-8
 
 (define-vop (verify-arg-count)
   (:policy :fast-safe)
-  (:translate sb!c::%verify-arg-count)
   (:args (nargs :scs (any-reg)))
-  (:arg-types positive-fixnum (:constant t))
-  (:info count)
+  (:arg-types positive-fixnum (:constant t) (:constant t))
+  (:info min max)
   (:vop-var vop)
   (:save-p :compute-only)
   (:generator 3
-   (inst twi :ne nargs (fixnumize count))))
-
-;;; Signal various errors.
-(macrolet ((frob (name error translate &rest args)
-             `(define-vop (,name)
-                ,@(when translate
-                    `((:policy :fast-safe)
-                      (:translate ,translate)))
-                (:args ,@(mapcar #'(lambda (arg)
-                                     `(,arg :scs (any-reg descriptor-reg)))
-                                 args))
-                (:vop-var vop)
-                (:save-p :compute-only)
-                (:generator 1000
-                  (error-call vop ',error ,@args)))))
-  (frob arg-count-error invalid-arg-count-error
-    sb!c::%arg-count-error nargs)
-  (frob type-check-error object-not-type-error sb!c::%type-check-error
-    object type)
-  (frob layout-invalid-error layout-invalid-error sb!c::%layout-invalid-error
-    object layout)
-  (frob odd-key-args-error odd-key-args-error
-        sb!c::%odd-key-args-error)
-  (frob unknown-key-arg-error unknown-key-arg-error
-        sb!c::%unknown-key-arg-error key)
-  (frob nil-fun-returned-error nil-fun-returned-error nil fun))
+    (cond ((not min)
+           (inst twi :ne nargs (fixnumize max)))
+          (max
+           (when (plusp min)
+             (inst twi :llt nargs (fixnumize min)))
+           (inst twi :lgt nargs (fixnumize max)))
+          ((plusp min)
+           (inst twi :llt nargs (fixnumize min))))))
 
 (define-vop (step-instrument-before-vop)
   (:temporary (:scs (descriptor-reg)) stepping)
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 3
-    ;; Get the symbol-value of SB!IMPL::*STEPPING*
-    #!-sb-thread
-    (load-symbol-value stepping sb!impl::*stepping*)
-    #!+sb-thread
+    ;; Get the symbol-value of SB-IMPL::*STEPPING*
+    #-sb-thread
+    (load-symbol-value stepping sb-impl::*stepping*)
+    #+sb-thread
     (loadw stepping thread-base-tn thread-stepping-slot)
-    (inst cmpw stepping null-tn)
-    ;; If it's not null, trap.
+    (inst cmpwi stepping 0)
+    ;; If it's not zero, trap.
     (inst beq DONE)
     ;; CONTEXT-PC will be pointing here when the interrupt is handled,
     ;; not after the UNIMP.
-    (note-this-location vop :step-before-vop)
+    (note-this-location vop :internal-error)
     ;; CALLEE-REGISTER-OFFSET isn't needed for before-traps, so we
     ;; can just use a bare SINGLE-STEP-BEFORE-TRAP as the code.
     (inst unimp single-step-before-trap)
