@@ -10,28 +10,27 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; The component that is currently being compiled. TNs are allocated
 ;;; in this component.
 (defvar *component-being-compiled*)
 
-;;; DO-PACKED-TNS (TN-Var Component [Result]) Declaration* Form*
+;;; DO-PACKED-TNS (TN-Var Component [Result]) Form*
 ;;;
 ;;; Iterate over all packed TNs allocated in COMPONENT.
 (defmacro do-packed-tns ((tn component &optional result) &body body)
-  (let ((n-component (gensym)))
-    `(let ((,n-component (component-info ,component)))
-       (do ((,tn (ir2-component-normal-tns ,n-component) (tn-next ,tn)))
-           ((null ,tn))
-         ,@body)
-       (do ((,tn (ir2-component-restricted-tns ,n-component) (tn-next ,tn)))
-           ((null ,tn))
-         ,@body)
-       (do ((,tn (ir2-component-wired-tns ,n-component) (tn-next ,tn)))
-           ((null ,tn)
-            ,result)
-         ,@body))))
+  (with-unique-names (n-component tns more-tns outer inner)
+    `(prog* ((,n-component (component-info ,component))
+             (,tn       (ir2-component-normal-tns ,n-component))
+             (,tns      (ir2-component-restricted-tns ,n-component))
+             (,more-tns (ir2-component-wired-tns ,n-component)))
+       (when ,tn (go ,inner))
+       ,outer (when (eq ,tns :done) (return ,result))
+              (shiftf ,tn ,tns ,more-tns :done)
+              (unless ,tn (go ,outer))
+       ,inner (progn ,@body)
+              (if (setq ,tn (tn-next ,tn)) (go ,inner) (go ,outer)))))
 
 (defun set-ir2-physenv-live-tns (value instance)
   (setf (ir2-physenv-live-tns instance) value))
@@ -110,21 +109,24 @@
 ;;; Create a packed TN of the specified primitive-type in the
 ;;; *COMPONENT-BEING-COMPILED*. We use the SCs from the primitive type
 ;;; to determine which SCs it can be packed in.
-(defun make-normal-tn (type)
-  (declare (type primitive-type type))
+(defun make-normal-tn (primitive-type &optional (type *universal-type*))
+  (declare (type primitive-type primitive-type)
+           (type ctype type))
   (let* ((component (component-info *component-being-compiled*))
          (res (make-tn (incf (ir2-component-global-tn-counter component))
-                       :normal type nil)))
+                       :normal primitive-type nil)))
+    (setf (tn-type res) (single-value-type type))
     (push-in tn-next res (ir2-component-normal-tns component))
     res))
 
 ;;; Create a normal packed TN with representation indicated by SCN.
-(defun make-representation-tn (ptype scn)
+(defun make-representation-tn (ptype scn &optional (type *universal-type*))
   (declare (type primitive-type ptype) (type sc-number scn))
   (let* ((component (component-info *component-being-compiled*))
          (res (make-tn (incf (ir2-component-global-tn-counter component))
                        :normal ptype
                        (svref *backend-sc-numbers* scn))))
+    (setf (tn-type res) (single-value-type type))
     (push-in tn-next res (ir2-component-normal-tns component))
     res))
 
@@ -132,7 +134,7 @@
 ;;; and FSC to record where it goes, and then put it on the current component's
 ;;; Wired-TNs list. Ptype is the TN's primitive-type, which may be NIL in VOP
 ;;; temporaries.
-(defun make-wired-tn (ptype scn offset)
+(defun make-wired-tn (ptype scn offset &optional (type *universal-type*))
   (declare (type (or primitive-type null) ptype)
            (type sc-number scn) (type unsigned-byte offset))
   (let* ((component (component-info *component-being-compiled*))
@@ -140,20 +142,34 @@
                        :normal ptype
                        (svref *backend-sc-numbers* scn))))
     (setf (tn-offset res) offset)
+    (when ptype
+      (setf (tn-type res) (if type
+                              (single-value-type type)
+                              *universal-type*)))
     (push-in tn-next res (ir2-component-wired-tns component))
     res))
+(defun sb-vm::make-wired-tn* (prim-type-name scn offset)
+  (make-wired-tn (primitive-type-or-lose prim-type-name) scn offset))
 
 ;;; Create a packed TN restricted to the SC with number SCN. Ptype is as
 ;;; for MAKE-WIRED-TN.
-(defun make-restricted-tn (ptype scn)
+(defun make-restricted-tn (ptype scn &optional (type *universal-type*))
   (declare (type (or primitive-type null) ptype) (type sc-number scn))
   (let* ((component (component-info *component-being-compiled*))
          (res (make-tn (incf (ir2-component-global-tn-counter component))
                        :normal ptype
                        (svref *backend-sc-numbers* scn))))
+    (when ptype
+      (setf (tn-type res) (if type
+                              (single-value-type type)
+                              *universal-type*)))
     (push-in tn-next res (ir2-component-restricted-tns component))
     res))
 
+(defun make-unused-tn ()
+  (make-tn (incf (ir2-component-global-tn-counter (component-info *component-being-compiled*)))
+           :unused nil nil))
+
 ;;; Make TN be live throughout PHYSENV. Return TN. In the DEBUG case,
 ;;; the TN is treated normally in blocks in the environment which
 ;;; reference the TN, allowing targeting to/from the TN. This results
@@ -184,6 +200,7 @@
   tn)
 
 ;;; Specify that SAVE be used as the save location for TN. TN is returned.
+#-fp-and-pc-standard-save
 (defun specify-save-tn (tn save)
   (declare (type tn tn save))
   (aver (eq (tn-kind save) :normal))
@@ -196,32 +213,65 @@
          (component-info *component-being-compiled*)))
   tn)
 
-;;; Create a constant TN. The implementation dependent
+;;; Create a constant TN. The backend dependent
 ;;; IMMEDIATE-CONSTANT-SC function is used to determine whether the
 ;;; constant has an immediate representation.
-(defun make-constant-tn (constant)
+;;; FIXME: this can create multiple boxed TNs (consuming more than one
+;;; slot in the boxed constant vector) for a given leaf when called by
+;;; EMIT-MOVES-AND-COERCIONS. That's wasteful.
+(defun make-constant-tn (constant &optional force-boxed)
   (declare (type constant constant))
-  (let* ((component (component-info *component-being-compiled*))
-         (immed (immediate-constant-sc (constant-value constant)))
-         (sc (svref *backend-sc-numbers*
-                    (or immed (sc-number-or-lose 'constant))))
-         (res (make-tn 0 :constant (primitive-type (leaf-type constant)) sc)))
-    (unless immed
-      (let ((constants (ir2-component-constants component)))
-        (setf (tn-offset res) (fill-pointer constants))
-        (vector-push-extend constant constants)))
-    (push-in tn-next res (ir2-component-constant-tns component))
-    (setf (tn-leaf res) constant)
-    res))
+  (or (leaf-info constant)
+      (multiple-value-bind (immed null-offset)
+          (immediate-constant-sc (constant-value constant))
+        ;; currently NULL-OFFSET is used only on ARM64
+        (if null-offset
+            (setf (leaf-info constant)
+                  (component-live-tn
+                   (make-wired-tn (primitive-type (leaf-type constant))
+                                  immed
+                                  null-offset)))
+            (let* ((boxed (or (not immed)
+                              (boxed-immediate-sc-p immed)))
+                   (component (component-info *component-being-compiled*))
+                   ;; If a constant has either an immediate or boxed
+                   ;; representation (e.g. double-float) postpone the SC
+                   ;; choice until SELECT-REPRESENTATIONS.
+                   (sc (cond (boxed
+                              (if immed
+                                  (svref *backend-sc-numbers* immed)
+                                  (sc-or-lose 'constant)))
+                             (force-boxed
+                              (setf immed nil)
+                              (sc-or-lose 'constant))))
+                   (res (make-tn 0 :constant (primitive-type (leaf-type constant)) sc)))
+              ;; Objects of type SYMBOL can be immediate but they still go in the constants
+              ;; because liveness depends on pointer tracing without looking at code-fixups.
+              (when (and sc
+                         (or (not immed)
+                             #+immobile-space
+                             (let ((val (constant-value constant)))
+                               (or (and (symbolp val) (not (sb-vm:static-symbol-p val)))
+                                   (typep val 'wrapper)))))
+                (let ((constants (ir2-component-constants component)))
+                  (setf (tn-offset res)
+                        (vector-push-extend constant constants))))
+              (when sc
+                (setf (leaf-info constant) res))
+              (push-in tn-next res (ir2-component-constant-tns component))
+              (setf (tn-type res) (leaf-type constant))
+              (setf (tn-leaf res) constant)
+              res)))))
 
 (defun make-load-time-value-tn (handle type)
   (let* ((component (component-info *component-being-compiled*))
          (sc (svref *backend-sc-numbers*
-                    (sc-number-or-lose 'constant)))
+                    sb-vm:constant-sc-number))
          (res (make-tn 0 :constant (primitive-type type) sc))
          (constants (ir2-component-constants component)))
-    (setf (tn-offset res) (fill-pointer constants))
-    (vector-push-extend (cons :load-time-value handle) constants)
+    (setf (tn-offset res) (fill-pointer constants)
+          (tn-type res) type)
+    (vector-push-extend (list :load-time-value handle res) constants)
     (push-in tn-next res (ir2-component-constant-tns component))
     res))
 
@@ -231,7 +281,8 @@
   (let* ((component (component-info *component-being-compiled*))
          (res (make-tn (incf (ir2-component-global-tn-counter component))
                        :alias (tn-primitive-type tn) nil)))
-    (setf (tn-save-tn res) tn)
+    (setf (tn-save-tn res) tn
+          (tn-type res) (tn-type tn))
     (push-in tn-next res
              (ir2-component-alias-tns component))
     res))
@@ -239,26 +290,30 @@
 ;;; Return a load-time constant TN with the specified KIND and INFO.
 ;;; If the desired CONSTANTS entry already exists, then reuse it,
 ;;; otherwise allocate a new load-time constant slot.
+;;; FIXME: this function deserves a comment about why it creates a new TN even
+;;; when KIND + INFO are already found. It's as if we need to maintain a 1:1
+;;; relation between TN and TN-REF for a subset of constants, but it's unclear
+;;; why that should be necessary. The shape of this algorithm suggests more than
+;;; mere oversight, and that it was a deliberate choice.
 (defun make-load-time-constant-tn (kind info)
   (declare (type keyword kind))
   (let* ((component (component-info *component-being-compiled*))
          (res (make-tn 0
                        :constant
                        *backend-t-primitive-type*
-                       (svref *backend-sc-numbers*
-                              (sc-number-or-lose 'constant))))
+                       (svref *backend-sc-numbers* sb-vm:constant-sc-number)))
          (constants (ir2-component-constants component)))
-
-    (do ((i 0 (1+ i)))
+    (setf (tn-type res) *universal-type*)
+    (do ((i 1 (1+ i)))
         ((= i (length constants))
          (setf (tn-offset res) i)
-         (vector-push-extend (cons kind info) constants))
+         (vector-push-extend (list kind info res) constants))
       (let ((entry (aref constants i)))
         (when (and (consp entry)
                    (eq (car entry) kind)
-                   (or (eq (cdr entry) info)
+                   (or (eq (cadr entry) info)
                        (and (consp info)
-                            (equal (cdr entry) info))))
+                            (equal (cadr entry) info))))
           (setf (tn-offset res) i)
           (return))))
 
@@ -274,9 +329,12 @@
 (defun reference-tn (tn write-p)
   (declare (type tn tn) (type boolean write-p))
   (let ((res (make-tn-ref tn write-p)))
-    (if write-p
-        (push-in tn-ref-next res (tn-writes tn))
-        (push-in tn-ref-next res (tn-reads tn)))
+    (unless (eql (tn-kind tn) :unused)
+      (when (tn-primitive-type tn)
+        (aver (setf (tn-ref-type res) (tn-type tn))))
+      (if write-p
+          (push-in tn-ref-next res (tn-writes tn))
+          (push-in tn-ref-next res (tn-reads tn))))
     res))
 
 ;;; Make TN-REFS to reference each TN in TNs, linked together by
@@ -327,11 +385,7 @@
            (type template template) (type tn x y))
   (let ((arg (reference-tn x nil))
         (result (reference-tn y t)))
-    (multiple-value-bind (first last)
-        (funcall (template-emit-function template) node block template arg
-                 result)
-      (insert-vop-sequence first last block before)
-      last)))
+    (emit-and-insert-vop node block template arg result before)))
 
 ;;; like EMIT-MOVE-TEMPLATE, except that we pass in INFO args too
 (defun emit-load-template (node block template x y info &optional before)
@@ -339,11 +393,7 @@
            (type template template) (type tn x y))
   (let ((arg (reference-tn x nil))
         (result (reference-tn y t)))
-    (multiple-value-bind (first last)
-        (funcall (template-emit-function template) node block template arg
-                 result info)
-      (insert-vop-sequence first last block before)
-      last)))
+    (emit-and-insert-vop node block template arg result before info)))
 
 ;;; like EMIT-MOVE-TEMPLATE, except that the VOP takes two args
 (defun emit-move-arg-template (node block template x f y &optional before)
@@ -353,29 +403,26 @@
         (f-ref (reference-tn f nil))
         (y-ref (reference-tn y t)))
     (setf (tn-ref-across x-ref) f-ref)
-    (multiple-value-bind (first last)
-        (funcall (template-emit-function template) node block template x-ref
-                 y-ref)
-      (insert-vop-sequence first last block before)
-      last)))
+    (emit-and-insert-vop node block template x-ref y-ref before)))
 
 ;;; like EMIT-MOVE-TEMPLATE, except that the VOP takes no args
 (defun emit-context-template (node block template y &optional before)
   (declare (type node node) (type ir2-block block)
            (type template template) (type tn y))
   (let ((y-ref (reference-tn y t)))
-    (multiple-value-bind (first last)
-        (funcall (template-emit-function template) node block template nil
-                 y-ref)
-      (insert-vop-sequence first last block before)
-      last)))
+    (emit-and-insert-vop node block template nil y-ref before)))
 
 ;;; Return the label marking the start of Block, assigning one if necessary.
 (defun block-label (block)
   (declare (type cblock block))
   (let ((2block (block-info block)))
     (or (ir2-block-%label 2block)
-        (setf (ir2-block-%label 2block) (gen-label)))))
+        (setf (ir2-block-%label 2block) (gen-label "basic block")))))
+(defun block-trampoline (block)
+  (declare (type cblock block))
+  (let ((2block (block-info block)))
+    (or (ir2-block-%trampoline-label 2block)
+        (setf (ir2-block-%trampoline-label 2block) (gen-label "trampoline")))))
 
 ;;; Return true if Block is emitted immediately after the block ended by Node.
 (defun drop-thru-p (node block)
@@ -383,35 +430,46 @@
   (let ((next-block (ir2-block-next (block-info (node-block node)))))
     (aver (eq node (block-last (node-block node))))
     (eq next-block (block-info block))))
+(defun register-drop-thru (block)
+  (declare (type cblock block))
+  (let ((2block (block-info block)))
+    (setf (ir2-block-dropped-thru-to 2block) t))
+  nil)
 
-;;; Link a list of VOPs from First to Last into Block, Before the specified
-;;; VOP. If Before is NIL, insert at the end.
-(defun insert-vop-sequence (first last block before)
-  (declare (type vop first last) (type ir2-block block)
+;;; Insert a VOP into BLOCK, before the specified
+;;; BEFORE VOP. If BEFORE is NIL, insert at the end.
+(defun insert-vop (vop block before)
+  (declare (type vop vop) (type ir2-block block)
            (type (or vop null) before))
   (if before
       (let ((prev (vop-prev before)))
-        (setf (vop-prev first) prev)
+        (setf (vop-prev vop) prev)
         (if prev
-            (setf (vop-next prev) first)
-            (setf (ir2-block-start-vop block) first))
-        (setf (vop-next last) before)
-        (setf (vop-prev before) last))
+            (setf (vop-next prev) vop)
+            (setf (ir2-block-start-vop block) vop))
+        (setf (vop-next vop) before)
+        (setf (vop-prev before) vop))
       (let ((current (ir2-block-last-vop block)))
-        (setf (vop-prev first) current)
-        (setf (ir2-block-last-vop block) last)
+        (setf (vop-prev vop) current)
+        (setf (ir2-block-last-vop block) vop)
         (if current
-            (setf (vop-next current) first)
-            (setf (ir2-block-start-vop block) first))))
+            (setf (vop-next current) vop)
+            (setf (ir2-block-start-vop block) vop))))
   (values))
 
+(defun emit-and-insert-vop (node block template arg result before
+                            &optional info)
+  (let ((vop (emit-vop node block template arg result info)))
+    (insert-vop vop block before)
+    vop))
+
 ;;; Delete all of the TN-REFs associated with VOP and remove VOP from the IR2.
 (defun delete-vop (vop)
   (declare (type vop vop))
   (do ((ref (vop-refs vop) (tn-ref-next-ref ref)))
       ((null ref))
-    (delete-tn-ref ref))
-
+    (unless (eql (tn-kind (tn-ref-tn ref)) :unused)
+      (delete-tn-ref ref)))
   (let ((prev (vop-prev vop))
         (next (vop-next vop))
         (block (vop-block vop)))
@@ -422,7 +480,7 @@
         (setf (vop-prev next) prev)
         (setf (ir2-block-last-vop block) prev)))
 
-  (values))
+  nil)
 
 ;;; Return a list of N normal TNs of the specified primitive type.
 (defun make-n-tns (n ptype)
@@ -445,15 +503,13 @@
   (aver (eq (tn-kind tn) :constant))
   (constant-value (tn-leaf tn)))
 
-(defun immediate-tn-p (tn)
+(defun constant-tn-p (tn)
   (declare (type tn tn))
   (let ((leaf (tn-leaf tn)))
     ;; Leaves with KIND :CONSTANT can have NIL as the leaf if they
     ;; represent load time values.
     (and leaf
-         (eq (tn-kind tn) :constant)
-         (eq (immediate-constant-sc (constant-value leaf))
-             (sc-number-or-lose 'sb!vm::immediate)))))
+         (eq (tn-kind tn) :constant))))
 
 ;;; Force TN to be allocated in a SC that doesn't need to be saved: an
 ;;; unbounded non-save-p SC. We don't actually make it a real "restricted" TN,
