@@ -10,27 +10,88 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!C")
+(in-package "SB-C")
 
 ;;; the largest number of TNs whose liveness changes that we can have
 ;;; in any block
-(def!constant local-tn-limit 64)
+(defconstant local-tn-limit 64)
 
 (deftype local-tn-number () `(integer 0 (,local-tn-limit)))
 (deftype local-tn-count () `(integer 0 ,local-tn-limit))
 (deftype local-tn-vector () `(simple-vector ,local-tn-limit))
 (deftype local-tn-bit-vector () `(simple-bit-vector ,local-tn-limit))
 
-;;; type of an SC number
-(deftype sc-number () `(integer 0 (,sc-number-limit)))
+;;; vectors indexed by SC numbers
+(deftype sc-vector () `(simple-vector ,sb-vm:sc-number-limit))
+(deftype sc-bit-vector () `(simple-bit-vector ,sb-vm:sc-number-limit))
 
-;;; types for vectors indexed by SC numbers
-(deftype sc-vector () `(simple-vector ,sc-number-limit))
-(deftype sc-bit-vector () `(simple-bit-vector ,sc-number-limit))
+;;; Bitset representation of a set of locations in a finite SC.
+(deftype sc-locations ()
+  `(unsigned-byte ,sb-vm:finite-sc-offset-limit))
+
+(defun make-sc-locations (locations)
+  (reduce (lambda (value location)
+            (check-type location sb-vm:finite-sc-offset)
+            (dpb 1 (byte 1 location) value))
+          locations :initial-value 0))
+
+(declaim (inline sc-offset-to-sc-locations)
+         (ftype (sfunction (sb-vm:finite-sc-offset) sc-locations)
+                sc-offset-to-sc-locations))
+(defun sc-offset-to-sc-locations (offset)
+  (dpb 1 (byte 1 offset) 0))
+
+(declaim (inline sc-locations-count)
+         (ftype (sfunction (sc-locations) (integer 0 #.sb-vm:finite-sc-offset-limit))
+                sc-locations-count))
+(defun sc-locations-count (locations)
+  (logcount locations))
+
+(declaim (inline sc-locations-first)
+         (ftype (sfunction (sc-locations) sb-vm:finite-sc-offset)
+                sc-locations-first))
+(defun sc-locations-first (locations)
+  (1- (integer-length (logxor locations (1- locations)))))
+
+(declaim (inline sc-locations-member)
+         (ftype (sfunction (sb-vm:finite-sc-offset sc-locations) boolean)
+                sc-locations-member))
+(defun sc-locations-member (location locations)
+  (logbitp location locations))
+
+;;; This used to have two local functions in it, but when it did,
+;;; using ABCL as the build host crashed thusly with no backtrace:
+;;; (funcall (macro-function 'do-sc-locations)
+;;;           '(do-sc-locations (el (sc-locations sc) nil (sc-element-size sc))
+;;;             (feep))
+;;;            nil)
+;;;  => Debugger invoked on condition of type TYPE-ERROR
+;;;     The value NIL is not of type STRUCTURE-OBJECT.
+(defmacro do-sc-locations ((location locations
+                            &optional result increment (limit 'sb-vm:finite-sc-offset-limit))
+                           &body body)
+  (let ((mid (floor sb-vm:finite-sc-offset-limit 2)))
+    (once-only ((locations locations)
+                (increment `(the sb-vm:finite-sc-offset ,(or increment 1))))
+      (flet ((make-guarded-block (start end)
+                 (unless (and (integerp limit) (> start limit))
+                   (let ((mask (dpb -1 (byte mid start) 0)))
+                     `((when (logtest ,mask ,locations)
+                         (loop named #:noname
+                               for ,location
+                               from ,start below ,end by ,increment
+                               when (logbitp ,location ,locations)
+                                 do (locally (declare (type sb-vm:finite-sc-offset
+                                                            ,location))
+                                      ,@body))))))))
+        `(block nil
+           ,@(make-guarded-block 0   mid)
+           ,@(make-guarded-block mid limit)
+           ,result)))))
 
 ;;; the different policies we can use to determine the coding strategy
 (deftype ltn-policy ()
-  '(member :safe :small :fast :fast-safe))
+  '(member :safe :small :small-safe :fast :fast-safe))
 
 ;;;; PRIMITIVE-TYPEs
 
@@ -41,18 +102,14 @@
 ;;; representations of that value.
 (defstruct (primitive-type (:copier nil))
   ;; the name of this PRIMITIVE-TYPE
-  (name nil :type symbol)
+  (name nil :type symbol :read-only t)
   ;; a list of the SC numbers for all the SCs that a TN of this type
   ;; can be allocated in
-  (scs nil :type list)
+  (scs nil :type list :read-only t)
   ;; the Lisp type equivalent to this type. If this type could never be
-  ;; returned by PRIMITIVE-TYPE, then this is the NIL (or empty) type
-  (specifier (missing-arg) :type type-specifier)
-  ;; the template used to check that an object is of this type. This is a
-  ;; template of one argument and one result, both of primitive-type T. If
-  ;; the argument is of the correct type, then it is delivered into the
-  ;; result. If the type is incorrect, then an error is signalled.
-  (check nil :type (or template null)))
+  ;; returned by PRIMITIVE-TYPE, then this is the NIL (or empty) type.
+  ;; TYPE-SPECIFIER is too general - this doesn't allow CLASS/CLASSOID.
+  (specifier (missing-arg) :type (or symbol list) :read-only t))
 
 (defprinter (primitive-type)
   name)
@@ -167,11 +224,16 @@
   ;; the assembler label that points to the beginning of the code for
   ;; this block, or NIL when we haven't assigned a label yet
   (%label nil)
+  ;; the assembler label that points to the trampoline for this block,
+  ;; or NIL if unassigned yet. Only meaningful for local call targets.
+  (%trampoline-label nil)
+  ;; T if the preceding block assumes it can drop thru to %label
+  (dropped-thru-to nil)
   ;; list of LOCATION-INFO structures describing all the interesting
   ;; (to the debugger) locations in this block
   (locations nil :type list))
 
-(defprinter (ir2-block)
+(defprinter (ir2-block :identity t)
   (pushed :test pushed)
   (popped :test popped)
   (start-vop :test start-vop)
@@ -261,28 +323,37 @@
   ;; constant pool. A non-immediate :CONSTANT TN with offset 0 refers
   ;; to the constant in element 0, etc. Normal constants are
   ;; represented by the placing the CONSTANT leaf in this vector. A
-  ;; load-time constant is distinguished by being a cons (KIND .
-  ;; WHAT). KIND is a keyword indicating how the constant is computed,
-  ;; and WHAT is some context.
+  ;; load-time constant is distinguished by being a cons
+  ;; (KIND WHAT TN).
+  ;; KIND is a keyword indicating how the constant is computed, and
+  ;; WHAT is some context.
   ;;
   ;; These load-time constants are recognized:
   ;;
-  ;; (:entry . <function>)
+  ;; (:entry <function>)
   ;;    Is replaced by the code pointer for the specified function.
   ;;    This is how compiled code (including DEFUN) gets its hands on
   ;;    a function. <function> is the XEP lambda for the called
   ;;    function; its LEAF-INFO should be an ENTRY-INFO structure.
   ;;
-  ;; (:label . <label>)
-  ;;    Is replaced with the byte offset of that label from the start
-  ;;    of the code vector (including the header length.)
+  ;; (:fdefinition <name>)
+  ;;    Is replaced with the fdefn for NAME.
+  ;;
+  ;; (:known-fun <name>)
+  ;;    Is replaced with #'NAME for a system-internal function.
+  ;;
+  ;; (:load-time-value <handle>)
+  ;;    Is replaced with the result of executing the forms
+  ;;    to compute <handle>.
   ;;
   ;; A null entry in this vector is a placeholder for implementation
   ;; overhead that is eventually stuffed in somehow.
-  (constants (make-array 10 :fill-pointer 0 :adjustable t) :type vector)
-  ;; some kind of info about the component's run-time representation.
-  ;; This is filled in by the VM supplied SELECT-COMPONENT-FORMAT function.
-  format
+  ;; Prior to performing SORT-BOXED-CONSTANTS, the index 0 is reserved
+  ;; to signify something (other than NIL) if stored as a TN-OFFSET,
+  ;; though nothing makes use of this at present.
+  (constants (make-array 10 :fill-pointer 1 :adjustable t
+                         :initial-element :ignore)
+             :type vector :read-only t)
   ;; a list of the ENTRY-INFO structures describing all of the entries
   ;; into this component. Filled in by entry analysis.
   (entries nil :type list)
@@ -299,8 +370,9 @@
   ;; dynamic vop count info. This is needed by both ir2-convert and
   ;; setup-dynamic-count-info. (But only if we are generating code to
   ;; collect dynamic statistics.)
-  #!+sb-dyncount
-  (dyncount-info nil :type (or null dyncount-info)))
+  #+sb-dyncount
+  (dyncount-info nil :type (or null dyncount-info))
+  (avx2-used-p nil))
 
 ;;; An ENTRY-INFO condenses all the information that the dumper needs
 ;;; to create each XEP's function entry data structure. ENTRY-INFO
@@ -320,12 +392,17 @@
   ;; some string that is intended to be informative.
   (name "<not computed>" :type (or simple-string list symbol))
   ;; the argument list that the function was defined with.
-  (arguments nil :type list)
+  (arguments :unknown :type (or list (eql :unknown)))
+  ;; source form and/or docstring
+  (form/doc nil :type (or list string (cons t string)))
   ;; a function type specifier representing the arguments and results
   ;; of this function
   (type 'function :type (or list (member function)))
-  ;; docstring and/or xref information for the XEP
-  (info nil :type (or null simple-vector string (cons string simple-vector))))
+  (xref))
+(defun entry-info-type/xref (entry)
+  (let ((type (entry-info-type entry))
+        (xref (entry-info-xref entry)))
+    (if (and type xref) (cons type xref) (or type xref))))
 
 ;;; An IR2-PHYSENV is used to annotate non-LET LAMBDAs with their
 ;;; passing locations. It is stored in the PHYSENV-INFO.
@@ -349,6 +426,14 @@
   ;; implementations we may use a call instruction that requires the
   ;; return PC to be passed in a particular place.
   (return-pc-pass (missing-arg) :type tn :read-only t)
+  ;; a label that marks the first instruction after the RETURN-PC has
+  ;; been moved from its passing location to its save location.
+  #-fp-and-pc-standard-save
+  (lra-saved-pc nil :type (or label null))
+  ;; a label that marks the first instruction after the OLD-FP has
+  ;; been moved from its passing location to its save location.
+  #-fp-and-pc-standard-save
+  (cfp-saved-pc nil :type (or label null))
   ;; True if this function has a frame on the number stack. This is
   ;; set by representation selection whenever it is possible that some
   ;; function in our tail set will make use of the number stack.
@@ -365,12 +450,17 @@
   ;; the environment is properly initialized, i.e. arguments moved
   ;; from their passing locations, etc. This is the start of the
   ;; function as far as the debugger is concerned.
-  (environment-start nil :type (or label null)))
+  (environment-start nil :type (or label null))
+  (closure-save-tn nil :type (or tn null))
+  #+unwind-to-frame-and-call-vop
+  (bsp-save-tn nil :type (or tn null)))
+
 (defprinter (ir2-physenv)
   closure
   old-fp
   return-pc
-  return-pc-pass)
+  return-pc-pass
+  closure-save-tn)
 
 ;;; A RETURN-INFO is used by GTN to represent the return strategy and
 ;;; locations for all the functions in a given TAIL-SET. It is stored
@@ -386,7 +476,8 @@
   (count (missing-arg) :type (or index (member :unknown)))
   ;; If count isn't :UNKNOWN, then this is a list of the
   ;; primitive-types of each value.
-  (types () :type list)
+  (primitive-types () :type list)
+  (types nil :type list)
   ;; If kind is :FIXED, then this is the list of the TNs that we
   ;; return the values in.
   (locations () :type list))
@@ -403,68 +494,29 @@
   ;; unwind-block, so we leave this slot null.
   (home nil :type (or tn null))
   ;; the saved control stack pointer
-  (save-sp (missing-arg) :type tn)
+  (save-sp nil :type (or tn null))
   ;; the list of dynamic state save TNs
+  #-unbind-in-unwind
   (dynamic-state (list* (make-stack-pointer-tn)
                         (make-dynamic-state-tns))
                  :type list)
   ;; the target label for NLX entry
-  (target (gen-label) :type label))
+  (target (gen-label) :type label)
+  (block-tn nil :type (or tn null)))
 (defprinter (ir2-nlx-info)
   home
   save-sp
+  #-unbind-in-unwind
   dynamic-state)
-
-(defstruct (cloop (:conc-name loop-)
-                  (:predicate loop-p)
-                  (:constructor make-loop)
-                  (:copier copy-loop))
-  ;; The kind of loop that this is.  These values are legal:
-  ;;
-  ;;    :OUTER
-  ;;        This is the outermost loop structure, and represents all the
-  ;;        code in a component.
-  ;;
-  ;;    :NATURAL
-  ;;        A normal loop with only one entry.
-  ;;
-  ;;    :STRANGE
-  ;;        A segment of a "strange loop" in a non-reducible flow graph.
-  (kind (missing-arg) :type (member :outer :natural :strange))
-  ;; The first and last blocks in the loop.  There may be more than one tail,
-  ;; since there may be multiple back branches to the same head.
-  (head nil :type (or cblock null))
-  (tail nil :type list)
-  ;; A list of all the blocks in this loop or its inferiors that have a
-  ;; successor outside of the loop.
-  (exits nil :type list)
-  ;; The loop that this loop is nested within.  This is null in the outermost
-  ;; loop structure.
-  (superior nil :type (or cloop null))
-  ;; A list of the loops nested directly within this one.
-  (inferiors nil :type list)
-  (depth 0 :type fixnum)
-  ;; The head of the list of blocks directly within this loop.  We must recurse
-  ;; on INFERIORS to find all the blocks.
-  (blocks nil :type (or null cblock))
-  ;; Backend saves the first emitted block of each loop here.
-  (info nil))
-
-(defprinter (cloop :conc-name loop-)
-  kind
-  head
-  tail
-  exits
-  depth)
 
 ;;;; VOPs and templates
 
 ;;; A VOP is a Virtual Operation. It represents an operation and the
 ;;; operands to the operation.
-(def!struct (vop (:constructor make-vop (block node info args results))
-                 (:copier nil))
+(defstruct (vop (:constructor make-vop (block node info args results))
+                (:copier nil))
   ;; VOP-INFO structure containing static info about the operation
-  (info nil :type (or vop-info null))
+  (info nil :type vop-info)
   ;; the IR2-BLOCK this VOP is in
   (block (missing-arg) :type ir2-block)
   ;; VOPs evaluated after and before this one. Null at the
@@ -492,17 +544,12 @@
   ;; are read and before results are written. This is only filled in
   ;; when VOP-INFO-SAVE-P is non-null.
   (save-set nil :type (or local-tn-bit-vector null)))
-(defprinter (vop)
-  (info :prin1 (vop-info-name info))
-  args
-  results
-  (codegen-info :test codegen-info))
 
 ;;; A TN-REF object contains information about a particular reference
 ;;; to a TN. The information in TN-REFs largely determines how TNs are
 ;;; packed.
-(def!struct (tn-ref (:constructor make-tn-ref (tn write-p))
-                    (:copier nil))
+(defstruct (tn-ref (:constructor make-tn-ref (tn write-p))
+                   (:copier nil))
   ;; the TN referenced
   (tn (missing-arg) :type tn)
   ;; Is this is a write reference? (as opposed to a read reference)
@@ -524,16 +571,20 @@
   ;; vice-versa.
   (target nil :type (or null tn-ref))
   ;; the load TN allocated for this operand, if any
-  (load-tn nil :type (or tn null)))
-(defprinter (tn-ref)
-  tn
-  write-p
-  (vop :test vop :prin1 (vop-info-name (vop-info vop))))
+  (load-tn nil :type (or tn null))
+  ;; on CISC microprocessors, a representation of the memory load
+  ;; (or store) based on this TN-REF (for which its TN is the base register)
+  ;; and not directly an operand from/into which the TN-REF flows.
+  (memory-access)
+  ;; The type of the LVAR the TN of this TN-REF is used for.
+  (type nil :type (or ctype null)))
 
 ;;; A TEMPLATE object represents a particular IR2 coding strategy for
 ;;; a known function.
-(def!struct (template (:constructor nil)
-                      #-sb-xc-host (:pure t))
+(defstruct (template (:constructor nil)
+                     (:copier nil)
+                     #-sb-xc-host
+                     (:pure t))
   ;; the symbol name of this VOP. This is used when printing the VOP
   ;; and is also used to provide a handle for definition and
   ;; translation.
@@ -590,20 +641,7 @@
   ;; the number of trailing arguments to VOP or %PRIMITIVE that we
   ;; bundle into a list and pass into the emit function. This provides
   ;; a way to pass uninterpreted stuff directly to the code generator.
-  (info-arg-count 0 :type index)
-  ;; a function that emits the VOPs for this template. Arguments:
-  ;;  1] Node for source context.
-  ;;  2] IR2-BLOCK that we place the VOP in.
-  ;;  3] This structure.
-  ;;  4] Head of argument TN-REF list.
-  ;;  5] Head of result TN-REF list.
-  ;;  6] If INFO-ARG-COUNT is non-zero, then a list of the magic
-  ;;     arguments.
-  ;;
-  ;; Two values are returned: the first and last VOP emitted. This vop
-  ;; sequence must be linked into the VOP Next/Prev chain for the
-  ;; block. At least one VOP is always emitted.
-  (emit-function (missing-arg) :type function))
+  (info-arg-count 0 :type index))
 (defprinter (template)
   name
   arg-types
@@ -618,13 +656,7 @@
 ;;; A VOP-INFO object holds the constant information for a given
 ;;; virtual operation. We include TEMPLATE so that functions with a
 ;;; direct VOP equivalent can be translated easily.
-(def!struct (vop-info
-             (:include template)
-             (:make-load-form-fun ignore-it))
-  ;; side effects of this VOP and side effects that affect the value
-  ;; of this VOP
-  (effects (missing-arg) :type attributes)
-  (affected (missing-arg) :type attributes)
+(defstruct (vop-info (:include template) (:copier nil))
   ;; If true, causes special casing of TNs live after this VOP that
   ;; aren't results:
   ;; -- If T, all such TNs that are allocated in a SC with a defined
@@ -673,10 +705,6 @@
   ;; operand SC restriction.
   (arg-load-scs nil :type list)
   (result-load-scs nil :type list)
-  ;; if true, a function that is called with the VOP to do operand
-  ;; targeting. This is done by modifying the TN-REF-TARGET slots in
-  ;; the TN-REFS so that they point to other TN-REFS in the same VOP.
-  (target-fun nil :type (or null function))
   ;; a function that emits assembly code for a use of this VOP when it
   ;; is called with the VOP structure. This is null if this VOP has no
   ;; specified generator (i.e. if it exists only to be inherited by
@@ -690,18 +718,39 @@
   ;; counts as one, and all the more args/results together count as 1.
   (num-args 0 :type index)
   (num-results 0 :type index)
-  ;; a vector of the temporaries the vop needs. See EMIT-GENERIC-VOP
+  ;; a vector of the temporaries the vop needs. See EMIT-VOP
   ;; in vmdef for information on how the temps are encoded.
-  (temps nil :type (or null (specializable-vector (unsigned-byte 16))))
+  (temps nil :type (or null (simple-array (unsigned-byte 16) 1)))
   ;; the order all the refs for this vop should be put in. Each
   ;; operand is assigned a number in the following ordering: args,
   ;; more-args, results, more-results, temps. This vector represents
   ;; the order the operands should be put into in the next-ref link.
-  (ref-ordering nil :type (or null (specializable-vector (unsigned-byte 8))))
+  (ref-ordering nil :type (or null (simple-array (unsigned-byte 8) 1)))
   ;; a vector of the various targets that should be done. Each element
   ;; encodes the source ref (shifted 8, it is also encoded in
   ;; MAX-VOP-TN-REFS) and the dest ref index.
-  (targets nil :type (or null (specializable-vector (unsigned-byte 16)))))
+  (targets nil :type (or null (simple-array (unsigned-byte 16) 1)))
+  (optimizer nil :type (or null function))
+  (optional-results nil :type list)
+  move-vop-p)
+(declaim (inline vop-name))
+(defun vop-name (vop)
+  (declare (type vop vop))
+  (vop-info-name (vop-info vop)))
+
+;; These printers follow the definition of VOP-INFO because they
+;; want to inline VOP-INFO-NAME, and it's less code to move them here
+;; than to move the defstructs of VOP-INFO and TEMPLATE.
+(defprinter (vop)
+  (info :prin1 (vop-info-name info))
+  args
+  results
+  (codegen-info :test codegen-info))
+(defprinter (tn-ref)
+  tn
+  write-p
+  (vop :test vop :prin1 (vop-name vop)))
+
 
 ;;;; SBs and SCs
 
@@ -744,22 +793,32 @@
 
 ;;; The SB structure represents the global information associated with
 ;;; a storage base.
-(def!struct (sb (:make-load-form-fun just-dump-it-normally))
+(defstruct (storage-base (:copier nil) (:conc-name sb-))
   ;; name, for printing and reference
-  (name nil :type symbol)
+  (name nil :type symbol :read-only t)
   ;; the kind of storage base (which determines the packing
   ;; algorithm)
-  (kind :non-packed :type (member :finite :unbounded :non-packed))
+  (kind :non-packed :type (member :finite :unbounded :non-packed) :read-only t)
   ;; the number of elements in the SB. If finite, this is the total
   ;; size. If unbounded, this is the size that the SB is initially
   ;; allocated at.
-  (size 0 :type index))
-(defprinter (sb)
-  name)
+  (size 0 :type index :read-only t))
 
 ;;; A FINITE-SB holds information needed by the packing algorithm for
 ;;; finite SBs.
-(def!struct (finite-sb (:include sb))
+(defstruct (finite-sb-template (:include storage-base) (:copier nil)
+                               (:predicate nil) (:conc-name finite-sb-))
+  ;; for indirecting access of read/write slots of this SB into
+  ;; a proxy object specific to one compiler invocation.
+  (index 0 :type (mod #+sb-xc-host 8 ; arbitrary limit
+                      #-sb-xc-host #.(length *finite-sbs*)) :read-only t)
+  ;; the minimum number of location by which to grow this SB
+  ;; if it is :unbounded
+  (size-increment 1 :type index :read-only t)
+  ;; current-size must always be a multiple of this. It is assumed
+  ;; to be a power of two.
+  (size-alignment 1 :type index :read-only t))
+(defstruct (finite-sb (:copier nil) (:predicate nil) (:conc-name fsb-))
   ;; the number of locations currently allocated in this SB
   (current-size 0 :type index)
   ;; the last location packed in, used by pack to scatter TNs to
@@ -777,7 +836,6 @@
   ;; is set, then the location is in use somewhere in the block, and
   ;; thus has a conflict for always-live TNs.
   (always-live '#() :type simple-vector)
-  (always-live-count '#() :type simple-vector)
   ;; a vector containing the TN currently live in each location in the
   ;; SB, or NIL if the location is unused. This is used during load-tn pack.
   (live-tns '#() :type simple-vector)
@@ -785,21 +843,55 @@
   ;; might not be virgin, and thus must be reinitialized when PACK
   ;; starts. Less then the length of those vectors when not all of the
   ;; length was used on the previously packed component.
-  (last-block-count 0 :type index))
+  (last-block-count 0 :type index)
+  (wired-map 0 :type sb-vm:finite-sc-offset-map))
+(declaim (freeze-type storage-base finite-sb-template finite-sb))
 
-;;; the SC structure holds the storage base that storage is allocated
+;;; Give this a toplevel value so that it can be declaimed ALWAYS-BOUND.
+;;; The compiler will never look at the toplevel value though.
+(defvar *finite-sbs*
+  #-sb-xc-host
+  (make-array #.(count :non-packed *backend-sbs* :key #'sb-kind :test #'neq)
+              :initial-element (make-unbound-marker)))
+#-sb-xc-host
+(progn
+  (declaim (type (simple-vector #.(length *finite-sbs*)) *finite-sbs*)
+           (always-bound *finite-sbs*))
+  (eval-when (:compile-toplevel :load-toplevel :execute)
+    (setf (info :variable :wired-tls '*finite-sbs*) :always-thread-local)))
+
+;;; All these are SETFable
+(defmacro finite-sb-current-size (sb)
+  `(fsb-current-size (svref *finite-sbs* (finite-sb-index ,sb))))
+(defmacro finite-sb-last-offset (sb)
+  `(fsb-last-offset (svref *finite-sbs* (finite-sb-index ,sb))))
+(defmacro finite-sb-conflicts (sb)
+  `(fsb-conflicts (svref *finite-sbs* (finite-sb-index ,sb))))
+(defmacro finite-sb-always-live (sb)
+  `(fsb-always-live (svref *finite-sbs* (finite-sb-index ,sb))))
+(defmacro finite-sb-wired-map (sb)
+  `(fsb-wired-map (svref *finite-sbs* (finite-sb-index ,sb))))
+
+(defmacro finite-sb-live-tns (sb)
+  `(fsb-live-tns (svref *finite-sbs* (finite-sb-index ,sb))))
+(defmacro finite-sb-last-block-count (sb)
+  `(fsb-last-block-count (svref *finite-sbs* (finite-sb-index ,sb))))
+
+;;; the STORAGE-CLASS structure holds the storage base that storage is allocated
 ;;; in and information used to select locations within the SB
-(def!struct (sc (:copier nil))
+(defstruct (storage-class (:conc-name "SC-") (:copier nil) (:predicate nil))
   ;; name, for printing and reference
   (name nil :type symbol)
   ;; the number used to index SC cost vectors
   (number 0 :type sc-number)
   ;; the storage base that this SC allocates storage from
-  (sb nil :type (or sb null))
+  (sb nil :type (or storage-base null))
   ;; the size of elements in this SC, in units of locations in the SB
   (element-size 0 :type index)
-  ;; if our SB is finite, a list of the locations in this SC
-  (locations nil :type list)
+  ;; if our SB is finite, a vector of the locations in this SC
+  (locations (missing-arg) :type sc-locations :read-only t)
+  ;; information for the assembler when moving to/from the locations
+  (operand-size nil :type (or null keyword) :read-only t)
   ;; a list of the alternate (save) SCs for this SC
   (alternate-scs nil :type list)
   ;; a list of the constant SCs that can me moved into this SC
@@ -812,9 +904,9 @@
   ;; of the corresponding move functions. If loading is impossible,
   ;; then the entries are NIL. LOAD-COSTS is initialized to have a 0
   ;; for this SC.
-  (move-funs (make-array sc-number-limit :initial-element nil)
+  (move-funs (make-array sb-vm:sc-number-limit :initial-element nil)
              :type sc-vector)
-  (load-costs (make-array sc-number-limit :initial-element nil)
+  (load-costs (make-array sb-vm:sc-number-limit :initial-element nil)
               :type sc-vector)
   ;; a vector mapping from SC numbers to possibly
   ;; representation-specific move and coerce VOPs. Each entry is a
@@ -834,34 +926,35 @@
   ;; TNs wired in the standard argument registers, since there may
   ;; already be live TNs wired in those locations holding the values
   ;; that we are setting up for unknown-values return.
-  (move-vops (make-array sc-number-limit :initial-element nil)
+  (move-vops (make-array sb-vm:sc-number-limit :initial-element nil)
              :type sc-vector)
   ;; the costs corresponding to the MOVE-VOPS. Separate because this
   ;; info is needed at meta-compile time, while the MOVE-VOPs don't
   ;; exist till load time. If no move is defined, then the entry is
   ;; NIL.
-  (move-costs (make-array sc-number-limit :initial-element nil)
+  (move-costs (make-array sb-vm:sc-number-limit :initial-element nil)
               :type sc-vector)
   ;; similar to Move-VOPs, except that we only ever use the entries
   ;; for this SC and its alternates, since we never combine complex
   ;; representation conversion with argument passing.
-  (move-arg-vops (make-array sc-number-limit :initial-element nil)
+  (move-arg-vops (make-array sb-vm:sc-number-limit :initial-element nil)
                  :type sc-vector)
   ;; true if this SC or one of its alternates in in the NUMBER-STACK SB.
   (number-stack-p nil :type boolean)
   ;; alignment restriction. The offset must be an even multiple of this.
+  ;; this must be a power of two.
   (alignment 1 :type (and index (integer 1)))
-  ;; a list of locations that we avoid packing in during normal
+  ;; a vector of locations that we avoid packing in during normal
   ;; register allocation to ensure that these locations will be free
   ;; for operand loading. This prevents load-TN packing from thrashing
   ;; by spilling a lot.
-  (reserve-locations nil :type list))
-(defprinter (sc)
+  (reserve-locations (missing-arg) :type sc-locations :read-only t))
+(defprinter (storage-class :conc-name "SC-")
   name)
 
 ;;;; TNs
 
-(def!struct (tn (:include sset-element)
+(defstruct (tn (:include sset-element)
                (:constructor make-random-tn)
                (:constructor make-tn (number kind primitive-type sc))
                (:copier nil))
@@ -889,7 +982,7 @@
   ;;   :SAVE-ONCE
   ;;    A TN used for saving a :NORMAL TN across function calls. The
   ;;    lifetime information slots are unitialized: get the original
-  ;;    TN our of the SAVE-TN slot and use it for conflicts. SAVE-ONCE
+  ;;    TN out of the SAVE-TN slot and use it for conflicts. SAVE-ONCE
   ;;    is like :SAVE, except that it is only save once at the single
   ;;    writer of the original TN.
   ;;
@@ -919,13 +1012,23 @@
   ;;    as :NORMAL, but then at the end merges the conflict info into
   ;;    the original TN and replaces all uses of the alias with the
   ;;    original TN. SAVE-TN holds the aliased TN.
+  ;;   :UNUSED
+  ;;    Unused result
   (kind (missing-arg)
         :type (member :normal :environment :debug-environment
-                      :save :save-once :specified-save :load :constant
-                      :component :alias))
+                      :save :save-once  :load :constant
+                      :component :alias :unused
+                      #-fp-and-pc-standard-save :specified-save
+                      :arg-pass))
   ;; the primitive-type for this TN's value. Null in restricted or
   ;; wired TNs.
   (primitive-type nil :type (or primitive-type null))
+  ;; The type of the LVAR this TN is used for.
+  ;; Meaningless after copy-propagate or other optimizations are performed
+  ;; and TN-REF-TYPE should be used instead.
+  ;; The only purpose of this slot is for REFERENCE-TN to populate
+  ;; TN-REF-TYPE.
+  (type nil :type (or ctype null))
   ;; If this TN represents a variable or constant, then this is the
   ;; corresponding LEAF.
   (leaf nil :type (or leaf null))
@@ -955,7 +1058,14 @@
   ;; This list is sorted by block number (i.e. reverse DFO), allowing
   ;; the intersection between the lifetimes for two global TNs to be
   ;; easily found. If null, then this TN is a local TN.
-  (global-conflicts nil :type (or global-conflicts null))
+  ;; KLUDGE: The defstructs for TN and GLOBAL-CONFLICTS are mutually
+  ;; referential, and absent a block-compilation feature, one or the other
+  ;; of the structures can't inline type checks for its referent.
+  ;; Stating this type as (OR NULL GLOBAL-CONFLICTS) instead of the reverse
+  ;; avoids a forward-reference that would crash cold-init.
+  ;; But since TYPEP short-circuits, the constructor is happy with NIL.
+  ;; [See the comments in LAYOUT-OF for further detail]
+  (global-conflicts nil :type (or null global-conflicts))
   ;; During lifetime analysis, this is used as a pointer into the
   ;; conflicts chain, for scanning through blocks in reverse DFO.
   (current-conflict nil)
@@ -965,26 +1075,35 @@
   (save-tn nil :type (or tn null))
   ;; After pack, the SC we packed into. Beforehand, the SC we want to
   ;; pack into, or null if we don't know.
-  (sc nil :type (or sc null))
+  (sc nil :type (or storage-class null))
   ;; the offset within the SB that this TN is packed into. This is what
   ;; indicates that the TN is packed
-  (offset nil :type (or index null))
+  ;; The integer bound ensures that TN-BYTE-OFFSET remains a fixnum.
+  ;; The default stack size is 2MB, so this is plenty big.
+  (offset nil :type (or null
+                        (unsigned-byte #.(- sb-vm:n-positive-fixnum-bits
+                                            sb-vm:word-shift))))
   ;; some kind of info about how important this TN is
   (cost 0 :type fixnum)
   ;; If a :ENVIRONMENT or :DEBUG-ENVIRONMENT TN, this is the
   ;; physical environment that the TN is live throughout.
   (physenv nil :type (or physenv null))
-  ;; The depth of the deepest loop that this TN is used in.
-  (loop-depth 0 :type fixnum))
-(def!method print-object ((tn tn) stream)
-  (print-unreadable-object (tn stream :type t)
+  ;; Used by pack-iterative
+  (vertex nil))
+
+(declaim (freeze-type tn))
+(defmethod print-object ((tn tn) stream)
+  (cond ((not (boundp '*compilation*))
+         (print-unreadable-object (tn stream :type t :identity t)))
+        (t
+         (print-unreadable-object (tn stream :type t)
     ;; KLUDGE: The distinction between PRINT-TN and PRINT-OBJECT on TN is
     ;; not very mnemonic. -- WHN 20000124
-    (print-tn-guts tn stream)))
+           (print-tn-guts tn stream)))))
 
 ;;; The GLOBAL-CONFLICTS structure represents the conflicts for global
 ;;; TNs. Each global TN has a list of these structures, one for each
-;;; block that it is live in. In addition to repsenting the result of
+;;; block that it is live in. In addition to representing the result of
 ;;; lifetime analysis, the global conflicts structure is used during
 ;;; lifetime analysis to represent the set of TNs live at the start of
 ;;; the IR2 block.
@@ -993,8 +1112,7 @@
             (:copier nil))
   ;; the IR2-BLOCK that this structure represents the conflicts for
   (block (missing-arg) :type ir2-block)
-  ;; thread running through all the GLOBAL-CONFLICTSs for BLOCK. This
-  ;; thread is sorted by TN number
+  ;; thread running through all the GLOBAL-CONFLICTSs for BLOCK.
   (next-blockwise nil :type (or global-conflicts null))
   ;; the way that TN is used by BLOCK
   ;;
