@@ -9,52 +9,21 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
-
-;;; Make an environment-live stack TN for saving the SP for NLX entry.
-(!def-vm-support-routine make-nlx-sp-tn (env)
-  (physenv-live-tn
-   (make-representation-tn *fixnum-primitive-type* any-reg-sc-number)
-   env))
+(in-package "SB-VM")
 
 ;;; Make a TN for the argument count passing location for a non-local entry.
-(!def-vm-support-routine make-nlx-entry-arg-start-location ()
-  (make-wired-tn *fixnum-primitive-type* any-reg-sc-number rbx-offset))
+(defun make-nlx-entry-arg-start-location ()
+    (make-wired-tn *fixnum-primitive-type* any-reg-sc-number rbx-offset))
 
-(defun catch-block-ea (tn)
+(defun catch-block-ea (tn &optional (offset 0))
   (aver (sc-is tn catch-block))
-  (make-ea :qword :base rbp-tn
-           :disp (frame-byte-offset (+ -1 (tn-offset tn) catch-block-size))))
+  (ea (frame-byte-offset (- (+ -1 (tn-offset tn) catch-block-size) offset)) rbp-tn))
 
+(defun unwind-block-ea (tn &optional (offset 0))
+  (aver (sc-is tn unwind-block))
+  (ea (frame-byte-offset (- (+ -1 (tn-offset tn) unwind-block-size offset) offset)) rbp-tn))
 
 ;;;; Save and restore dynamic environment.
-;;;;
-;;;; These VOPs are used in the reentered function to restore the
-;;;; appropriate dynamic environment. Currently we only save the
-;;;; Current-Catch and the alien stack pointer. (Before sbcl-0.7.0,
-;;;; when there were IR1 and byte interpreters, we had to save
-;;;; the interpreter "eval stack" too.)
-;;;;
-;;;; We don't need to save/restore the current UNWIND-PROTECT, since
-;;;; UNWIND-PROTECTs are implicitly processed during unwinding.
-;;;;
-;;;; We don't need to save the BSP, because that is handled automatically.
-
-(define-vop (save-dynamic-state)
-  (:results (catch :scs (descriptor-reg))
-            (alien-stack :scs (descriptor-reg)))
-  (:generator 13
-    (load-tl-symbol-value catch *current-catch-block*)
-    (load-tl-symbol-value alien-stack *alien-stack*)))
-
-(define-vop (restore-dynamic-state)
-  (:args (catch :scs (descriptor-reg))
-         (alien-stack :scs (descriptor-reg)))
-  #!+sb-thread (:temporary (:sc unsigned-reg) temp)
-  (:generator 10
-    (store-tl-symbol-value catch *current-catch-block* temp)
-    (store-tl-symbol-value alien-stack *alien-stack* temp)))
-
 (define-vop (current-stack-pointer)
   (:results (res :scs (any-reg control-stack)))
   (:generator 1
@@ -69,18 +38,39 @@
 
 ;;; Compute the address of the catch block from its TN, then store into the
 ;;; block the current Fp, Env, Unwind-Protect, and the entry PC.
+#+sb-thread
+(progn
+  ;; MOVAPD instruction faults if not properly aligned
+  (assert (evenp (/ (info :variable :wired-tls '*binding-stack-pointer*) n-word-bytes)))
+  (assert (= (- (info :variable :wired-tls '*current-catch-block*)
+                (info :variable :wired-tls '*binding-stack-pointer*))
+             n-word-bytes))
+  (assert (= (- unwind-block-current-catch-slot unwind-block-bsp-slot) 1)))
+
 (define-vop (make-unwind-block)
   (:args (tn))
   (:info entry-label)
   (:temporary (:sc unsigned-reg) temp)
+  #+sb-thread
+  (:temporary (:sc complex-double-reg) xmm-temp)
   (:results (block :scs (any-reg)))
   (:generator 22
-    (inst lea block (catch-block-ea tn))
+    (inst lea block (unwind-block-ea tn))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
-    (storew temp block unwind-block-current-uwp-slot)
-    (storew rbp-tn block unwind-block-current-cont-slot)
-    (inst lea temp (make-fixup nil :code-object entry-label))
-    (storew temp block catch-block-entry-pc-slot)))
+    (storew temp block unwind-block-uwp-slot)
+    (storew rbp-tn block unwind-block-cfp-slot)
+    (inst lea temp (rip-relative-ea entry-label))
+    (storew temp block unwind-block-entry-pc-slot)
+    #+sb-thread
+    (let ((bsp (info :variable :wired-tls '*binding-stack-pointer*)))
+      (inst movapd xmm-temp (thread-tls-ea bsp))
+      (inst movupd (ea (* unwind-block-bsp-slot n-word-bytes) block) xmm-temp))
+    #-sb-thread
+    (progn
+      (load-binding-stack-pointer temp)
+      (storew temp block unwind-block-bsp-slot)
+      (load-tl-symbol-value temp *current-catch-block*)
+      (storew temp block unwind-block-current-catch-slot))))
 
 ;;; like MAKE-UNWIND-BLOCK, except that we also store in the specified
 ;;; tag, and link the block into the CURRENT-CATCH list
@@ -90,44 +80,55 @@
   (:info entry-label)
   (:results (block :scs (any-reg)))
   (:temporary (:sc descriptor-reg) temp)
+  #+sb-thread
+  (:temporary (:sc complex-double-reg) xmm-temp)
   (:generator 44
     (inst lea block (catch-block-ea tn))
     (load-tl-symbol-value temp *current-unwind-protect-block*)
-    (storew temp block  unwind-block-current-uwp-slot)
-    (storew rbp-tn block  unwind-block-current-cont-slot)
-    (inst lea temp (make-fixup nil :code-object entry-label))
+    (storew temp block catch-block-uwp-slot)
+    (storew rbp-tn block catch-block-cfp-slot)
+    (inst lea temp (rip-relative-ea entry-label))
     (storew temp block catch-block-entry-pc-slot)
     (storew tag block catch-block-tag-slot)
-    (load-tl-symbol-value temp *current-catch-block*)
-    (storew temp block catch-block-previous-catch-slot)
-    (store-tl-symbol-value block *current-catch-block* temp)))
+    #+sb-thread
+    (let ((bsp #1=(info :variable :wired-tls '*binding-stack-pointer*)))
+      #.(assert (and (= (- (info :variable :wired-tls '*current-catch-block*) #1#) n-word-bytes)
+                     (= (- catch-block-previous-catch-slot catch-block-bsp-slot) 1)))
+      (inst movapd xmm-temp (thread-tls-ea bsp))
+      (inst movupd (ea (* catch-block-bsp-slot n-word-bytes) block) xmm-temp)
+      (store-tl-symbol-value block *current-catch-block*))
+    #-sb-thread
+    (progn
+      (load-tl-symbol-value temp *current-catch-block*)
+      (storew temp block catch-block-previous-catch-slot)
+      (store-tl-symbol-value block *current-catch-block*)
+      (load-binding-stack-pointer temp)
+      (storew temp block catch-block-bsp-slot))))
 
-;;; Just set the current unwind-protect to TN's address. This instantiates an
+;;; Just set the current unwind-protect to UWP. This instantiates an
 ;;; unwind block as an unwind-protect.
 (define-vop (set-unwind-protect)
-  (:args (tn))
-  (:temporary (:sc unsigned-reg) new-uwp #!+sb-thread tls)
+  (:args (uwp :scs (any-reg)))
   (:generator 7
-    (inst lea new-uwp (catch-block-ea tn))
-    (store-tl-symbol-value new-uwp *current-unwind-protect-block* tls)))
+    (store-tl-symbol-value uwp *current-unwind-protect-block*)))
 
-(define-vop (unlink-catch-block)
-  (:temporary (:sc unsigned-reg) #!+sb-thread tls block)
+(define-vop (%catch-breakup)
+  (:args (current-block))
+  (:temporary (:sc unsigned-reg) block)
   (:policy :fast-safe)
-  (:translate %catch-breakup)
   (:generator 17
-    (load-tl-symbol-value block *current-catch-block*)
-    (loadw block block catch-block-previous-catch-slot)
-    (store-tl-symbol-value block *current-catch-block* tls)))
+    (inst mov block (catch-block-ea current-block
+                                    catch-block-previous-catch-slot))
+    (store-tl-symbol-value block *current-catch-block*)))
 
-(define-vop (unlink-unwind-protect)
-    (:temporary (:sc unsigned-reg) block #!+sb-thread tls)
+(define-vop (%unwind-protect-breakup)
+  (:args (current-block))
+  (:temporary (:sc unsigned-reg) block)
   (:policy :fast-safe)
-  (:translate %unwind-protect-breakup)
   (:generator 17
-    (load-tl-symbol-value block *current-unwind-protect-block*)
-    (loadw block block unwind-block-current-uwp-slot)
-    (store-tl-symbol-value block *current-unwind-protect-block* tls)))
+     (inst mov block (unwind-block-ea current-block
+                                      unwind-block-uwp-slot))
+     (store-tl-symbol-value block *current-unwind-protect-block*)))
 
 ;;;; NLX entry VOPs
 (define-vop (nlx-entry)
@@ -148,7 +149,8 @@
           ((= nvals 1)
            (let ((no-values (gen-label)))
              (inst mov (tn-ref-tn values) nil-value)
-             (inst jrcxz no-values)
+             (inst test rcx-tn rcx-tn)
+             (inst jmp :z no-values)
              (loadw (tn-ref-tn values) start -1)
              (emit-label no-values)))
           (t
@@ -175,7 +177,7 @@
                     (inst mov tn move-temp)))))
              (let ((defaulting-done (gen-label)))
                (emit-label defaulting-done)
-               (assemble (*elsewhere*)
+               (assemble (:elsewhere)
                  (dolist (default (defaults))
                    (emit-label (car default))
                    (when (cddr default)
@@ -185,106 +187,116 @@
     (inst mov rsp-tn sp)))
 
 (define-vop (nlx-entry-multiple)
-  (:args (top)
-         (source)
+  (:args (top :target result
+              :scs (any-reg))
+         (source :to :save)
          (count :target rcx))
-  ;; Again, no SC restrictions for the args, 'cause the loading would
-  ;; happen before the entry label.
   (:info label)
+  (:before-load
+    (emit-label label)
+    (note-this-location vop :non-local-entry))
   (:temporary (:sc unsigned-reg :offset rcx-offset :from (:argument 2)) rcx)
-  (:temporary (:sc unsigned-reg :offset rsi-offset) rsi)
-  (:temporary (:sc unsigned-reg :offset rdi-offset) rdi)
-  (:results (result :scs (any-reg) :from (:argument 0))
+  (:temporary (:sc unsigned-reg) loop-index temp)
+  (:results (result :scs (any-reg))
             (num :scs (any-reg control-stack)))
   (:save-p :force-to-stack)
+  (:args-var top-tn-ref)
   (:vop-var vop)
   (:generator 30
-    (emit-label label)
-    (note-this-location vop :non-local-entry)
-
-    (inst lea rsi (make-ea :qword :base source :disp (- n-word-bytes)))
     ;; The 'top' arg contains the %esp value saved at the time the
     ;; catch block was created and points to where the thrown values
     ;; should sit.
-    (move rdi top)
-    (move result rdi)
+    (if (eq (tn-kind result) :unused)
+        (setf result top)
+        (move result top))
 
-    (inst sub rdi n-word-bytes)
-    (move rcx count)                    ; fixnum words == bytes
-    (move num rcx)
-    (inst shr rcx word-shift)           ; word count for <rep movs>
-    ;; If we got zero, we be done.
-    (inst jrcxz DONE)
-    ;; Copy them down.
-    (inst std)
-    (inst rep)
-    (inst movs :qword)
-    (inst cld)
+    (unless (eq (tn-kind num) :unused)
+      (move num count))
+    (move rcx count)
+    (zeroize loop-index)
+    (inst test rcx rcx)
+    (inst jmp :z DONE)
+    LOOP
+    (inst sub loop-index n-word-bytes)
+    (inst mov temp (ea source loop-index))
+    (inst mov (ea result loop-index) temp)
+
+    (inst sub rcx (fixnumize 1))
+    (inst jmp :nz LOOP)
     DONE
     ;; Reset the CSP at last moved arg.
-    (inst lea rsp-tn (make-ea :qword :base rdi :disp n-word-bytes))))
+    (inst lea rsp-tn (ea result loop-index))))
 
 
 ;;; This VOP is just to force the TNs used in the cleanup onto the stack.
 (define-vop (uwp-entry)
   (:info label)
   (:save-p :force-to-stack)
-  (:results (block) (start) (count))
-  (:ignore block start count)
   (:vop-var vop)
   (:generator 0
     (emit-label label)
     (note-this-location vop :non-local-entry)))
 
+(define-vop (uwp-entry-block)
+  (:info label)
+  (:save-p :force-to-stack)
+  (:results (block))
+  (:vop-var vop)
+  (:generator 0
+    (emit-label label)
+    (note-this-location vop :non-local-entry)
+    ;; Get the saved block in UNWIND
+    (inst mov block (ea (* 3 n-word-bytes) rsp-tn))))
+
 (define-vop (unwind-to-frame-and-call)
     (:args (ofp :scs (descriptor-reg))
            (uwp :scs (descriptor-reg))
-           (function :scs (descriptor-reg)))
-  (:arg-types system-area-pointer system-area-pointer t)
+           (function :scs (descriptor-reg) :to :load :target saved-function)
+           (bsp :scs (any-reg descriptor-reg))
+           (catch-block :scs (any-reg descriptor-reg)))
+  (:arg-types system-area-pointer system-area-pointer t t t)
   (:temporary (:sc sap-reg) temp)
+  (:temporary (:sc descriptor-reg :offset rbx-offset) saved-function)
   (:temporary (:sc unsigned-reg :offset rax-offset) block)
+  (:temporary (:sc unsigned-reg :offset r11-offset) extra-temp-reg)
+  (:vop-var vop)
   (:generator 22
     ;; Store the function into a non-stack location, since we'll be
     ;; unwinding the stack and destroying register contents before we
-    ;; use it.
-    (store-tl-symbol-value function
-                           *unwind-to-frame-function*
-                           temp)
+    ;; use it.  It turns out that RBX is preserved as part of the
+    ;; normal multiple-value handling of an unwind, so use that.
+    (move saved-function function)
 
     ;; Allocate space for magic UWP block.
     (inst sub rsp-tn (* unwind-block-size n-word-bytes))
     ;; Set up magic catch / UWP block.
     (move block rsp-tn)
     (loadw temp uwp sap-pointer-slot other-pointer-lowtag)
-    (storew temp block unwind-block-current-uwp-slot)
+    (storew temp block unwind-block-uwp-slot)
     (loadw temp ofp sap-pointer-slot other-pointer-lowtag)
-    (storew temp block unwind-block-current-cont-slot)
+    (storew temp block unwind-block-cfp-slot)
 
-    (inst lea temp-reg-tn (make-fixup nil :code-object entry-label))
-    (storew temp-reg-tn
-            block
-            catch-block-entry-pc-slot)
+    (inst lea extra-temp-reg (rip-relative-ea entry-label))
+    (storew extra-temp-reg block unwind-block-entry-pc-slot)
+    (storew bsp block unwind-block-bsp-slot)
+    (storew catch-block block unwind-block-current-catch-slot)
 
     ;; Run any required UWPs.
-    (inst lea temp-reg-tn (make-fixup 'unwind :assembly-routine))
-    (inst jmp temp-reg-tn)
+    (invoke-asm-routine 'jmp 'unwind vop)
     ENTRY-LABEL
 
-    ;; Load function from symbol
-    (load-tl-symbol-value block *unwind-to-frame-function*)
+    ;; Move our saved function to where we want it now.
+    (move block saved-function)
 
     ;; No parameters
     (zeroize rcx-tn)
 
     ;; Clear the stack
-    (inst lea rsp-tn
-          (make-ea :qword :base rbp-tn
-                   :disp (* (- sp->fp-offset 3) n-word-bytes)))
+    (inst lea rsp-tn (ea (* (- sp->fp-offset 3) n-word-bytes) rbp-tn))
 
     ;; Push the return-pc so it looks like we just called.
     (pushw rbp-tn (frame-word-offset return-pc-save-offset))
 
     ;; Call it
-    (inst jmp (make-ea :qword :base block
-                       :disp (- (* closure-fun-slot n-word-bytes)
-                                fun-pointer-lowtag)))))
+    (inst jmp (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag)
+                  block))))
