@@ -9,7 +9,7 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(in-package "SB!VM")
+(in-package "SB-VM")
 
 ;;;; data object ref/set stuff
 
@@ -21,14 +21,16 @@
   (:generator 1
     (loadw result object offset lowtag)))
 
+;;; This vop is magical in (at least) 2 ways:
+;;; 1. it is always selected by name (in ir2-convert-setter or -setfer)
+;;; 2. the ir2 converter makes it return a value despite absence of :results
 (define-vop (set-slot)
   (:args (object :scs (descriptor-reg))
          (value :scs (descriptor-reg any-reg immediate)))
   (:info name offset lowtag)
   (:ignore name)
-  (:results)
   (:generator 1
-     (storew (encode-value-if-immediate value) object offset lowtag)))
+    (storew (encode-value-if-immediate value) object offset lowtag)))
 
 (define-vop (compare-and-swap-slot)
   (:args (object :scs (descriptor-reg) :to :eval)
@@ -55,7 +57,7 @@
          (old :scs (descriptor-reg any-reg) :target eax)
          (new :scs (descriptor-reg any-reg)))
   (:temporary (:sc descriptor-reg :offset eax-offset) eax)
-  #!+sb-thread
+  #+sb-thread
   (:temporary (:sc descriptor-reg) tls)
   (:results (result :scs (descriptor-reg any-reg)))
   (:policy :fast-safe)
@@ -67,11 +69,12 @@
     (let ((unbound (generate-error-code vop 'unbound-symbol-error symbol))
           (check (gen-label)))
       (move eax old)
-      #!+sb-thread
+      #+sb-thread
       (progn
         (loadw tls symbol symbol-tls-index-slot other-pointer-lowtag)
         ;; Thread-local area, no LOCK needed.
-        (inst cmpxchg (make-ea :dword :base tls) new :fs)
+        (with-tls-ea (EA :base tls :base-already-live-p t)
+          (inst cmpxchg EA new :maybe-fs))
         (inst cmp eax no-tls-value-marker-widetag)
         (inst jmp :ne check)
         (move eax old))
@@ -90,11 +93,11 @@
 (define-vop (fast-symbol-global-value cell-ref)
   (:variant symbol-value-slot other-pointer-lowtag)
   (:policy :fast)
-  (:translate symbol-global-value))
+  (:translate sym-global-val))
 
 (define-vop (symbol-global-value)
   (:policy :fast-safe)
-  (:translate symbol-global-value)
+  (:translate sym-global-val)
   (:args (object :scs (descriptor-reg) :to (:result 1)))
   (:results (value :scs (descriptor-reg any-reg)))
   (:vop-var vop)
@@ -105,7 +108,7 @@
       (inst cmp value unbound-marker-widetag)
       (inst jmp :e err-lab))))
 
-#!+sb-thread
+#+sb-thread
 (progn
   (define-vop (set)
     (:args (symbol :scs (descriptor-reg))
@@ -115,9 +118,10 @@
       (let ((global-val (gen-label))
             (done (gen-label)))
         (loadw tls symbol symbol-tls-index-slot other-pointer-lowtag)
-        (inst cmp (make-ea :dword :base tls) no-tls-value-marker-widetag :fs)
-        (inst jmp :z global-val)
-        (inst mov (make-ea :dword :base tls) value :fs)
+        (with-tls-ea (EA :base tls :base-already-live-p t)
+          (inst cmp EA no-tls-value-marker-widetag :maybe-fs)
+          (inst jmp :z global-val)
+          (inst mov EA value :maybe-fs))
         (inst jmp done)
         (emit-label global-val)
         (storew value symbol symbol-value-slot other-pointer-lowtag)
@@ -126,7 +130,7 @@
   ;; With Symbol-Value, we check that the value isn't the trap object. So
   ;; Symbol-Value of NIL is NIL.
   (define-vop (symbol-value)
-    (:translate symbol-value)
+    (:translate symeval)
     (:policy :fast-safe)
     (:args (object :scs (descriptor-reg) :to (:result 1)))
     (:results (value :scs (descriptor-reg any-reg)))
@@ -137,7 +141,8 @@
              (err-lab (generate-error-code vop 'unbound-symbol-error object))
              (ret-lab (gen-label)))
         (loadw value object symbol-tls-index-slot other-pointer-lowtag)
-        (inst mov value (make-ea :dword :base value) :fs)
+        (with-tls-ea (EA :base value :base-already-live-p t)
+          (inst mov value EA :maybe-fs))
         (inst cmp value no-tls-value-marker-widetag)
         (inst jmp :ne check-unbound-label)
         (loadw value object symbol-value-slot other-pointer-lowtag)
@@ -153,25 +158,26 @@
     ;; unbound", which is used in the implementation of COPY-SYMBOL.  --
     ;; CSR, 2003-04-22
     (:policy :fast)
-    (:translate symbol-value)
+    (:translate symeval)
     (:generator 8
       (let ((ret-lab (gen-label)))
         (loadw value object symbol-tls-index-slot other-pointer-lowtag)
-        (inst mov value (make-ea :dword :base value) :fs)
+        (with-tls-ea (EA :base value :base-already-live-p t)
+          (inst mov value EA :maybe-fs))
         (inst cmp value no-tls-value-marker-widetag)
         (inst jmp :ne ret-lab)
         (loadw value object symbol-value-slot other-pointer-lowtag)
         (emit-label ret-lab)))))
 
-#!-sb-thread
+#-sb-thread
 (progn
   (define-vop (symbol-value symbol-global-value)
-    (:translate symbol-value))
+    (:translate symeval))
   (define-vop (fast-symbol-value fast-symbol-global-value)
-    (:translate symbol-value))
+    (:translate symeval))
   (define-vop (set %set-symbol-global-value)))
 
-#!+sb-thread
+#+sb-thread
 (define-vop (boundp)
   (:translate boundp)
   (:policy :fast-safe)
@@ -181,21 +187,22 @@
   (:generator 9
     (let ((check-unbound-label (gen-label)))
       (loadw value object symbol-tls-index-slot other-pointer-lowtag)
-      (inst mov value (make-ea :dword :base value) :fs)
+      (with-tls-ea (EA :base value :base-already-live-p t)
+        (inst mov value EA :maybe-fs))
       (inst cmp value no-tls-value-marker-widetag)
       (inst jmp :ne check-unbound-label)
       (loadw value object symbol-value-slot other-pointer-lowtag)
       (emit-label check-unbound-label)
       (inst cmp value unbound-marker-widetag))))
 
-#!-sb-thread
+#-sb-thread
 (define-vop (boundp)
   (:translate boundp)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg)))
   (:conditional :ne)
   (:generator 9
-    (inst cmp (make-ea-for-object-slot object symbol-value-slot
+    (inst cmp (object-slot-ea object symbol-value-slot
                                        other-pointer-lowtag)
           unbound-marker-widetag)))
 
@@ -206,13 +213,15 @@
   (:args (symbol :scs (descriptor-reg)))
   (:results (res :scs (any-reg)))
   (:result-types positive-fixnum)
+  (:args-var args)
   (:generator 2
     ;; The symbol-hash slot of NIL holds NIL because it is also the
-    ;; cdr slot, so we have to strip off the two low bits to make sure
+    ;; car slot, so we have to strip off the two low bits to make sure
     ;; it is a fixnum.  The lowtag selection magic that is required to
     ;; ensure this is explained in the comment in objdef.lisp
     (loadw res symbol symbol-hash-slot other-pointer-lowtag)
-    (inst and res (lognot #b11))))
+    (unless (not-nil-tn-ref-p args)
+      (inst and res (lognot #b11)))))
 
 ;;;; fdefinition (FDEFN) objects
 
@@ -220,6 +229,8 @@
   (:variant fdefn-fun-slot other-pointer-lowtag))
 
 (define-vop (safe-fdefn-fun)
+  (:translate safe-fdefn-fun)
+  (:policy :fast-safe)
   (:args (object :scs (descriptor-reg) :to (:result 1)))
   (:results (value :scs (descriptor-reg any-reg)))
   (:vop-var vop)
@@ -236,17 +247,14 @@
   (:args (function :scs (descriptor-reg) :target result)
          (fdefn :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg) raw)
-  (:temporary (:sc byte-reg) type)
   (:results (result :scs (descriptor-reg)))
   (:generator 38
-    (load-type type function (- fun-pointer-lowtag))
-    (inst lea raw
-          (make-ea-for-object-slot function simple-fun-code-offset
-                                   fun-pointer-lowtag))
-    (inst cmp type simple-fun-header-widetag)
-    (inst jmp :e normal-fn)
-    (inst lea raw (make-fixup "closure_tramp" :foreign))
-    NORMAL-FN
+    (inst mov raw (make-fixup 'closure-tramp :assembly-routine))
+    (inst cmp (make-ea :byte :base function :disp (- fun-pointer-lowtag))
+          simple-fun-widetag)
+    (inst cmov :e raw
+          (make-ea :dword :base function
+                   :disp (- (* simple-fun-self-slot n-word-bytes) fun-pointer-lowtag)))
     (storew function fdefn fdefn-fun-slot other-pointer-lowtag)
     (storew raw fdefn fdefn-raw-addr-slot other-pointer-lowtag)
     (move result function)))
@@ -254,56 +262,56 @@
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
-  (:results (result :scs (descriptor-reg)))
+  (:args (fdefn :scs (descriptor-reg)))
   (:generator 38
     (storew nil-value fdefn fdefn-fun-slot other-pointer-lowtag)
-    (storew (make-fixup "undefined_tramp" :foreign)
-            fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-    (move result fdefn)))
+    (storew (make-fixup 'undefined-tramp :assembly-routine)
+            fdefn fdefn-raw-addr-slot other-pointer-lowtag)))
 
 ;;;; binding and unbinding
 
 ;;; BIND -- Establish VAL as a binding for SYMBOL. Save the old value and
 ;;; the symbol on the binding stack and stuff the new value into the
 ;;; symbol.
+;;; See the "Chapter 9: Specials" of the SBCL Internals Manual.
+;;
+#+sb-thread
+(macrolet
+    ((def (vopname symbol-arg info sc-offset load-tls-index)
+       `(define-vop (,vopname)
+          (:args (val :scs (any-reg descriptor-reg)) ,@symbol-arg)
+          ,@info
+          (:temporary (:sc unsigned-reg ,@sc-offset) tls-index)
+          (:temporary (:sc unsigned-reg) bsp #+win32 temp)
+          (:generator 10 ; cost is irrelevant. we explicitly pick a vop
+           (load-binding-stack-pointer bsp)
+           ,@load-tls-index
+           (inst add bsp (* binding-size n-word-bytes))
+           (store-binding-stack-pointer bsp)
+     ;; with-tls-ea on win32 causes tls-index to be an absolute address
+     ;; which is problematic when UNBIND uses with-tls-ea too.
+           #+win32(move temp tls-index)
+           (with-tls-ea (EA :base tls-index :base-already-live-p t)
+            (inst push EA :maybe-fs)
+            (popw bsp (- binding-value-slot binding-size))
+            (storew #-win32 tls-index
+                    #+win32 temp
+                    bsp (- binding-symbol-slot binding-size))
+            (inst mov EA val :maybe-fs))))))
+  (def bind ; bind a known symbol
+       nil ((:info symbol)) nil
+       ((inst mov tls-index (make-fixup symbol :symbol-tls-index))))
+  (def dynbind ; bind a symbol in a PROGV form
+       ((symbol :scs (descriptor-reg))) nil (:offset eax-offset)
+       ((inst mov tls-index (tls-index-of symbol))
+        (inst test tls-index tls-index)
+        (inst jmp :ne tls-index-valid)
+        (inst mov tls-index symbol)
+        (inst call (make-fixup 'alloc-tls-index :assembly-routine))
+        TLS-INDEX-VALID)))
 
-;;; FIXME: Split into DYNBIND and BIND: DYNBIND needs to ensure
-;;; TLS-INDEX, whereas BIND should assume it is already in place. Make
-;;; LET &co compile into BIND, and PROGV into DYNBIND, plus ensure
-;;; TLS-INDEX at compile-time, and make loader and dumper preserve
-;;; the existence of a TLS-INDEX.
-#!+sb-thread
-(define-vop (bind)
-  (:args (val :scs (any-reg descriptor-reg))
-         (symbol :scs (descriptor-reg)))
-  (:temporary (:sc unsigned-reg) tls-index bsp)
-  (:generator 10
-    (let ((tls-index-valid (gen-label)))
-      (load-binding-stack-pointer bsp)
-      (loadw tls-index symbol symbol-tls-index-slot other-pointer-lowtag)
-      (inst add bsp (* binding-size n-word-bytes))
-      (store-binding-stack-pointer bsp)
-      (inst or tls-index tls-index)
-      (inst jmp :ne tls-index-valid)
-      (inst mov tls-index symbol)
-      (inst call (make-fixup
-                  (ecase (tn-offset tls-index)
-                    (#.eax-offset 'alloc-tls-index-in-eax)
-                    (#.ebx-offset 'alloc-tls-index-in-ebx)
-                    (#.ecx-offset 'alloc-tls-index-in-ecx)
-                    (#.edx-offset 'alloc-tls-index-in-edx)
-                    (#.edi-offset 'alloc-tls-index-in-edi)
-                    (#.esi-offset 'alloc-tls-index-in-esi))
-                  :assembly-routine))
-      (emit-label tls-index-valid)
-      (inst push (make-ea :dword :base tls-index) :fs)
-      (popw bsp (- binding-value-slot binding-size))
-      (storew symbol bsp (- binding-symbol-slot binding-size))
-      (inst mov (make-ea :dword :base tls-index) val :fs))))
-
-#!-sb-thread
-(define-vop (bind)
+#-sb-thread
+(define-vop (dynbind)
   (:args (val :scs (any-reg descriptor-reg))
          (symbol :scs (descriptor-reg)))
   (:temporary (:sc unsigned-reg) temp bsp)
@@ -316,24 +324,24 @@
     (storew symbol bsp (- binding-symbol-slot binding-size))
     (storew val symbol symbol-value-slot other-pointer-lowtag)))
 
-#!+sb-thread
+#+sb-thread
 (define-vop (unbind)
   (:temporary (:sc unsigned-reg) temp bsp tls-index)
   (:generator 0
     (load-binding-stack-pointer bsp)
     ;; Load SYMBOL from stack, and get the TLS-INDEX.
-    (loadw temp bsp (- binding-symbol-slot binding-size))
-    (loadw tls-index temp symbol-tls-index-slot other-pointer-lowtag)
+    (loadw tls-index bsp (- binding-symbol-slot binding-size))
     ;; Load VALUE from stack, then restore it to the TLS area.
     (loadw temp bsp (- binding-value-slot binding-size))
-    (inst mov (make-ea :dword :base tls-index) temp :fs)
+    (with-tls-ea (EA :base tls-index :base-already-live-p t)
+      (inst mov EA temp :maybe-fs))
     ;; Zero out the stack.
-    (storew 0 bsp (- binding-symbol-slot binding-size))
-    (storew 0 bsp (- binding-value-slot binding-size))
     (inst sub bsp (* binding-size n-word-bytes))
+    (storew 0 bsp binding-symbol-slot)
+    (storew 0 bsp binding-value-slot)
     (store-binding-stack-pointer bsp)))
 
-#!-sb-thread
+#-sb-thread
 (define-vop (unbind)
   (:temporary (:sc unsigned-reg) symbol value bsp)
   (:generator 0
@@ -349,55 +357,33 @@
 
 (define-vop (unbind-to-here)
   (:args (where :scs (descriptor-reg any-reg)))
-  (:temporary (:sc unsigned-reg) symbol value bsp #!+sb-thread tls-index)
+  (:temporary (:sc unsigned-reg) symbol value bsp)
   (:generator 0
     (load-binding-stack-pointer bsp)
     (inst cmp where bsp)
     (inst jmp :e done)
 
     LOOP
-    (loadw symbol bsp (- binding-symbol-slot binding-size))
-    (inst or symbol symbol)
+    (inst sub bsp (* binding-size n-word-bytes))
+    (loadw symbol bsp binding-symbol-slot)
+    (inst test symbol symbol)
     (inst jmp :z skip)
     ;; Bind stack debug sentinels have the unbound marker in the symbol slot
     (inst cmp symbol unbound-marker-widetag)
     (inst jmp :eq skip)
-    (loadw value bsp (- binding-value-slot binding-size))
-    #!-sb-thread (storew value symbol symbol-value-slot other-pointer-lowtag)
-
-    #!+sb-thread (loadw
-                  tls-index symbol symbol-tls-index-slot other-pointer-lowtag)
-    #!+sb-thread (inst mov (make-ea :dword :base tls-index) value :fs)
-    (storew 0 bsp (- binding-symbol-slot binding-size))
+    (loadw value bsp binding-value-slot)
+    #-sb-thread (storew value symbol symbol-value-slot other-pointer-lowtag)
+    #+sb-thread (with-tls-ea (EA :base symbol :base-already-live-p t)
+                   (inst mov EA value :maybe-fs))
+    (storew 0 bsp binding-symbol-slot)
 
     SKIP
-    (storew 0 bsp (- binding-value-slot binding-size))
-    (inst sub bsp (* binding-size n-word-bytes))
+    (storew 0 bsp binding-value-slot)
     (inst cmp where bsp)
     (inst jmp :ne loop)
     (store-binding-stack-pointer bsp)
 
     DONE))
-
-(define-vop (bind-sentinel)
-  (:temporary (:sc unsigned-reg) bsp)
-  (:generator 1
-     (load-binding-stack-pointer bsp)
-     (inst add bsp (* binding-size n-word-bytes))
-     (storew unbound-marker-widetag bsp (- binding-symbol-slot binding-size))
-     (storew ebp-tn bsp (- binding-value-slot binding-size))
-     (store-binding-stack-pointer bsp)))
-
-(define-vop (unbind-sentinel)
-  (:temporary (:sc unsigned-reg) bsp)
-  (:generator 1
-     (load-binding-stack-pointer bsp)
-     (storew 0 bsp (- binding-value-slot binding-size))
-     (storew 0 bsp (- binding-symbol-slot binding-size))
-     (inst sub bsp (* binding-size n-word-bytes))
-     (store-binding-stack-pointer bsp)))
-
-
 
 ;;;; closure indexing
 
@@ -413,11 +399,19 @@
   funcallable-instance-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %funcallable-instance-info)
 
-(define-vop (closure-ref slot-ref)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-ref)
+  (:args (object :scs (descriptor-reg)))
+  (:results (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (loadw value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
-(define-vop (closure-init slot-set)
-  (:variant closure-info-offset fun-pointer-lowtag))
+(define-vop (closure-init)
+  (:args (object :scs (descriptor-reg))
+         (value :scs (descriptor-reg any-reg)))
+  (:info offset)
+  (:generator 4
+    (storew value object (+ closure-info-offset offset) fun-pointer-lowtag)))
 
 (define-vop (closure-init-from-fp)
   (:args (object :scs (descriptor-reg)))
@@ -435,7 +429,7 @@
 
 ;;;; structure hackery
 
-(define-vop (instance-length)
+(define-vop ()
   (:policy :fast-safe)
   (:translate %instance-length)
   (:args (struct :scs (descriptor-reg)))
@@ -443,7 +437,7 @@
   (:result-types positive-fixnum)
   (:generator 4
     (loadw res struct 0 instance-pointer-lowtag)
-    (inst shr res n-widetag-bits)))
+    (inst shr res instance-length-shift)))
 
 (define-full-reffer instance-index-ref *
   instance-slots-offset instance-pointer-lowtag
@@ -455,91 +449,85 @@
   (any-reg descriptor-reg) *
   %instance-set)
 
-(define-full-compare-and-swap %compare-and-swap-instance-ref instance
+(define-full-compare-and-swap %instance-cas instance
   instance-slots-offset instance-pointer-lowtag
   (any-reg descriptor-reg) *
-  %compare-and-swap-instance-ref)
+  %instance-cas)
+(define-full-compare-and-swap %raw-instance-cas/word instance
+  instance-slots-offset instance-pointer-lowtag
+  (unsigned-reg) unsigned-num %raw-instance-cas/word)
 
 ;;;; code object frobbing
 
 (define-full-reffer code-header-ref * 0 other-pointer-lowtag
   (any-reg descriptor-reg) * code-header-ref)
 
-(define-full-setter code-header-set * 0 other-pointer-lowtag
-  (any-reg descriptor-reg) * code-header-set)
+(define-vop (code-header-set)
+  (:translate code-header-set)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (unsigned-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * unsigned-num *)
+  (:temporary (:sc unsigned-reg :offset eax-offset) eax) ; for the asm routine
+  (:temporary (:sc unsigned-reg :offset edx-offset) edx)
+  (:temporary (:sc unsigned-reg :offset edi-offset) edi)
+  (:ignore eax edx edi)
+  (:generator 10
+    (inst push value)
+    (inst push index)
+    (inst push object)
+    (inst call (make-fixup 'code-header-set :assembly-routine))))
 
 ;;;; raw instance slot accessors
 
-(defun make-ea-for-raw-slot (object index instance-length n-words)
-  (if (integerp instance-length)
-      ;; For RAW-INSTANCE-INIT/* VOPs, which know the exact instance length
-      ;; at compile time.
-      (make-ea :dword
-               :base object
-               :disp (- (* (- instance-length instance-slots-offset index (1- n-words))
-                           n-word-bytes)
-                        instance-pointer-lowtag))
-      (flet ((make-ea-using-value (value)
-               (make-ea :dword :base object
-                        :index instance-length
-                        :scale 4
-                        :disp (- (* (- instance-slots-offset n-words)
-                                    n-word-bytes)
-                                 instance-pointer-lowtag
-                                 (* value n-word-bytes)))))
-        (if (typep index 'tn)
-            (sc-case index
-              (any-reg (make-ea :dword
-                                :base object
-                                :index instance-length
-                                :disp (- (* (- instance-slots-offset n-words)
-                                            n-word-bytes)
-                                         instance-pointer-lowtag)))
-              (immediate (make-ea-using-value (tn-value index))))
-            (make-ea-using-value index)))))
+(defun instance-slot-ea (object index &optional (displacement 0))
+  ;; instance-init vops pass a literal integer, ref/set can use an immediate tn
+  (let ((imm-index (cond ((integerp index) index)
+                         ((sc-is index immediate) (tn-value index)))))
+    (make-ea :dword :base object
+             ;; If index is a register, it needs no scaling - it has tag bits.
+             :index (unless imm-index index)
+             :disp (- (ash (+ (or imm-index 0) displacement instance-slots-offset)
+                           word-shift) instance-pointer-lowtag))))
 
-(define-vop (raw-instance-ref/word)
+(define-vop ()
   (:translate %raw-instance-ref/word)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg)) (index :scs (any-reg immediate)))
   (:arg-types * tagged-num)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (value :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (inst mov value (make-ea-for-raw-slot object index tmp 1))))
+    (inst mov value (instance-slot-ea object index))))
 
-(define-vop (raw-instance-set/word)
+(define-vop ()
   (:translate %raw-instance-set/word)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate))
-         (value :scs (unsigned-reg) :target result))
-  (:arg-types * tagged-num unsigned-num)
-  (:temporary (:sc unsigned-reg) tmp)
-  (:results (result :scs (unsigned-reg)))
-  (:result-types unsigned-num)
-  (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (inst mov (make-ea-for-raw-slot object index tmp 1) value)
-    (move result value)))
-
-(define-vop (raw-instance-init/word)
-  (:args (object :scs (descriptor-reg))
          (value :scs (unsigned-reg)))
-  (:arg-types * unsigned-num)
-  (:info instance-length index)
+  (:arg-types * tagged-num unsigned-num)
+  (:generator 5 (inst mov (instance-slot-ea object index) value)))
+
+(define-vop ()
+  (:translate %raw-instance-ref/signed-word)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg)) (index :scs (any-reg immediate)))
+  (:arg-types * tagged-num)
+  (:results (value :scs (signed-reg)))
+  (:result-types signed-num)
   (:generator 5
-    (inst mov (make-ea-for-raw-slot object index instance-length 1) value)))
+    (inst mov value (instance-slot-ea object index))))
+
+(define-vop ()
+  (:translate %raw-instance-set/signed-word)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg immediate))
+         (value :scs (signed-reg)))
+  (:arg-types * tagged-num signed-num)
+  (:generator 5 (inst mov (instance-slot-ea object index) value)))
 
 (define-vop (raw-instance-atomic-incf/word)
   (:translate %raw-instance-atomic-incf/word)
@@ -548,284 +536,182 @@
          (index :scs (any-reg immediate))
          (diff :scs (unsigned-reg) :target result))
   (:arg-types * tagged-num unsigned-num)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (result :scs (unsigned-reg)))
   (:result-types unsigned-num)
   (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (inst xadd (make-ea-for-raw-slot object index tmp 1) diff :lock)
+    (inst xadd (instance-slot-ea object index) diff :lock)
     (move result diff)))
 
-(define-vop (raw-instance-ref/single)
+(define-vop ()
   (:translate %raw-instance-ref/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg)) (index :scs (any-reg immediate)))
   (:arg-types * tagged-num)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (value :scs (single-reg)))
   (:result-types single-float)
   (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
     (with-empty-tn@fp-top(value)
-      (inst fld (make-ea-for-raw-slot object index tmp 1)))))
+      (inst fld (instance-slot-ea object index)))))
 
-(define-vop (raw-instance-set/single)
+(define-vop ()
   (:translate %raw-instance-set/single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate))
-         (value :scs (single-reg) :target result))
-  (:arg-types * tagged-num single-float)
-  (:temporary (:sc unsigned-reg) tmp)
-  (:results (result :scs (single-reg)))
-  (:result-types single-float)
-  (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (unless (zerop (tn-offset value))
-      (inst fxch value))
-    (inst fst (make-ea-for-raw-slot object index tmp 1))
-    (cond
-      ((zerop (tn-offset value))
-        (unless (zerop (tn-offset result))
-          (inst fst result)))
-      ((zerop (tn-offset result))
-        (inst fst value))
-      (t
-        (unless (location= value result)
-          (inst fst result))
-        (inst fxch value)))))
-
-(define-vop (raw-instance-init/single)
-  (:args (object :scs (descriptor-reg))
          (value :scs (single-reg)))
-  (:arg-types * single-float)
-  (:info instance-length index)
+  (:arg-types * tagged-num single-float)
   (:generator 5
-    (with-tn@fp-top (value)
-      (inst fst (make-ea-for-raw-slot object index instance-length 1)))))
+    (with-tn@fp-top (value) (inst fst (instance-slot-ea object index)))))
 
-(define-vop (raw-instance-ref/double)
+(define-vop ()
   (:translate %raw-instance-ref/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg)) (index :scs (any-reg immediate)))
   (:arg-types * tagged-num)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (value :scs (double-reg)))
   (:result-types double-float)
   (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
     (with-empty-tn@fp-top(value)
-      (inst fldd (make-ea-for-raw-slot object index tmp 2)))))
+      (inst fldd (instance-slot-ea object index)))))
 
-(define-vop (raw-instance-set/double)
+(define-vop ()
   (:translate %raw-instance-set/double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate))
-         (value :scs (double-reg) :target result))
-  (:arg-types * tagged-num double-float)
-  (:temporary (:sc unsigned-reg) tmp)
-  (:results (result :scs (double-reg)))
-  (:result-types double-float)
-  (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (unless (zerop (tn-offset value))
-      (inst fxch value))
-    (inst fstd (make-ea-for-raw-slot object index tmp 2))
-    (cond
-      ((zerop (tn-offset value))
-        (unless (zerop (tn-offset result))
-          (inst fstd result)))
-      ((zerop (tn-offset result))
-        (inst fstd value))
-      (t
-        (unless (location= value result)
-          (inst fstd result))
-        (inst fxch value)))))
-
-(define-vop (raw-instance-init/double)
-  (:args (object :scs (descriptor-reg))
          (value :scs (double-reg)))
-  (:arg-types * double-float)
-  (:info instance-length index)
+  (:arg-types * tagged-num double-float)
   (:generator 5
-    (with-tn@fp-top (value)
-      (inst fstd (make-ea-for-raw-slot object index instance-length 2)))))
+    (with-tn@fp-top (value) (inst fstd (instance-slot-ea object index)))))
 
-(define-vop (raw-instance-ref/complex-single)
+(define-vop ()
   (:translate %raw-instance-ref/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate)))
   (:arg-types * positive-fixnum)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (value :scs (complex-single-reg)))
   (:result-types complex-single-float)
   (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
     (let ((real-tn (complex-single-reg-real-tn value)))
       (with-empty-tn@fp-top (real-tn)
-        (inst fld (make-ea-for-raw-slot object index tmp 2))))
+        (inst fld (instance-slot-ea object index))))
     (let ((imag-tn (complex-single-reg-imag-tn value)))
       (with-empty-tn@fp-top (imag-tn)
-        (inst fld (make-ea-for-raw-slot object index tmp 1))))))
+        (inst fld (instance-slot-ea object index 1))))))
 
-(define-vop (raw-instance-set/complex-single)
+(define-vop ()
   (:translate %raw-instance-set/complex-single)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate))
-         (value :scs (complex-single-reg) :target result))
-  (:arg-types * positive-fixnum complex-single-float)
-  (:temporary (:sc unsigned-reg) tmp)
-  (:results (result :scs (complex-single-reg)))
-  (:result-types complex-single-float)
-  (:generator 5
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (let ((value-real (complex-single-reg-real-tn value))
-          (result-real (complex-single-reg-real-tn result)))
-      (cond ((zerop (tn-offset value-real))
-             ;; Value is in ST0.
-             (inst fst (make-ea-for-raw-slot object index tmp 2))
-             (unless (zerop (tn-offset result-real))
-               ;; Value is in ST0 but not result.
-               (inst fst result-real)))
-            (t
-             ;; Value is not in ST0.
-             (inst fxch value-real)
-             (inst fst (make-ea-for-raw-slot object index tmp 2))
-             (cond ((zerop (tn-offset result-real))
-                    ;; The result is in ST0.
-                    (inst fst value-real))
-                   (t
-                    ;; Neither value or result are in ST0
-                    (unless (location= value-real result-real)
-                      (inst fst result-real))
-                    (inst fxch value-real))))))
-    (let ((value-imag (complex-single-reg-imag-tn value))
-          (result-imag (complex-single-reg-imag-tn result)))
-      (inst fxch value-imag)
-      (inst fst (make-ea-for-raw-slot object index tmp 1))
-      (unless (location= value-imag result-imag)
-        (inst fst result-imag))
-      (inst fxch value-imag))))
-
-(define-vop (raw-instance-init/complex-single)
-  (:args (object :scs (descriptor-reg))
          (value :scs (complex-single-reg)))
-  (:arg-types * complex-single-float)
-  (:info instance-length index)
+  (:arg-types * positive-fixnum complex-single-float)
   (:generator 5
     (let ((value-real (complex-single-reg-real-tn value)))
       (with-tn@fp-top (value-real)
-        (inst fst (make-ea-for-raw-slot object index instance-length 2))))
+        (inst fst (instance-slot-ea object index))))
     (let ((value-imag (complex-single-reg-imag-tn value)))
-      (with-tn@fp-top (value-imag)
-        (inst fst (make-ea-for-raw-slot object index instance-length 1))))))
+      (inst fxch value-imag)
+      (inst fst (instance-slot-ea object index 1))
+      (inst fxch value-imag))))
 
-(define-vop (raw-instance-ref/complex-double)
+(define-vop ()
   (:translate %raw-instance-ref/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate)))
   (:arg-types * positive-fixnum)
-  (:temporary (:sc unsigned-reg) tmp)
   (:results (value :scs (complex-double-reg)))
   (:result-types complex-double-float)
   (:generator 7
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
     (let ((real-tn (complex-double-reg-real-tn value)))
       (with-empty-tn@fp-top (real-tn)
-        (inst fldd (make-ea-for-raw-slot object index tmp 4))))
+        (inst fldd (instance-slot-ea object index))))
     (let ((imag-tn (complex-double-reg-imag-tn value)))
       (with-empty-tn@fp-top (imag-tn)
-        (inst fldd (make-ea-for-raw-slot object index tmp 2))))))
+        (inst fldd (instance-slot-ea object index 2))))))
 
-(define-vop (raw-instance-set/complex-double)
+(define-vop ()
   (:translate %raw-instance-set/complex-double)
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg immediate))
-         (value :scs (complex-double-reg) :target result))
-  (:arg-types * positive-fixnum complex-double-float)
-  (:temporary (:sc unsigned-reg) tmp)
-  (:results (result :scs (complex-double-reg)))
-  (:result-types complex-double-float)
-  (:generator 20
-    (loadw tmp object 0 instance-pointer-lowtag)
-    (inst shr tmp n-widetag-bits)
-    (when (sc-is index any-reg)
-      (inst shl tmp n-fixnum-tag-bits)
-      (inst sub tmp index))
-    (let ((value-real (complex-double-reg-real-tn value))
-          (result-real (complex-double-reg-real-tn result)))
-      (cond ((zerop (tn-offset value-real))
-             ;; Value is in ST0.
-             (inst fstd (make-ea-for-raw-slot object index tmp 4))
-             (unless (zerop (tn-offset result-real))
-               ;; Value is in ST0 but not result.
-               (inst fstd result-real)))
-            (t
-             ;; Value is not in ST0.
-             (inst fxch value-real)
-             (inst fstd (make-ea-for-raw-slot object index tmp 4))
-             (cond ((zerop (tn-offset result-real))
-                    ;; The result is in ST0.
-                    (inst fstd value-real))
-                   (t
-                    ;; Neither value or result are in ST0
-                    (unless (location= value-real result-real)
-                      (inst fstd result-real))
-                    (inst fxch value-real))))))
-    (let ((value-imag (complex-double-reg-imag-tn value))
-          (result-imag (complex-double-reg-imag-tn result)))
-      (inst fxch value-imag)
-      (inst fstd (make-ea-for-raw-slot object index tmp 2))
-      (unless (location= value-imag result-imag)
-        (inst fstd result-imag))
-      (inst fxch value-imag))))
-
-(define-vop (raw-instance-init/complex-double)
-  (:args (object :scs (descriptor-reg))
          (value :scs (complex-double-reg)))
-  (:arg-types * complex-double-float)
-  (:info instance-length index)
+  (:arg-types * positive-fixnum complex-double-float)
   (:generator 20
     (let ((value-real (complex-double-reg-real-tn value)))
       (with-tn@fp-top (value-real)
-        (inst fstd (make-ea-for-raw-slot object index instance-length 4))))
+        (inst fstd (instance-slot-ea object index))))
     (let ((value-imag (complex-double-reg-imag-tn value)))
-      (with-tn@fp-top (value-imag)
-        (inst fstd (make-ea-for-raw-slot object index instance-length 2))))))
+      (inst fxch value-imag)
+      (inst fstd (instance-slot-ea object index 2))
+      (inst fxch value-imag))))
+
+;;;;
+
+(defknown %cons-cas-pair (cons t t t t) (values t t))
+(defknown %vector-cas-pair (simple-vector index t t t t) (values t t))
+;; %INSTANCE-CAS-PAIR only operates on tagged slots (for now)
+(defknown %instance-cas-pair (instance index t t t t) (values t t))
+
+(defun generate-dblcas (memory-operand old-lo old-hi new-lo new-hi
+                        eax ebx ecx edx result-lo result-hi)
+  (move eax old-lo)
+  (move edx old-hi)
+  (move ebx new-lo)
+  (move ecx new-hi)
+  (inst cmpxchg8b memory-operand :lock)
+  ;; EDX:EAX hold the actual old contents of memory.
+  ;; Manually analyze result lifetimes to avoid clobbering.
+  (cond ((and (location= result-lo edx) (location= result-hi eax))
+         (inst xchg eax edx)) ; unlikely, but possible
+        ((location= result-lo edx) ; result-hi is not eax
+         (move result-hi edx) ; move high part first
+         (move result-lo eax))
+        (t                    ; result-lo is not edx
+         (move result-lo eax) ; move low part first
+         (move result-hi edx))))
+
+(macrolet
+    ((define-cmpxchg-vop (name memory-operand more-stuff &optional index-arg)
+       `(define-vop (,name)
+          (:policy :fast-safe)
+          ,@more-stuff
+          (:args (data :scs (descriptor-reg) :to :eval)
+                 ,@index-arg
+                 (expected-old-lo :scs (descriptor-reg any-reg) :target eax)
+                 (expected-old-hi :scs (descriptor-reg any-reg) :target edx)
+                 (new-lo :scs (descriptor-reg any-reg) :target ebx)
+                 (new-hi :scs (descriptor-reg any-reg) :target ecx))
+          (:results (result-lo :scs (descriptor-reg any-reg))
+                    (result-hi :scs (descriptor-reg any-reg)))
+          (:temporary (:sc unsigned-reg :offset eax-offset
+                       :from (:argument 2) :to (:result 0)) eax)
+          (:temporary (:sc unsigned-reg :offset edx-offset
+                       :from (:argument 3) :to (:result 0)) edx)
+          (:temporary (:sc unsigned-reg :offset ebx-offset
+                       :from (:argument 4) :to (:result 0)) ebx)
+          (:temporary (:sc unsigned-reg :offset ecx-offset
+                       :from (:argument 5) :to (:result 0)) ecx)
+          (:generator 7
+           (generate-dblcas ,memory-operand
+                            expected-old-lo expected-old-hi new-lo new-hi
+                            eax ebx ecx edx result-lo result-hi)))))
+  (define-cmpxchg-vop compare-and-exchange-pair
+      (make-ea :dword :base data :disp (- list-pointer-lowtag))
+      ((:translate %cons-cas-pair)))
+  (define-cmpxchg-vop compare-and-exchange-pair-indexed
+      (make-ea :dword :base data :disp offset :index index
+                      :scale (ash n-word-bytes (- n-fixnum-tag-bits)))
+      ((:variant-vars offset))
+      ((index :scs (descriptor-reg any-reg) :to :eval))))
+
+(define-vop (%vector-cas-pair compare-and-exchange-pair-indexed)
+  (:translate %vector-cas-pair)
+  (:variant (- (* n-word-bytes vector-data-offset) other-pointer-lowtag)))
+
+(define-vop (%instance-cas-pair compare-and-exchange-pair-indexed)
+  (:translate %instance-cas-pair)
+  (:variant (- (* n-word-bytes instance-slots-offset) instance-pointer-lowtag)))
+
