@@ -17,6 +17,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <ctype.h>
 #ifndef LISP_FEATURE_WIN32
 #include <libgen.h>
 #endif
@@ -29,19 +30,21 @@
 #include <sys/file.h>
 #include <sys/param.h>
 #include <sys/stat.h>
-#include <signal.h>
+#include "runtime.h"
 #ifndef LISP_FEATURE_WIN32
 #include <sched.h>
+#else
+#include <shellapi.h>
 #endif
 #include <errno.h>
 #include <locale.h>
 #include <limits.h>
 
-#if defined(SVR4) || defined(__linux__)
 #include <time.h>
-#endif
 
+#ifndef LISP_FEATURE_WIN32
 #include "signal.h"
+#endif
 
 #include "runtime.h"
 #include "vars.h"
@@ -61,40 +64,12 @@
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
 
+struct timespec lisp_init_time;
 
-#ifdef irix
-#include <string.h>
-#include "interr.h"
-#endif
+static char libpath[] = "../lib/sbcl";
+char *sbcl_runtime_home;
+char *sbcl_runtime;
 
-#ifndef SBCL_HOME
-#define SBCL_HOME SBCL_PREFIX"/lib/sbcl/"
-#endif
-
-#ifdef LISP_FEATURE_HPUX
-extern void *return_from_lisp_stub;
-#include "genesis/closure.h"
-#include "genesis/simple-fun.h"
-#endif
-
-
-/* SIGINT handler that invokes the monitor (for when Lisp isn't up to it) */
-static void
-sigint_handler(int signal, siginfo_t *info, os_context_t *context)
-{
-    lose("\nSIGINT hit at 0x%08lX\n",
-         (unsigned long) *os_context_pc_addr(context));
-}
-
-/* (This is not static, because we want to be able to call it from
- * Lisp land.) */
-void
-sigint_init(void)
-{
-    SHOW("entering sigint_init()");
-    install_handler(SIGINT, sigint_handler);
-    SHOW("leaving sigint_init()");
-}
 
 /*
  * helper functions for dealing with command line args
@@ -105,7 +80,7 @@ successful_malloc(size_t size)
 {
     void* result = malloc(size);
     if (0 == result) {
-        lose("malloc failure\n");
+        lose("malloc failure");
     } else {
         return result;
     }
@@ -118,7 +93,7 @@ copied_string(char *string)
     return strcpy(successful_malloc(1+strlen(string)), string);
 }
 
-char *
+static char *
 copied_existing_filename_or_null(char *filename)
 {
     struct stat filename_stat;
@@ -130,7 +105,7 @@ copied_existing_filename_or_null(char *filename)
 }
 
 #ifndef LISP_FEATURE_WIN32
-char *
+static char *
 copied_realpath(const char *pathname)
 {
     char *messy, *tidy;
@@ -151,18 +126,22 @@ copied_realpath(const char *pathname)
 
     tidy = successful_malloc(PATH_MAX + 1);
     if (realpath((messy ? messy : pathname), tidy) == NULL) {
-        free(messy);
+        if (messy)
+            free(messy);
         free(tidy);
         return NULL;
     }
 
+    if (messy)
+        free(messy);
+
     return tidy;
 }
 #endif /* LISP_FEATURE_WIN32 */
 
 /* miscellaneous chattiness */
 
-void
+static void
 print_help()
 {
     puts(
@@ -173,12 +152,21 @@ Common runtime options:\n\
   --core <filename>          Use the specified core file instead of the default.\n\
   --dynamic-space-size <MiB> Size of reserved dynamic space in megabytes.\n\
   --control-stack-size <MiB> Size of reserved control stack in megabytes.\n\
+  --tls-limit                Maximum number of thread-local symbols.\n\
 \n\
 Common toplevel options:\n\
   --sysinit <filename>       System-wide init-file to use instead of default.\n\
   --userinit <filename>      Per-user init-file to use instead of default.\n\
   --no-sysinit               Inhibit processing of any system-wide init-file.\n\
   --no-userinit              Inhibit processing of any per-user init-file.\n\
+  --disable-debugger         Invoke sb-ext:disable-debugger.\n\
+  --noprint                  Run a Read-Eval Loop without printing results.\n\
+  --script [<filename>]      Skip #! line, disable debugger, avoid verbosity.\n\
+  --quit                     Exit with code 0 after option processing.\n\
+  --non-interactive          Sets both --quit and --disable-debugger.\n\
+Common toplevel options that are processed in order:\n\
+  --eval <form>              Form to eval when processing this option.\n\
+  --load <filename>          File to load when processing this option.\n\
 \n\
 User options are not processed by SBCL. All runtime options must\n\
 appear before toplevel options, and all toplevel options must\n\
@@ -189,13 +177,13 @@ should be installed along with SBCL, and is also available from the\n\
 website <http://www.sbcl.org/>.\n");
 }
 
-void
+static void
 print_version()
 {
     printf("SBCL %s\n", SBCL_VERSION_STRING);
 }
 
-void
+static void
 print_banner()
 {
     printf(
@@ -212,41 +200,59 @@ distribution for more information.\n\
 /* Look for a core file to load, first in the directory named by the
  * SBCL_HOME environment variable, then in a hardcoded default
  * location.  Returns a malloced copy of the core filename. */
-char *
+static char *
 search_for_core ()
 {
-    char *sbcl_home = getenv("SBCL_HOME");
+    char *env_sbcl_home = getenv("SBCL_HOME");
     char *lookhere;
     char *stem = "/sbcl.core";
     char *core;
+    struct stat filename_stat;
 
-    if (!(sbcl_home && *sbcl_home)) sbcl_home = SBCL_HOME;
-    lookhere = (char *) calloc(strlen(sbcl_home) +
+    if (!(env_sbcl_home && *env_sbcl_home) ||
+        stat(env_sbcl_home, &filename_stat))
+        env_sbcl_home = sbcl_runtime_home;
+    lookhere = (char *) calloc(strlen(env_sbcl_home) +
+                               strlen(libpath) +
                                strlen(stem) +
                                1,
                                sizeof(char));
-    sprintf(lookhere, "%s%s", sbcl_home, stem);
+    sprintf(lookhere, "%s%s%s", env_sbcl_home, libpath, stem);
     core = copied_existing_filename_or_null(lookhere);
 
-    if (!core) {
-        lose("can't find core file at %s\n", lookhere);
+    if (core) {
+        free(lookhere);
+    } else {
+        free(lookhere);
+        core = copied_existing_filename_or_null ("sbcl.core");
+        if (!core) {
+            lookhere = (char *) calloc(strlen(env_sbcl_home) +
+                                       strlen(stem) +
+                                       1,
+                                       sizeof(char));
+            sprintf(lookhere, "%s%s", env_sbcl_home, stem);
+            core = copied_existing_filename_or_null (lookhere);
+            if (core) {
+                free(lookhere);
+            } else {
+                return NULL;
+            }
+        }
     }
 
-    free(lookhere);
-
     return core;
 }
 
 /* Try to find the path to an executable from argv[0], this is only
  * used when os_get_runtime_executable_path() returns NULL */
 #ifdef LISP_FEATURE_WIN32
-char *
+static char *
 search_for_executable(const char *argv0)
 {
     return NULL;
 }
 #else /* LISP_FEATURE_WIN32 */
-char *
+static char *
 search_for_executable(const char *argv0)
 {
     char *search, *start, *end, *buf;
@@ -276,6 +282,18 @@ search_for_executable(const char *argv0)
             return search;
         }
     }
+    /* The above for-loop fails to process the last part of PATH if PATH does
+     * not end with ':'. We may consider appending an extra ':' to the end of
+     * SEARCH.  -- houjingyi 2013-05-24 */
+    if (start != NULL && *start != '\0') {
+        snprintf(buf, PATH_MAX + 1, "%s/%s", start, argv0);
+        if (access(buf, F_OK) == 0) {
+            free(search);
+            search = copied_realpath(buf);
+            free(buf);
+            return search;
+        }
+    }
 
     free(search);
     free(buf);
@@ -283,76 +301,191 @@ search_for_executable(const char *argv0)
 }
 #endif /* LISP_FEATURE_WIN32 */
 
-char **posix_argv;
+static size_t
+parse_size_arg(char *arg, char *arg_name)
+{
+  char *tail, *power_name;
+  size_t power, res;
+
+  res = strtoul(arg, &tail, 0);
+
+  if (arg == tail) {
+    lose("%s argument is not a number: %s", arg_name, arg);
+  } else if (tail[0]) {
+    int i, size;
+    power_name = copied_string(tail);
+    size = strlen(power_name);
+    for (i=0; i<size; i++)
+      power_name[i] = toupper(power_name[i]);
+  } else {
+    power = 20;
+    power_name = NULL;
+  }
+  if (power_name) {
+    if ((0==strcmp("KB", power_name)) ||
+        (0==strcmp("KIB", power_name))) {
+      power = 10;
+    } else if ((0==strcmp("MB", power_name)) ||
+               (0==strcmp("MIB", power_name))) {
+      power = 20;
+    } else if ((0==strcmp("GB", power_name)) ||
+               (0==strcmp("GIB", power_name))) {
+      power = 30;
+    } else {
+      lose("%s argument has an unknown suffix: %s", arg_name, tail);
+    }
+    free(power_name);
+  }
+  if ((res <= 0) ||
+      (res >= (SIZE_MAX >> power))) {
+    lose("%s argument is out of range: %s", arg_name, arg);
+  }
+  res <<= power;
+  return res;
+}
+
+#ifdef LISP_FEATURE_WIN32
+    wchar_t
+#else
+    char
+#endif
+    **posix_argv;
+
 char *core_string;
 
-struct runtime_options *runtime_options;
+static void print_environment(int argc, char *argv[])
+{
+    int n = 0;
+    printf("; Commandline arguments:\n");
+    while (n < argc) {
+        printf(";  %2d: \"%s\"\n", n, argv[n]);
+        ++n;
+    }
+    n = 0;
+    printf(";\n; Environment:\n");
+    while (ENVIRON[n]) {
+        printf(";  %2d: \"%s\"\n", n, ENVIRON[n]);
+        ++n;
+    }
+}
+
+char * sb_realpath (char *);
+char *dir_name(char *path) {
+    if (path == NULL)
+        return NULL;
+
+    char* result;
+    char slashchar =
+#ifdef LISP_FEATURE_WIN32
+        '\\';
+#else
+    '/';
+#endif
+
+    char *slash = strrchr(path, slashchar);
+
+    if (slash) {
+        int prefixlen = slash - path + 1; // keep the slash in the prefix
+        result = successful_malloc(prefixlen + 1);
+        memcpy(result, path, prefixlen);
+        result[prefixlen] = 0;
+        return result;
+    } else {
+        return NULL;
+    }
+}
 
-char *saved_runtime_path = NULL;
 
+extern void write_protect_immobile_space();
+struct lisp_startup_options lisp_startup_options;
 int
-main(int argc, char *argv[], char *envp[])
+initialize_lisp(int argc, char *argv[], char *envp[])
 {
 #ifdef LISP_FEATURE_WIN32
     /* Exception handling support structure. Evil Win32 hack. */
     struct lisp_exception_frame exception_frame;
 #endif
+#ifdef LISP_FEATURE_UNIX
+    clock_gettime(
+#ifdef LISP_FEATURE_LINUX
+        CLOCK_MONOTONIC_COARSE
+#else
+        CLOCK_MONOTONIC
+#endif
+        , &lisp_init_time);
+#endif
 
     /* the name of the core file we're to execute. Note that this is
      * a malloc'ed string which should be freed eventually. */
     char *core = 0;
-    char **sbcl_argv = 0;
+
+#ifdef LISP_FEATURE_WIN32
+    wchar_t
+#else
+        char
+#endif
+        **sbcl_argv = 0;
     os_vm_offset_t embedded_core_offset = 0;
-    char *runtime_path = 0;
 
     /* other command line options */
-    boolean noinform = 0;
     boolean end_runtime_options = 0;
-    boolean disable_lossage_handler_p = 0;
+    boolean disable_lossage_handler_p
+#if defined(LISP_FEATURE_SB_LDB)
+        = 0;
+#else
+        = 1;
+#endif
+
+    boolean debug_environment_p = 0;
 
     lispobj initial_function;
-    const char *sbcl_home = getenv("SBCL_HOME");
+    int merge_core_pages = -1;
+    struct memsize_options memsize_options;
+    memsize_options.present_in_core = 0;
+
+    boolean have_hardwired_spaces = os_preinit(argv, envp);
 
     interrupt_init();
-    block_blockable_signals(0, 0);
-
-    setlocale(LC_ALL, "");
-
-    runtime_options = NULL;
-
-    /* Save the argv[0] derived runtime path in case
-     * os_get_runtime_executable_path(1) isn't able to get an
-     * externally-usable path later on. */
-    saved_runtime_path = search_for_executable(argv[0]);
+#ifdef LISP_FEATURE_UNIX
+    /* Not sure why anyone sends signals to this process so early.
+     * But win32 models the signal mask as part of 'struct thread'
+     * which doesn't exist yet, so don't do this */
+    block_blockable_signals(0);
+#endif
 
     /* Check early to see if this executable has an embedded core,
      * which also populates runtime_options if the core has runtime
      * options */
-    runtime_path = os_get_runtime_executable_path(0);
-    if (runtime_path || saved_runtime_path) {
-        os_vm_offset_t offset = search_for_embedded_core(
-            runtime_path ? runtime_path : saved_runtime_path);
+    if (!(sbcl_runtime = os_get_runtime_executable_path()))
+        sbcl_runtime = search_for_executable(argv[0]);
+
+    if (!(sbcl_runtime_home = dir_name(argv[0])))
+      if (!(sbcl_runtime_home = dir_name(sbcl_runtime)))
+        sbcl_runtime_home = libpath;
+
+    if (sbcl_runtime) {
+        os_vm_offset_t offset = search_for_embedded_core(sbcl_runtime, &memsize_options);
         if (offset != -1) {
             embedded_core_offset = offset;
-            core = (runtime_path ? runtime_path :
-                    copied_string(saved_runtime_path));
-        } else {
-            free(runtime_path);
+            core = sbcl_runtime;
         }
     }
 
-
     /* Parse our part of the command line (aka "runtime options"),
      * stripping out those options that we handle. */
-    if (runtime_options != NULL) {
-        dynamic_space_size = runtime_options->dynamic_space_size;
-        thread_control_stack_size = runtime_options->thread_control_stack_size;
+    if (memsize_options.present_in_core) {
+        dynamic_space_size = memsize_options.dynamic_space_size;
+        thread_control_stack_size = memsize_options.thread_control_stack_size;
+        dynamic_values_bytes = memsize_options.thread_tls_bytes;
+#ifndef LISP_FEATURE_WIN32
         sbcl_argv = argv;
+#else
+        int wargc;
+        sbcl_argv = CommandLineToArgvW(GetCommandLineW(), &wargc);
+#endif
     } else {
         int argi = 1;
 
-        runtime_options = successful_malloc(sizeof(struct runtime_options));
-
         while (argi < argc) {
             char *arg = argv[argi];
             if (0 == strcmp(arg, "--script")) {
@@ -360,21 +493,21 @@ main(int argc, char *argv[], char *envp[])
                  * runtime option, it is equivalent to --noinform.
                  * This exits, and does not increment argi, so that
                  * TOPLEVEL-INIT sees the option. */
-                noinform = 1;
+                lisp_startup_options.noinform = 1;
                 end_runtime_options = 1;
                 disable_lossage_handler_p = 1;
                 lose_on_corruption_p = 1;
                 break;
             } else if (0 == strcmp(arg, "--noinform")) {
-                noinform = 1;
+                lisp_startup_options.noinform = 1;
                 ++argi;
             } else if (0 == strcmp(arg, "--core")) {
                 if (core) {
-                    lose("more than one core file specified\n");
+                    lose("more than one core file specified");
                 } else {
                     ++argi;
                     if (argi >= argc) {
-                        lose("missing filename for --core argument\n");
+                        lose("missing filename for --core argument");
                     }
                     core = copied_string(argv[argi]);
                     ++argi;
@@ -393,38 +526,38 @@ main(int argc, char *argv[], char *envp[])
                 ++argi;
                 if (argi >= argc)
                     lose("missing argument for --dynamic-space-size");
-                errno = 0;
-                dynamic_space_size = strtol(argv[argi++], 0, 0) << 20;
-                if (errno)
-                    lose("argument to --dynamic-space-size is not a number");
+                  dynamic_space_size = parse_size_arg(argv[argi++],
+                                                      "--dynamic-space-size");
 #               ifdef MAX_DYNAMIC_SPACE_END
                 if (!((DYNAMIC_SPACE_START <
                        DYNAMIC_SPACE_START+dynamic_space_size) &&
                       (DYNAMIC_SPACE_START+dynamic_space_size <=
-                       MAX_DYNAMIC_SPACE_END)))
-                    lose("specified --dynamic-space-size too large");
+                       MAX_DYNAMIC_SPACE_END))) {
+                  char* suffix = "";
+                  char* size = argv[argi-1];
+                  if (!strchr(size, 'B') && !strchr(size, 'b')) {
+                    suffix = " [MB]";
+                  }
+                  lose("--dynamic-space-size argument %s%s is too large, max %lu KB",
+                       size, suffix,
+                       (MAX_DYNAMIC_SPACE_END-DYNAMIC_SPACE_START) / 1024);
+                }
 #               endif
             } else if (0 == strcmp(arg, "--control-stack-size")) {
                 ++argi;
                 if (argi >= argc)
                     lose("missing argument for --control-stack-size");
                 errno = 0;
-                thread_control_stack_size = strtol(argv[argi++], 0, 0) << 20;
-                if (errno)
-                    lose("argument to --control-stack-size is not a number");
+                thread_control_stack_size = parse_size_arg(argv[argi++], "--control-stack-size");
+            } else if (0 == strcmp(arg, "--tls-limit")) {
+                // this is not named "tls-size" because "size" is not the
+                // best measurement for how many symbols to allow
+                ++argi;
+                if (argi >= argc)
+                    lose("missing argument for --tls-limit");
+                dynamic_values_bytes = N_WORD_BYTES * atoi(argv[argi++]);
             } else if (0 == strcmp(arg, "--debug-environment")) {
-                int n = 0;
-                printf("; Commandline arguments:\n");
-                while (n < argc) {
-                    printf(";  %2d: \"%s\"\n", n, argv[n]);
-                    ++n;
-                }
-                n = 0;
-                printf(";\n; Environment:\n");
-                while (ENVIRON[n]) {
-                    printf(";  %2d: \"%s\"\n", n, ENVIRON[n]);
-                    ++n;
-                }
+                debug_environment_p = 1;
                 ++argi;
             } else if (0 == strcmp(arg, "--disable-ldb")) {
                 disable_lossage_handler_p = 1;
@@ -436,6 +569,12 @@ main(int argc, char *argv[], char *envp[])
                 end_runtime_options = 1;
                 ++argi;
                 break;
+            } else if (0 == strcmp(arg, "--merge-core-pages")) {
+                ++argi;
+                merge_core_pages = 1;
+            } else if (0 == strcmp(arg, "--no-merge-core-pages")) {
+                ++argi;
+                merge_core_pages = 0;
             } else {
                 /* This option was unrecognized as a runtime option,
                  * so it must be a toplevel option or a user option,
@@ -450,6 +589,7 @@ main(int argc, char *argv[], char *envp[])
         {
             char *argi0 = argv[argi];
             int argj = 1;
+#ifndef LISP_FEATURE_WIN32
             /* (argc - argi) for the arguments, one for the binary,
                and one for the terminating NULL. */
             sbcl_argv = successful_malloc((2 + argc - argi) * sizeof(char *));
@@ -463,58 +603,105 @@ main(int argc, char *argv[], char *envp[])
                  * error. */
                 if (!end_runtime_options &&
                     0 == strcmp(arg, "--end-runtime-options")) {
-                    lose("bad runtime option \"%s\"\n", argi0);
+                    lose("bad runtime option \"%s\"", argi0);
                 }
                 sbcl_argv[argj++] = arg;
             }
+#else
+            /* The runtime options are processed as chars above, which may
+             * not always work but may be good enough for now, as it
+             * has been for a long time. */
+            int wargc;
+            wchar_t** wargv;
+            wargv = CommandLineToArgvW(GetCommandLineW(), &wargc);
+            sbcl_argv = successful_malloc((2 + wargc - argi) * sizeof(wchar_t *));
+            sbcl_argv[0] = wargv[0];
+            while (argi < wargc) {
+                wchar_t *warg = wargv[argi++];
+                if (!end_runtime_options &&
+                    0 == wcscmp(warg, L"--end-runtime-options")) {
+                    lose("bad runtime option \"%s\"", argi0);
+                }
+                sbcl_argv[argj++] = warg;
+            }
+#endif
             sbcl_argv[argj] = 0;
         }
     }
 
     /* Align down to multiple of page_table page size, and to the appropriate
      * stack alignment. */
-    dynamic_space_size &= ~(PAGE_BYTES-1);
-    thread_control_stack_size &= ~(CONTROL_STACK_ALIGNMENT_BYTES-1);
+    dynamic_space_size &= ~(sword_t)(BACKEND_PAGE_BYTES-1);
+#ifdef LISP_FEATURE_GENCGC
+    dynamic_space_size &= ~(sword_t)(GENCGC_CARD_BYTES-1);
+#endif
+    thread_control_stack_size &= ~(sword_t)(CONTROL_STACK_ALIGNMENT_BYTES-1);
 
-    /* Preserve the runtime options for possible future core saving */
-    runtime_options->dynamic_space_size = dynamic_space_size;
-    runtime_options->thread_control_stack_size = thread_control_stack_size;
-
-    /* KLUDGE: os_vm_page_size is set by os_init(), and on some
-     * systems (e.g. Alpha) arch_init() needs need os_vm_page_size, so
-     * it must follow os_init(). -- WHN 2000-01-26 */
     os_init(argv, envp);
-    arch_init();
+    if (debug_environment_p) {
+        print_environment(argc, argv);
+    }
+    dyndebug_init();
+    // FIXME: if the 'have' flag is 0 and you've disabled disabling of ASLR
+    // then we haven't done an exec(), nor unmapped the mappings that were obtained
+    // already obtained (if any) so it is unhelpful to try again here.
+    allocate_lisp_dynamic_space(have_hardwired_spaces);
     gc_init();
-    validate();
 
     /* If no core file was specified, look for one. */
-    if (!core) {
-        core = search_for_core();
+    if (!core && !(core = search_for_core())) {
+      /* Try resolving symlinks */
+      if (sbcl_runtime) {
+        free(sbcl_runtime_home);
+        char* real = sb_realpath(sbcl_runtime);
+        if (!real)
+          goto lose;
+        sbcl_runtime_home = dir_name(real);
+        free(real);
+        if (!sbcl_runtime_home)
+          goto lose;
+        if(!(core = search_for_core()))
+          goto lose;
+      } else {
+      lose:
+        lose("Can't find sbcl.core");
+      }
     }
 
-    /* Make sure that SBCL_HOME is set and not the empty string,
-       unless loading an embedded core. */
-    if (!(sbcl_home && *sbcl_home) && embedded_core_offset == 0) {
-        char *envstring, *copied_core, *dir;
-        char *stem = "SBCL_HOME=";
-        copied_core = copied_string(core);
-        dir = dirname(copied_core);
-        envstring = (char *) calloc(strlen(stem) +
-                                    strlen(dir) +
-                                    1,
-                                    sizeof(char));
-        sprintf(envstring, "%s%s", stem, dir);
-        putenv(envstring);
-        free(copied_core);
-    }
+    if (embedded_core_offset)
+        lisp_startup_options.noinform = 1;
 
-    if (!noinform && embedded_core_offset == 0) {
+    if (!lisp_startup_options.noinform) {
         print_banner();
         fflush(stdout);
     }
 
-#if defined(SVR4) || defined(__linux__) || defined(__NetBSD__)
+    if (embedded_core_offset == 0) {
+        /* Here we make a last attempt at recognizing an embedded core,
+         * so that a file with an embedded core is a valid argument to
+         * --core.  We take care that any decisions on special behaviour
+         * (suppressed banner, embedded options) have already been made
+         * before we reach this block, so that there is no observable
+         * difference between "embedded" and "bare" images given to
+         * --core. */
+        os_vm_offset_t offset = search_for_embedded_core(core, 0);
+        if (offset != -1)
+            embedded_core_offset = offset;
+    }
+
+    globals_init();
+
+    /* Doing this immediately after the core has been located
+     * and before any random malloc() calls occur improves the chance
+     * of mapping dynamic space at our preferred address (if movable).
+     * If not movable, it was already mapped in allocate_spaces(). */
+    initial_function = load_core_file(core, embedded_core_offset,
+                                      merge_core_pages);
+    if (initial_function == NIL) {
+        lose("couldn't find initial function");
+    }
+
+#if defined(SVR4) || defined(__linux__) || defined(__NetBSD__) || defined(__HAIKU__)
     tzset();
 #endif
 
@@ -524,24 +711,22 @@ main(int argc, char *argv[], char *envp[])
     if (!disable_lossage_handler_p)
         enable_lossage_handler();
 
-    globals_init();
-
-    initial_function = load_core_file(core, embedded_core_offset);
-    if (initial_function == NIL) {
-        lose("couldn't find initial function\n");
-    }
-#ifdef LISP_FEATURE_HPUX
-    /* -1 = CLOSURE_FUN_OFFSET, 23 = SIMPLE_FUN_CODE_OFFSET, we are
-     * not in LANGUAGE_ASSEMBLY so we cant reach them. */
-    return_from_lisp_stub = (void *) ((char *)*((unsigned long *)
-                 ((char *)initial_function + -1)) + 23);
+    os_link_runtime();
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    /* Delayed until after dynamic space has been mapped, fixups made,
+     * and/or immobile-space linkage entries written,
+     * since it was too soon earlier to handle write faults. */
+    write_protect_immobile_space();
 #endif
 
-    gc_initialize_pointers();
-
     arch_install_interrupt_handlers();
 #ifndef LISP_FEATURE_WIN32
     os_install_interrupt_handlers();
+# ifdef LISP_FEATURE_SB_SAFEPOINT
+    ll_install_handler(SIGURG, thruption_handler);
+# elif defined LISP_FEATURE_SB_THREAD
+    ll_install_handler(SIG_STOP_FOR_GC, sig_stop_for_gc_handler);
+# endif
 #else
 /*     wos_install_interrupt_handlers(handler); */
     wos_install_interrupt_handlers(&exception_frame);
@@ -555,14 +740,6 @@ main(int argc, char *argv[], char *envp[])
 
     FSHOW((stderr, "/funcalling initial_function=0x%lx\n",
           (unsigned long)initial_function));
-#ifdef LISP_FEATURE_WIN32
-    fprintf(stderr, "\n\
-This is experimental prerelease support for the Windows platform: use\n\
-at your own risk.  \"Your Kitten of Death awaits!\"\n");
-    fflush(stdout);
-    fflush(stderr);
-#endif
-    create_initial_thread(initial_function);
-    lose("CATS.  CATS ARE NICE.\n");
+    create_main_lisp_thread(initial_function);
     return 0;
 }
