@@ -18,30 +18,23 @@
 #include "os.h"
 #include "arch.h"
 #include "lispregs.h"
-#include "signal.h"
 #include "alloc.h"
 #include "interrupt.h"
 #include "interr.h"
 #include "breakpoint.h"
 #include "thread.h"
-#include "pseudo-atomic.h"
+#include "getallocptr.h"
+#include "forwarding-ptr.h"
+#include "var-io.h"
+#include "code.h"
 
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
+#include "genesis/vector.h"
 
-#define BREAKPOINT_INST 0xcc    /* INT3 */
-#define UD2_INST 0x0b0f         /* UD2 */
-
-#ifndef LISP_FEATURE_UD2_BREAKPOINTS
+#define INT3_INST 0xcc
+#define UD2_INST 0x0b0f
 #define BREAKPOINT_WIDTH 1
-#else
-#define BREAKPOINT_WIDTH 2
-#endif
-
-unsigned long fast_random_state = 1;
-
-void arch_init(void)
-{}
 
 #ifndef LISP_FEATURE_WIN32
 os_vm_address_t
@@ -62,14 +55,16 @@ arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *context)
 int *
 context_eflags_addr(os_context_t *context)
 {
-#if defined __linux__ || defined __sun
+#if defined __linux__
     /* KLUDGE: As of kernel 2.2.14 on Red Hat 6.2, there's code in the
      * <sys/ucontext.h> file to define symbolic names for offsets into
      * gregs[], but it's conditional on __USE_GNU and not defined, so
      * we need to do this nasty absolute index magic number thing
      * instead. */
     return &context->uc_mcontext.gregs[16];
-#elif defined __FreeBSD__
+#elif defined(LISP_FEATURE_SUNOS)
+    return &context->uc_mcontext.gregs[EFL];
+#elif defined(LISP_FEATURE_FREEBSD) || defined(__DragonFly__)
     return &context->uc_mcontext.mc_eflags;
 #elif defined __OpenBSD__
     return &context->sc_eflags;
@@ -78,7 +73,7 @@ context_eflags_addr(os_context_t *context)
 #elif defined __NetBSD__
     return &(context->uc_mcontext.__gregs[_REG_EFL]);
 #elif defined LISP_FEATURE_WIN32
-    return (int *)&context->EFlags;
+    return (int *)&context->win32_context->EFlags;
 #else
 #error unsupported OS
 #endif
@@ -91,28 +86,25 @@ void arch_skip_instruction(os_context_t *context)
      * past it. Skip the code; after that, if the code is an
      * error-trap or cerror-trap then skip the data bytes that follow. */
 
-    int vlen;
     int code;
 
-
     /* Get and skip the Lisp interrupt code. */
     code = *(char*)(*os_context_pc_addr(context))++;
     switch (code)
         {
         case trap_Error:
         case trap_Cerror:
-            /* Lisp error arg vector length */
-            vlen = *(char*)(*os_context_pc_addr(context))++;
-            /* Skip Lisp error arg data bytes. */
-            while (vlen-- > 0) {
-                ++*os_context_pc_addr(context);
-            }
+            skip_internal_error(context);
             break;
 
         case trap_Breakpoint:           /* not tested */
         case trap_FunEndBreakpoint: /* not tested */
             break;
 
+#ifdef LISP_FEATURE_SB_SAFEPOINT
+        case trap_GlobalSafepoint:
+        case trap_CspSafepoint:
+#endif
         case trap_PendingInterrupt:
         case trap_Halt:
         case trap_SingleStepAround:
@@ -139,20 +131,20 @@ arch_internal_error_arguments(os_context_t *context)
 boolean
 arch_pseudo_atomic_atomic(os_context_t *context)
 {
-    return get_pseudo_atomic_atomic(arch_os_get_current_thread());
+    return get_pseudo_atomic_atomic(get_sb_vm_thread());
 }
 
 void
 arch_set_pseudo_atomic_interrupted(os_context_t *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     set_pseudo_atomic_interrupted(thread);
 }
 
 void
 arch_clear_pseudo_atomic_interrupted(os_context_t *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     clear_pseudo_atomic_interrupted(thread);
 }
 
@@ -164,16 +156,8 @@ unsigned int
 arch_install_breakpoint(void *pc)
 {
     unsigned int result = *(unsigned int*)pc;
-
-#ifndef LISP_FEATURE_UD2_BREAKPOINTS
-    *(char*)pc = BREAKPOINT_INST;               /* x86 INT3       */
+    *(char*)pc = INT3_INST;
     *((char*)pc+1) = trap_Breakpoint;           /* Lisp trap code */
-#else
-    *(char*)pc = UD2_INST & 0xff;
-    *((char*)pc+1) = UD2_INST >> 8;
-    *((char*)pc+2) = trap_Breakpoint;
-#endif
-
     return result;
 }
 
@@ -182,9 +166,6 @@ arch_remove_breakpoint(void *pc, unsigned int orig_inst)
 {
     *((char *)pc) = orig_inst & 0xff;
     *((char *)pc + 1) = (orig_inst & 0xff00) >> 8;
-#if BREAKPOINT_WIDTH > 1
-    *((char *)pc + 2) = (orig_inst & 0xff0000) >> 16;
-#endif
 }
 
 /* When single stepping, single_stepping holds the original instruction
@@ -287,7 +268,7 @@ sigtrap_handler(int signal, siginfo_t *info, os_context_t *context)
 
     /* This is just for info in case the monitor wants to print an
      * approximation. */
-    current_control_stack_pointer =
+    access_control_stack_pointer(get_sb_vm_thread()) =
         (lispobj *)*os_context_sp_addr(context);
 
 #ifdef LISP_FEATURE_SUNOS
@@ -311,17 +292,14 @@ sigtrap_handler(int signal, siginfo_t *info, os_context_t *context)
 
 void
 sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context) {
-    /* Triggering SIGTRAP using int3 is unreliable on OS X/x86, so
-     * we need to use illegal instructions for traps.
-     */
-#if defined(LISP_FEATURE_UD2_BREAKPOINTS) && !defined(LISP_FEATURE_MACH_EXCEPTION_HANDLER)
+#ifndef LISP_FEATURE_MACH_EXCEPTION_HANDLER
     if (*((unsigned short *)*os_context_pc_addr(context)) == UD2_INST) {
         *os_context_pc_addr(context) += 2;
         return sigtrap_handler(signal, siginfo, context);
     }
 #endif
     fake_foreign_function_call(context);
-    lose("Unhandled SIGILL");
+    lose("Unhandled SIGILL at %p.", (void*)*os_context_pc_addr(context));
 }
 #endif /* not LISP_FEATURE_WIN32 */
 
@@ -341,40 +319,89 @@ arch_install_interrupt_handlers()
      * CL way, I hope there will at least be a comment to explain
      * why.. -- WHN 2001-06-07 */
 #if !defined(LISP_FEATURE_WIN32) && !defined(LISP_FEATURE_MACH_EXCEPTION_HANDLER)
-    undoably_install_low_level_interrupt_handler(SIGILL , sigill_handler);
-    undoably_install_low_level_interrupt_handler(SIGTRAP, sigtrap_handler);
+    ll_install_handler(SIGILL , sigill_handler);
+    ll_install_handler(SIGTRAP, sigtrap_handler);
 #endif
     SHOW("returning from arch_install_interrupt_handlers()");
 }
 
-#ifdef LISP_FEATURE_LINKAGE_TABLE
-/* FIXME: It might be cleaner to generate these from the lisp side of
- * things.
- */
 
 void
-arch_write_linkage_table_jmp(char * reloc, void * fun)
+gencgc_apply_code_fixups(struct code *old_code, struct code *new_code)
 {
+    char* code_start_addr = code_text_start(new_code);
+    os_vm_size_t displacement = (char*)new_code - (char*)old_code;
+    lispobj fixups = new_code->fixups;
+    /* It will be a nonzero integer if valid, or 0 if there are no fixups */
+    if (fixups == 0)
+        return;
+
+    /* Could be pointing to a forwarding pointer. */
+    /* This is extremely unlikely, because the only referent of the fixups
+       is usually the code itself; so scavenging the vector won't occur
+       until after the code object is known to be live. As we're just now
+       enlivening the code, the fixups shouldn't have been forwarded.
+       Maybe the vector is on the special binding stack though ... */
+    if (is_lisp_pointer(fixups) &&
+        forwarding_pointer_p(native_pointer(fixups)))  {
+        /* If so, then follow it. */
+        /*SHOW("following pointer to a forwarding pointer");*/
+        fixups = forwarding_pointer_value(native_pointer(fixups));
+    }
+
+    if (fixnump(fixups) ||
+        (lowtag_of(fixups) == OTHER_POINTER_LOWTAG
+         && widetag_of(native_pointer(fixups)) == BIGNUM_WIDETAG)) {
+        /* Got the fixups for the code block. Now work through them
+           in order, first the absolute ones, then the relative.
+           Locations are sorted and delta-encoded for compactness. */
+        struct varint_unpacker unpacker;
+        varint_unpacker_init(&unpacker, fixups);
+        int prev_offset = 0, offset;
+        // Absolute fixups all refer to this object itself (the code
+        // boxed constants). Add this object's displacement to the
+        // value that currently exists at the fixup location.
+        while (varint_unpack(&unpacker, &offset) && offset != 0) {
+            offset += prev_offset;
+            prev_offset = offset;
+            *(char**)(code_start_addr + offset) += displacement;
+        }
+        prev_offset = 0;
+        // Relative fixups: assembly and foreign routines. Subtract this
+        // object's displacement from the value that currently exists.
+        while (varint_unpack(&unpacker, &offset) && offset != 0) {
+            offset += prev_offset;
+            prev_offset = offset;
+            *(char**)(code_start_addr + offset) -= displacement;
+        }
+    } else {
+        /* This used to just print a note to stderr, but bogus fixups seem to
+         * indicate real heap corruption, so a hard failure is in order. */
+        lose("fixup vector %x has a bad widetag: %#x",
+             fixups, widetag_of(native_pointer(fixups)));
+    }
+}
+
+void
+arch_write_linkage_table_entry(int index, void *target_addr, int datap)
+{
+    char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
+    if (datap) {
+        *(unsigned long *)reloc_addr = (unsigned long)target_addr;
+        return;
+    }
     /* Make JMP to function entry. JMP offset is calculated from next
      * instruction.
      */
-    long offset = (char *)fun - (reloc + 5);
+    long offset = (char *)target_addr - (reloc_addr + 5);
     int i;
 
-    *reloc++ = 0xe9;            /* opcode for JMP rel32 */
+    *reloc_addr++ = 0xe9;       /* opcode for JMP rel32 */
     for (i = 0; i < 4; i++) {
-        *reloc++ = offset & 0xff;
+        *reloc_addr++ = offset & 0xff;
         offset >>= 8;
     }
 
     /* write a nop for good measure. */
-    *reloc = 0x90;
+    *reloc_addr = 0x90;
 }
-
-void
-arch_write_linkage_table_ref(void * reloc, void * data)
-{
-    *(unsigned long *)reloc = (unsigned long)data;
-}
-
-#endif
