@@ -9,14 +9,61 @@
 ;;;; absolutely no warranty. See the COPYING and CREDITS files for
 ;;;; more information.
 
-(in-package :cl-user)
-
-(use-package :test-util)
-(use-package :assertoid)
-
 (defstruct foo)
 (defstruct bar x y)
 
+(defvar *things*)
+(defun make-things-for-sxhash-test (n)
+  (setf *things* (make-array n :fill-pointer 0))
+  (flet ((store (obj)
+           (vector-push-extend (list (sb-kernel:get-lisp-obj-address obj)
+                                     (sxhash obj)
+                                     obj)
+                               *things*)))
+    #-(or x86 x86-64) ; precise GC
+    (return-from make-things-for-sxhash-test (dotimes (i n t) (store (make-foo))))
+    #+sb-thread
+    (sb-thread:join-thread
+     (sb-thread:make-thread
+      (lambda () (dotimes (i n t) (store (make-foo))))))))
+(compile 'make-things-for-sxhash-test)
+
+;;; Assert that the C code which computes a perturbation of the object
+;;; address for lazy stable address-based hashing is the same as lisp.
+;;; Further, assert that each bit of the resulting positive-fixnum
+;;; can be in a 0 and 1 state (don't want any bits stuck at 0).
+(with-test (:name :address-based-sxhash-gcing)
+  (dotimes (runs 5)
+    (let ((tracker (make-array 4 :element-type 'sb-ext:word
+                               :initial-element 0)))
+      ;; with this many sxhashes we should see a 1 bit in
+      ;; in each bit position.
+      (when (make-things-for-sxhash-test 20)
+        (gc)
+        (sb-int:dovector (thing *things*)
+          (destructuring-bind (old-addr old-hash object) thing
+            (let* ((new-addr (sb-kernel:get-lisp-obj-address object))
+                   (new-hash (sxhash object)))
+              (setf (aref tracker 0) (logior (aref tracker 0) old-hash)
+                    (aref tracker 1) (logior (aref tracker 1)
+                                             (logxor old-hash most-positive-fixnum)))
+              (let* ((count-1s (logcount new-hash))
+                     (count-0s (- sb-vm:n-positive-fixnum-bits count-1s)))
+                (incf (aref tracker 2) count-1s)
+                (incf (aref tracker 3) count-0s))
+              (cond ((= new-addr old-addr)
+                     (warn "Can't test SXHASH after movement: didn't move"))
+                    ((not (eql new-hash old-hash))
+                     (error "SXHASH failure"))))))
+        ;; show the mask where we saw 1 bits (respectively 0),
+        ;; and total number of 1 (respectively 0) bits.)
+        #+nil
+        (format t "~@{[~64,'0b] ~d~%~}"
+                (aref tracker 0) (aref tracker 2)
+                (aref tracker 1) (aref tracker 3))
+        (assert (= (aref tracker 0) most-positive-fixnum))
+        (assert (= (aref tracker 1) most-positive-fixnum))))))
+
 ;;; SXHASH and PSXHASH should distribute hash values well over the
 ;;; space of possible values, so that collisions between the hash
 ;;; values of unequal objects should be very uncommon. (Except of
@@ -261,36 +308,27 @@
 
 ;;; This test works reliably on non-conservative platforms and
 ;;; somewhat reliably on conservative platforms with threads.
-#+(or (not (or x86 x86-64)) sb-thread)
-(progn
-
-(defparameter *ht* nil)
-
-(defvar *cons-here*)
-
-(declaim (notinline args))
-(defun take (&rest args)
-  (declare (ignore args)))
+(defun call-and-scrub-stack (thunk &aux results)
+  ;; Start by giving ourselves some headroom via a
+  ;; DX allocation around the actual allocation...
+  (let ((*s* (make-array 25)))
+    (declare (special *s*)
+             (dynamic-extent *s*))
+    (assert (stack-allocated-p *s*))
+    (setq results (multiple-value-list (funcall thunk))))
+  ;; ... and then arrange to have the no-longer-used parts of the
+  ;; control stack cleared.
+  (sb-sys:scrub-control-stack)
+  (values-list results))
+(compile 'call-and-scrub-stack)
 
 (defmacro alloc (&body body)
   "Execute BODY and try to reduce the chance of leaking a conservative root."
-  #-sb-thread
-  `(multiple-value-prog1
-       (progn ,@body)
-     (loop repeat 20000 do (setq *cons-here* (cons nil nil)))
-     ;; KLUDGE: Clean the argument passing regs.
-     (apply #'take (loop repeat 36 collect #'cons)))
-  #+sb-thread
-  (let ((values (gensym))
-        (sem (gensym)))
-    `(let ((,sem (sb-thread::make-semaphore))
-           ,values)
-       (sb-thread:make-thread (lambda ()
-                                (setq ,values
-                                      (multiple-value-list (progn ,@body)))
-                                (sb-thread::signal-semaphore ,sem)))
-       (sb-thread::wait-on-semaphore ,sem)
-       (values-list ,values))))
+  #+sb-thread `(sb-thread:join-thread
+                (sb-thread:make-thread
+                 (lambda ()
+                   (call-and-scrub-stack (lambda () ,@body)))))
+  #-sb-thread `(call-and-scrub-stack (lambda () ,@body)))
 
 (with-test (:name (:hash-table :weakness :eql :numbers))
   (flet ((random-number ()
@@ -324,39 +362,48 @@
     (values)))
 
 (defun print-ht (ht &optional (stream t))
-  (format stream "Weakness: ~S~%" (sb-impl::hash-table-weakness ht))
-  (format stream "Table: ~S~%" (sb-impl::hash-table-table ht))
+  (format stream "Weakness: ~S~%" (hash-table-weakness ht))
+  (format stream "Table: ~S~%" (sb-impl::hash-table-pairs ht))
   (format stream "Next: ~S~%" (sb-impl::hash-table-next-vector ht))
   (format stream "Index: ~S~%" (sb-impl::hash-table-index-vector ht))
   (format stream "Hash: ~S~%" (sb-impl::hash-table-hash-vector ht))
   (force-output stream))
 
-(with-test (:name (:hash-table :weakness :removal))
-  (loop for test in '(eq eql equal equalp) do
-        (format t "test: ~A~%" test)
-        (loop for weakness in '(:key :value :key-and-value :key-or-value)
-              do
-              (format t "weakness: ~A~%" weakness)
-              (let ((ht (make-hash-table :test 'equal :weakness weakness)))
+(macrolet
+    ((test-weakness (weakness)
+       `(with-test (:name (:hash-table :weakness ,weakness :removal))
+          (loop for test in '(eq eql equal equalp) do
+            (let ((ht (make-hash-table :test 'equal :weakness ,weakness)))
                 (alloc (add-removable-stuff ht :n 117 :size 1))
                 (loop for i upfrom 0
-                      do (format t "~A. count: ~A~%" i (hash-table-count ht))
+                      do ; (format t "~A. count: ~A~%" i (hash-table-count ht))
                       (force-output)
                       until (zerop (hash-table-count ht))
                       do
                       (when (= i 10)
-                        (print-ht ht)
+                        ; (print-ht ht)
                         #-(or x86 x86-64)
                         (assert nil)
                         ;; With conservative gc the test may not be
                         ;; bullet-proof so it's not an outright
                         ;; failure but a warning.
                         #+(or x86 x86-64)
-                        (progn
-                          (warn "Weak hash removal test failed for weakness ~A"
-                                weakness)
-                          (return)))
-                      (gc :full t))))))
+                        (if (eq *evaluator-mode* :compile)
+                            (assert nil)
+                            (progn
+                              (warn "Weak hash removal test failed for weakness ~A"
+                                    ,weakness)
+                              (return))))
+                      (gc :full t)))))))
+  ;; I separated these into 4 named tests to see if I could figure something out-
+  ;; If the interpreted lambda itself it kept alive by call-and-scrub,
+  ;; as it must be, the most recent values of local variables could linger.
+  ;; So with x86 using the interpreter, :KEY weakness generates a failure warning
+  ;; but how does :KEY-OR-VALUE _not_ generate a warning?
+  (test-weakness :key)
+  (test-weakness :value)
+  (test-weakness :key-and-value)
+  (test-weakness :key-or-value))
 
 (with-test (:name (:hash-table :weakness :string-interning))
   (let ((ht (make-hash-table :test 'equal :weakness :key))
@@ -389,8 +436,6 @@
                 (setf (gethash key h1) value))
           (sb-ext:gc :full t))))
 
-)
-
 ;;; DEFINE-HASH-TABLE-TEST
 
 (defstruct custom-hash-key name)
@@ -400,7 +445,7 @@
 (defun custom-hash-hash (x)
   (sxhash (custom-hash-key-name x)))
 (define-hash-table-test custom-hash-test custom-hash-hash)
-(with-test (:name define-hash-table-test.1)
+(with-test (:name :define-hash-table-test.1)
   (let ((table (make-hash-table :test 'custom-hash-test)))
     (setf (gethash (make-custom-hash-key :name "foo") table) :foo)
     (setf (gethash (make-custom-hash-key :name "bar") table) :bar)
@@ -421,7 +466,7 @@
     (lambda (x)
       (logand most-positive-fixnum
               (reduce #'+ (map 'list #'sxhash (subseq x 0 3))))))
-(with-test (:name define-hash-table-test.2)
+(with-test (:name :define-hash-table-test.2)
   (let ((table (make-hash-table :test 'head-eql)))
     (setf (gethash #(1 2 3 4) table) :|123|)
     (setf (gethash '(2 3 4 7) table) :|234|)
@@ -439,7 +484,7 @@
     (assert (eq :foo (gethash '(#\f #\o #\o 1 2 3) table)))
     (assert (eq 'head-eql (hash-table-test table)))))
 
-(with-test (:name make-hash-table/hash-fun)
+(with-test (:name :make-hash-table/hash-fun)
   (let ((table (make-hash-table
                 :test #'=
                 :hash-function (lambda (x)
@@ -452,4 +497,55 @@
     (assert (= 5 (gethash 1 table)))
     (assert (eq '= (hash-table-test table)))))
 
+(defstruct rslotty a
+  (uword 0 :type word)
+  (sword 0 :type sb-vm:signed-word)
+  (sf 0s0 :type single-float)
+  (df 0d0 :type double-float)
+  (csf #c(0s0 0s0) :type (complex single-float))
+  (cdf #c(0d0 0d0) :type (complex double-float)))
+
+(import 'sb-impl::psxhash)
+(with-test (:name :psxhash-raw-slots)
+  (let ((empty (psxhash (make-rslotty))))
+    ;; unequalp values produce unequal hashes
+    (assert (/= empty (psxhash (make-rslotty :uword 32))))
+    (assert (/= empty (psxhash (make-rslotty :sword -1800))))
+    (assert (/= empty (psxhash (make-rslotty :sf 1s0))))
+    (assert (/= empty (psxhash (make-rslotty :df 1d0))))
+    (assert (/= empty (psxhash (make-rslotty :csf #c(1s0 1s0)))))
+    (assert (/= empty (psxhash (make-rslotty :cdf #c(1d0 1d0)))))
+    ;; equalp values produce equal hashes
+    (assert (= empty (psxhash (make-rslotty :sf -0s0 :df -0d0))))
+    (assert (= empty (psxhash (make-rslotty :csf #c(-0s0 -0s0)))))
+    (assert (= empty (psxhash (make-rslotty :csf #c(0s0 -0s0)))))
+    (assert (= empty (psxhash (make-rslotty :csf #c(-0s0 0s0)))))
+    (assert (= empty (psxhash (make-rslotty :cdf #c(-0d0 -0d0)))))
+    (assert (= empty (psxhash (make-rslotty :cdf #c(0d0 -0d0)))))
+    (assert (= empty (psxhash (make-rslotty :cdf #c(-0d0 0d0)))))))
+
+(defun my= (a b) (= a b))
+(defun fixnum-hash (x) (sxhash (the fixnum x)))
+(defun fixnum-hash-worse (x) (logand (sxhash (the fixnum x)) 7))
+(define-hash-table-test my= fixnum-hash)
+
+(with-test (:name :hash-fun-is-function-designator)
+  ;; Users shouldn't write this baroque expression to make an EQL table.
+  (assert-error (make-hash-table :hash-function nil))
+  ;; nor this
+  (assert-error (make-hash-table :test #'eql :hash-function nil))
+  ;; :TEST, if unknown, does not imply a hash function
+  ;; even when it looks like it could.
+  (assert-error (make-hash-table :test #'=))
+  ;; and of course this doesn't work either because the preceding doesn't
+  (assert-error (make-hash-table :test #'= :hash-function nil))
+  ;; Try user functions
+  (let ((h (make-hash-table :test 'my=)))
+    (assert (eq (sb-impl::hash-table-hash-fun h) #'fixnum-hash)))
+  (let ((h (make-hash-table :test 'my= :hash-function 'fixnum-hash-worse)))
+    (assert (eq (sb-impl::hash-table-hash-fun h) #'fixnum-hash-worse)))
+  (let ((h (make-hash-table :test 'my= :hash-function #'fixnum-hash-worse)))
+    (assert (eq (sb-impl::hash-table-hash-fun h) #'fixnum-hash-worse)))
+  (assert-error (make-hash-table :test 'my= :hash-function nil))) ; no good
+
 ;;; success
